{*******************************************************}
{                                                       }
{       Turbo Pascal Version 7.0                        }
{       FreePacal                                       }
{       Delphi 4                                        }
{       Large text editor                               }
{     Text size > 10 Mb in real mode!                   }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}

{$I Define.inc}

unit Ext_Edit;
interface
uses Objects, Views, VMM, Ext_Type, Ext_Math, Drivers, Register, AddCntrl,
     ShardEdt, Ext_Clp, Clases, Ext_Sys;

const
  DefaultSaveExt = '.txt';
const
  smExtend = $01;
  smDouble = $02;
  lfSelectionChanged     = $0001;

const
{$IFNDEF PASCAL16}
   DefHalfEditorBlockSize   = defVMMPageSize div 2;
{$ELSE}
   DefHalfEditorBlockSize   = defVMMPageSize - SizeOf(String);
{$ENDIF}

type
  PTextBlock = ^TTextBlock;
  TTextBlock = object(TObject)
    TextSize:word;
    CountLine:word;
    MaxLineWidth:integer;
    constructor Init;
    constructor Load(var AFile:Text);
    destructor Done;virtual;
    procedure Save(AStream:PStream);
    procedure Update;
    function Str(Number:word):string;
    procedure ReplaseStr(Number:word;AStr:string);
    procedure DeleteStr(Number:word);
    procedure InsertStr(Number:word;AStr:string);
    function FindString(AStr:string; var X, Y:Longint):boolean;
  private
    Handle:TPtrHandle;
    Buffer:PCharArray;
    function StrPtr(Number:word):word;
    function StrLen(AStrPtr:word):word;
    procedure FindMaxLine;
    procedure MoveTo(Dest:PTextBlock; ALineNo:word); {७ᨬ 稭  CountLine}
  end;

type
  PFileObject = ^TFileObject;
  TFileObject = object(TObject)
    FileName:PString;
    Modified:boolean;
    CountLine:Longint;
    constructor Init(AFileName:FNameStr);
    destructor Done;virtual;
    function Str(Number:Longint):string;
    procedure ReplaseStr(Number:Longint; AStr:string);
    procedure DeleteStr(Number:Longint);
    procedure InsertStr(Number:Longint; AStr:string);
    function Save:boolean;
    function MaxLineWidth:integer;
    function FindString(AStr:string; SelectStart, SelectEnd:TLargePoint; var APos:TLargePoint):boolean;
    procedure DeleteText(TextStart, TextEnd:TLargePoint);
  private
    Editors:PCollection;
    EditorsPages:PCollection;
    function GetTextBlock(var Number:Longint; var Delta:Longint):PTextBlock;
    function fCountLine:Longint;
  end;

type
  PExtIndicator = ^TExtIndicator;
  TExtIndicator = object(TView)
    Location:TLargePoint;
    Modified:Boolean;
    constructor Init(var Bounds: TRect);
    procedure Draw; virtual;
    function GetPalette: PPalette; virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    procedure SetValue(ALocation: TLargePoint; AModified: Boolean);
  end;

type
  TDrawArray = array [0..MaxLineLength - 1] of Word;
type
  TEditorBookMark = record
    Valid  : boolean;
    Pos    : TLargePoint;
  end;
  TEditorBookmarkArray = array[0..9] of TEditorBookmark;

  PExtEditor = ^TExtEditor;
  TDrawColorEvent=procedure(Sender:PExtEditor; const S:string; var B:TDrawBuffer; Y: Longint);
  TExtEditor = object(TView)
    HScrollBar:PLargeScrollBar;
    VScrollBar:PLargeScrollBar;
    Indicator :PExtIndicator;
    CurPos    :TLargePoint;
    Delta     :TLargePoint;
    Limit     :TLargePoint;
    Overwrite :Boolean;
    SelectStart  :TLargePoint;
    SelectEnd    :TLargePoint;
    ReadOnly  :boolean;
    constructor Init(var Bounds: TRect; AHScrollBar, AVScrollBar:
       PLargeScrollBar; AIndicator: PExtIndicator; AFileName:FNameStr);
    constructor Load(var S: TStream);
    destructor Done;virtual;
    procedure Draw;virtual;
    procedure DrawLines(FirstLine, Count: Longint; DrawAll:boolean);
    procedure SetCurPos(X,Y: Longint);virtual;
    procedure ChangeBounds(var Bounds: TRect); virtual;
    function GetPalette: PPalette; virtual;
    procedure ConvertEvent(var Event: TEvent);
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure ScrollDraw; virtual;
    procedure ScrollTo(X, Y: Longint);
    procedure SetLimit(X, Y: Longint);
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    procedure UpdateIndicator;
    procedure LimitsChanged;
    function IsClipboard:boolean;
    function FileName:FNameStr;
    procedure UpdateCommands;virtual;
    function GetCurWord:string;

    procedure LineDown;
    procedure LineUp;
    procedure CharLeft;
    procedure CharRight;
    procedure LineStart;
    procedure LineEnd;
    procedure PageUp;
    procedure PageDown;
    procedure TextStart;
    procedure TextEnd;
    procedure WordRight;
    procedure WordLeft;
    procedure JumpMark(MarkIdx: integer);
    procedure DefineMark(MarkIdx: integer);
    procedure GotoLine;
    procedure TrackCursor(Center: boolean);
    procedure Find;
    procedure Replace;
    procedure DoSearchReplace;

    procedure ToggleInsMode;
    function Modified:boolean;

    procedure AddChar(const C:char; SelectText : Boolean);
    procedure DelChar;
    procedure DelWord;
    procedure DelStart;
    procedure DelEnd;
    procedure DelLine;
    procedure BackSpace;
    function NewLine: integer;
    function InsertLine(S: string; SelectText : Boolean): Boolean;

    procedure ClipPaste;
    procedure ClipCut;
    procedure ClipCopy;
    procedure CopyBlock;
    procedure MoveBlock;
    procedure IndentBlock;
    procedure UnindentBlock;
    procedure DelSelect;
    procedure WriteBlock;
    procedure ReadBlock;
    procedure HideSelect;
    procedure SetSelection(A, B: TLargePoint);
    function GetSelectedText:PStrings;
    procedure SetSelectedText(P:PStrings);
    procedure EndSelect;
    procedure StartSelect;
    function InsertFrom(Editor: PExtEditor): Boolean;
    procedure Store(var S: TStream);
    function isSelected:boolean;

  private
    FileObject:PFileObject;
    KeyState:Integer;
    Bookmarks:TEditorBookmarkArray;
    LockCount:integer;
    LockDrawLine:Longint;
    LockDrawCount:Longint;
    TempSelectStart  :TLargePoint;
    TempSelectEnd    :TLargePoint;
    FindTextSelected:boolean;
    LockFlag:word;
    NoSelect:boolean;

    procedure DrawCur;
    procedure CheckSels;
    procedure SelectionChanged;
    procedure Lock;
    procedure UnLock;
    function Search(var FindStr:string):boolean;
    procedure ChangeSelected(Count:integer);
  public
    OnDrawColorEvent:TDrawColorEvent;
  end;

type
  PExtFileEditor = ^TExtFileEditor;
  TExtFileEditor = object(TExtEditor)
    destructor Done;virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    function Valid(Command: Word): Boolean; virtual;
    function Save:boolean;
    function SaveAs:boolean;
    constructor Load(var S: TStream);
    procedure Store(var S: TStream);
    procedure UpdateCommands;virtual;
    function SaveFile(AFileName:FNameStr):boolean;
  end;

type
  PExtEditWindow = ^TExtEditWindow;
  TExtEditWindow = object(TWindow)
    Editor:PExtFileEditor;
    constructor Init(var Bounds: TRect;
                          FileName: FNameStr; ANumber: Integer);
    function GetTitle(MaxSize: Integer): TTitleStr; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Close; virtual;
    constructor Load(var S: TStream);
    procedure Store(var S: TStream);
    procedure Awaken; virtual;
    procedure SizeLimits(var Min, Max: TPoint);virtual;
  end;

type
  PStandartClipboard = ^TStandartClipboard;
  TStandartClipboard = object(TClipObject)
    constructor Init;
    destructor Done;virtual;
    procedure CopyToClp(var Buf; BufType:word; Len:word);virtual;
    procedure PasteFromClp(var Buf; BufType:word);virtual;
    procedure ClearClp;virtual;
    function DataSize(BufType:word):word;virtual;
  end;

type
  TEditorDialog = function(Dialog: Integer; Info: Pointer): Word;

function CreateEditWindow(AFileName:FNameStr; AHelpCtx:word):PExtEditWindow;
function StdEditorDialog(Dialog: Integer; Info: Pointer): word;far;
function DefEditorDialog(Dialog: Integer; Info: Pointer): Word;far;
procedure RegisterExtEditor;
procedure InitStandartClipboard;

const
{$IFDEF SINGLE}
  EditorDialog : TEditorDialog = DefEditorDialog;
{$ELSE}
  EditorDialog : TEditorDialog = StdEditorDialog;
{$ENDIF}
  Clipboard    : PExtEditor = nil;

{$IFNDEF DELPHI32}
const
  RExtEditor: TStreamRec = (
    ObjType: rcExtEditor;
    VmtLink: Ofs(TypeOf(TExtEditor)^);
    Load: @TExtEditor.Load;
    Store: @TExtEditor.Store
  );
{  RMemo: TStreamRec = (
    ObjType: 71;
    VmtLink: Ofs(TypeOf(TMemo)^);
    Load: @TMemo.Load;
    Store: @TMemo.Store
  );}
  RExtFileEditor: TStreamRec = (
    ObjType: rcExtFileEditor;
    VmtLink: Ofs(TypeOf(TExtFileEditor)^);
    Load: @TExtFileEditor.Load;
    Store: @TExtFileEditor.Store
  );
  RExtIndicator: TStreamRec = (
    ObjType: rcExtIndicator;
    VmtLink: Ofs(TypeOf(TExtIndicator)^);
    Load: @TExtIndicator.Load;
    Store: @TExtIndicator.Store
  );
  RExtEditWindow: TStreamRec = (
    ObjType: rcExtEditWindow;
    VmtLink: Ofs(TypeOf(TExtEditWindow)^);
    Load: @TExtEditWindow.Load;
    Store: @TExtEditWindow.Store
  );
{$ENDIF}
implementation
uses Ext_Dos, ResTools, App, Ext_Str, MsgBox, Dos, StdDlg, Std_Str;

{$I EXT_EDIT.INC}

  { TExtEditor }

constructor TExtEditor.Init(var Bounds: TRect; AHScrollBar, AVScrollBar:
       PLargeScrollBar; AIndicator: PExtIndicator; AFileName:FNameStr);
var P:PExtEditor;
begin
  inherited Init(Bounds);
  LockDrawLine:=MaxLongint;
  GrowMode := gfGrowHiX + gfGrowHiY;
  Options := Options or ofSelectable;

  EventMask := evMouseDown + evKeyDown + evCommand + evBroadcast;
  ShowCursor;
  HScrollBar := AHScrollBar;
  VScrollBar := AVScrollBar;
  Indicator := AIndicator;
  if AFileName='' then P:=nil
  else P:=Message(Application, evBroadcast, cmFindByName, @AFileName);
  if P=nil then FileObject:=New(PFileObject, Init(AFileName))
  else FileObject:=P^.FileObject;
  FileObject^.Editors^.Insert(@Self);
  LimitsChanged;
  UpdateIndicator;
end;

destructor TExtEditor.Done;
begin
  if isClipboard then Clipboard:=nil;
  SetLimit(0,0);
  if FileObject^.Editors^.Count>1 then FileObject^.Editors^.Delete(@Self)
  else Dispose(FileObject, Done);
  inherited Done;
end;

constructor TExtEditor.Load(var S: TStream);
var AFileName:FNameStr;
    P:PExtEditor;
begin
  inherited Load(S);
  GetPeerViewPtr(S, HScrollBar);
  GetPeerViewPtr(S, VScrollBar);
  GetPeerViewPtr(S, Indicator);
  S.Read(CurPos,SizeOf(TLargePoint));
  S.Read(Delta,SizeOf(TLargePoint));
  S.Read(Limit,SizeOf(TLargePoint));
  S.Read(Overwrite,SizeOf(Boolean));
  S.Read(SelectStart,SizeOf(TLargePoint));
  S.Read(SelectEnd,SizeOf(TLargePoint));
  S.Read(ReadOnly,SizeOf(boolean));
  S.Read(Bookmarks,SizeOf(TEditorBookmarkArray));
  AFileName:=ReadString(S);
{$IFNDEF FPC}
  LoadProcedure(S,@OnDrawColorEvent);
{$ENDIF}
  if AFileName='' then P:=nil
  else P:=Message(Application, evBroadcast, cmFindByName, @AFileName);
  if P=nil then FileObject:=New(PFileObject, Init(AFileName))
  else FileObject:=P^.FileObject;
  FileObject^.Editors^.Insert(@Self);
end;

procedure TExtEditor.Store(var S: TStream);
begin
  inherited Store(S);
  PutPeerViewPtr(S, HScrollBar);
  PutPeerViewPtr(S, VScrollBar);
  PutPeerViewPtr(S, Indicator);
  S.Write(CurPos,SizeOf(TLargePoint));
  S.Write(Delta,SizeOf(TLargePoint));
  S.Write(Limit,SizeOf(TLargePoint));
  S.Write(Overwrite,SizeOf(Boolean));
  S.Write(SelectStart,SizeOf(TLargePoint));
  S.Write(SelectEnd,SizeOf(TLargePoint));
  S.Write(ReadOnly,SizeOf(boolean));
  S.Write(Bookmarks,SizeOf(TEditorBookmarkArray));
  WriteString(S, FileName);
{$IFNDEF FPC}
  StoreProcedure(S,@OnDrawColorEvent);
{$ENDIF}
end;

function TExtEditor.FileName:FNameStr;
begin
  if FileObject^.FileName<>nil then FileName:=FileObject^.FileName^
  else FileName:='';
end;

procedure TExtEditor.LimitsChanged;
begin
  SetLimit(MaxLineLength, FileObject^.CountLine);
end;

procedure TExtEditor.DrawLines(FirstLine, Count: Longint; DrawAll:boolean);
var
  B: TDrawBuffer;
  S:string;
  MinX,MaxX,DX,i,ScreenX:Longint;
  Color,
  NormalColor,
  SelectColor: word;
  StartY,EndY:Longint;

procedure DoUpdate(P:PExtEditor);
begin
  if P<>@Self then P^.DrawLines(P^.Delta.Y,P^.Size.Y-1,false);
end;

begin
  if LockCount=0 then
  begin
    NormalColor:=GetColor(1);
    SelectColor:=GetColor(2);
    StartY:=Max(FirstLine, Delta.Y);
    EndY:=StartY+Min(Size.Y-1,Count+FirstLine-Delta.Y);
    Color:=NormalColor;
    for i:=StartY to EndY do
    begin
      S:=FileObject^.Str(I);
      if (SelectStart.X=SelectEnd.X) and (SelectStart.Y=SelectEnd.Y) then Color:=NormalColor
      else
      begin
        if (I>SelectStart.Y) and (I<SelectEnd.Y) then Color:=SelectColor
        else
          if (I=SelectStart.Y) or (I=SelectEnd.Y) then
          begin
            MoveChar(B, ' ', NormalColor, Size.X);
            MoveStr(B, Copy(S,Delta.X,Size.X), NormalColor);
            if I=SelectStart.Y then MinX:=SelectStart.X else MinX:=0;
            if I=SelectEnd.Y then MaxX:=SelectEnd.X-1 else MaxX:=255;
            for DX:=MinX to MaxX do
            begin
              ScreenX:=DX-(Delta.X);
              if (ScreenX>=0) and (ScreenX<Size.X) then
                B[ScreenX]:=(B[ScreenX] and $00ff) or (SelectColor shl 8);
            end;
            if Assigned(OnDrawColorEvent) then
               OnDrawColorEvent(@Self, stUpCase(S), B, I);
            WriteLine(0, i-Delta.Y, Size.X, 1, B);
            Continue;
          end
          else Color:=NormalColor;
      end;
      MoveChar(B, ' ', Color, Size.X);
      MoveStr(B, Copy(S,Delta.X,Size.X), Color);
      if Assigned(OnDrawColorEvent) then
         OnDrawColorEvent(@Self, stUpCase(S), B, I);
      WriteLine(0, i-Delta.Y, Size.X, 1, B);
    end;
    if (FileObject^.Editors^.Count>1) and DrawAll then
      for i:=0 to FileObject^.Editors^.Count-1 do
        DoUpdate(FileObject^.Editors^.At(i));
  end
  else {᫨ ᮢ ஢ -   न ᮢ}
  begin {LockDrawCount=0 -    ࠧ  ᨫ ᮢ       }
    if LockDrawCount=0 then
    begin
      LockDrawCount:=Count;
      LockDrawLine:=FirstLine;
    end
    else
    begin
      LockDrawCount:=Max(LockDrawLine+LockDrawCount,FirstLine+Count){-
                     Min(LockDrawLine, FirstLine)};
      LockDrawLine:=Min(LockDrawLine, FirstLine);
      LockDrawCount:=LockDrawCount-LockDrawLine;
    end;
  end;
end;

procedure TExtEditor.Lock;
begin
  inc(LockCount);
end;

procedure TExtEditor.UnLock;
begin
  if LockCount>0 then Dec(LockCount);
  if LockCount=0 then
  begin
    if LockDrawCount>0 then DrawLines(LockDrawLine,LockDrawCount,true);
    LockDrawLine:=MaxLongint;
    LockDrawCount:=0;
    if LockFlag and lfSelectionChanged<>0 then SelectionChanged;
    LockFlag:=0;
  end;
  if State and sfActive <> 0 then UpdateCommands;
end;

procedure TExtEditor.Draw;
begin
  DrawLines(Delta.Y,Size.Y-1,true);
end;

procedure TExtEditor.TrackCursor(Center: boolean);
var D: TLargePoint;
begin
  D:=Delta;
  if CurPos.Y<Delta.Y then D.Y:=CurPos.Y else
   if CurPos.Y>Delta.Y+Size.Y-1 then D.Y:=CurPos.Y-Size.Y+1;
  if CurPos.X<Delta.X then D.X:=CurPos.X else
   if CurPos.X>Delta.X+Size.X-1 then D.X:=CurPos.X-Size.X+1;
  if {((Delta.X<>D.X) or (Delta.Y<>D.Y)) and }Center then
  begin
     { loose centering for debugger PM }
     while (CurPos.Y-D.Y)<(Size.Y div 3) do Dec(D.Y);
     while (CurPos.Y-D.Y)>2*(Size.Y div 3) do Inc(D.Y);
  end;
  if (Delta.X<>D.X) or (Delta.Y<>D.Y) then ScrollTo(D.X,D.Y);
  DrawCur;
  UpdateIndicator;
end;

procedure TExtEditor.SetCurPos(X,Y: Longint);
var OldPos,OldSEnd,OldSStart: TLargePoint;
    Extended: boolean;
begin
  Lock;
  if FindTextSelected then
  begin
    SelectStart:=TempSelectStart;
    SelectEnd:=TempSelectEnd;
    FindTextSelected:=false;
    DrawView;
  end;
  X:=Max(0,Min(MaxLineLength+1,X));
  Y:=Max(0,Min(FileObject^.CountLine-1,Y));
  OldPos:=CurPos;
  OldSEnd:=SelectEnd;
  OldSStart:=SelectStart;
  CurPos.X:=X;
  CurPos.Y:=Y;
  TrackCursor(false);
  if (not NoSelect) and (ShouldExtend) then
  begin
    CheckSels;
    Extended:=false;
    if CompareLargePoint(OldPos,SelectStart)=0 then
    begin
      SetSelection(CurPos,SelectEnd);
      Extended:=true;
    end;
    CheckSels;
    if not Extended then
      if CompareLargePoint(OldPos, SelectEnd)=0 then
      begin
        SetSelection(SelectStart,CurPos);
        Extended:=true;
      end;
    CheckSels;
    if not Extended then
      if CompareLargePoint(OldPos,CurPos)<=0 then
      begin
        SetSelection(OldPos,CurPos);
        Extended:=true;
      end
      else
      begin
        SetSelection(CurPos,OldPos);
        Extended:=true;
      end;
    DrawView;
  end
  else
  if (EditorFlags and efPersistentBlocks)=0 then HideSelect;
  if CompareLargePoint(SelectStart,SelectEnd)=0 then SetSelection(CurPos,CurPos);
{  if (Flags and (efHighlightColumn+efHighlightRow))<>0 then DrawView;}
{  if ((CurPos.X<>OldPos.X) or (CurPos.Y<>OldPos.Y)) and
     ((Highlight.A.X<>HighLight.B.X) or (Highlight.A.Y<>HighLight.B.Y)) then
     HideHighlight;}
{  if (OldPos.Y<>CurPos.Y) and (0<=OldPos.Y) and (OldPos.Y<GetLineCount) then
     SetLineText(OldPos.Y,TrimRight(GetLineText(OldPos.Y)));
  if ((CurPos.X<>OldPos.X) or (CurPos.Y<>OldPos.Y)) and (GetErrorMessage<>'') then
    SetErrorMessage('');
  if ((CurPos.X<>OldPos.X) or (CurPos.Y<>OldPos.Y)) and (HighlightRow<>-1) then
    SetHighlightRow(-1);
  if ((CurPos.X<>OldPos.X) or (CurPos.Y<>OldPos.Y)) then
    AddAction(eaMoveCursor,OldPos,CurPos,'');}
  UnLock;
end;


procedure TExtEditor.ChangeBounds(var Bounds: TRect);
begin
  SetBounds(Bounds);
  Lock;
  SetLimit(MaxLineLength, Limit.Y);
  DrawView;
  UnLock;
end;

function TExtEditor.GetPalette: PPalette;
const
  P: String[Length(CEditor)] = CEditor;
begin
  GetPalette := @P;
end;

procedure TExtEditor.ConvertEvent(var Event: TEvent);
var
  Key: Word;
begin
  if Event.What = evKeyDown then
  begin
    if (GetShiftState and kbShift <> 0) and
      (Event.ScanCode >= $47) and (Event.ScanCode <= $51) then
      Event.CharCode := #0;
    Key := Event.KeyCode;
    if KeyState <> 0 then
    begin
      if (Lo(Key) >= $01) and (Lo(Key) <= $1A) then Inc(Key, $40);
      if (Lo(Key) >= $61) and (Lo(Key) <= $7A) then Dec(Key, $20);
    end;
    Key := ScanKeyMap(KeyMap[KeyState], Key);
    if (KeyState<>0) and (Key=0) then
      ClearEvent(Event); { eat second key if unrecognized after ^Q or ^K }
    KeyState := 0;
    if Key <> 0 then
      if Hi(Key) = $FF then
        begin
          KeyState := Lo(Key);
          ClearEvent(Event);
        end
      else
        begin
          Event.What := evCommand;
          Event.Command := Key;
        end;
  end;
end;


procedure TExtEditor.CheckSels;
begin
  if (SelectStart.Y>SelectEnd.Y) or
     ((SelectStart.Y=SelectEnd.Y) and (SelectStart.X>SelectEnd.X)) then
     SetSelection(SelectEnd,SelectStart);
end;

procedure TExtEditor.SelectionChanged;
var Enable: boolean;
begin
  if LockCount=0 then
  begin
    Enable:=isSelected and (Clipboard<>nil);
    SetCmdState(ToClipCmds,Enable);
    SetCmdState(NulClipCmds,Enable);
    Message(Application,evBroadcast,cmCommandSetChanged,nil);
  end
  else LockFlag:=LockFlag or lfSelectionChanged
end;

procedure TExtEditor.SetSelection(A, B: TLargePoint);
begin
  SelectStart:=A;
  SelectEnd:=B;
  SelectionChanged;
end;

procedure TExtEditor.HandleEvent(var Event: TEvent);
procedure DestroyLinks;
begin
  if Event.InfoPtr=HScrollBar then HScrollBar:=nil;
  if Event.InfoPtr=VScrollBar then VScrollBar:=nil;
end;
var Mouse:TPoint;
    D:TLargePoint;
{    SelectMode: Byte;}

begin
  inherited HandleEvent(Event);
  ConvertEvent(Event);
  case Event.What of
    evMouseDown :
(*      if MouseInView(Event.Where) then
{       if (Event.Buttons=mbRightButton) then
         begin
           MakeLocal(Event.Where,P); Inc(P.X); Inc(P.Y);
           LocalMenu(P);
           ClearEvent(Event);
         end else}
       if Event.Buttons=mbLeftButton then
        begin
          GetMousePos(P);
          StartP:=P;
          SetCurPtr(P.X,P.Y);
          repeat
            GetMousePos(P);
{            if (PointOfs(P)<PointOfs(StartP)
               then SetSelection(P,StartP)
               else SetSelection(StartP,P);}
            SetCurPos(P.X,P.Y);
            DrawView;
          until not MouseEvent(Event, evMouseMove+evMouseAuto);
          DrawView;
        end;*)
      begin
{        if Event.Double then SelectMode := SelectMode or smDouble;}
        repeat
          Lock;
          MakeLocal(Event.Where, Mouse);
          if Event.What = evMouseAuto then
          begin
            MakeLocal(Event.Where, Mouse);
            D := Delta;
            if Mouse.X < 0 then Dec(D.X);
            if Mouse.X >= Size.X then Inc(D.X);
            if Mouse.Y < 0 then Dec(D.Y);
            if Mouse.Y >= Size.Y then Inc(D.Y);
            ScrollTo(D.X, D.Y);
          end;
          SetCurPos(Mouse.X+Delta.X,Mouse.Y+Delta.Y);
{          SelectMode := SelectMode or smExtend;}
          Unlock;
        until not MouseEvent(Event, evMouseMove + evMouseAuto);
      end;
    evBroadcast:
      begin
        if (Event.Command = cmScrollBarChanged) and
           ((Event.InfoPtr = HScrollBar) or (Event.InfoPtr = VScrollBar)) then
           ScrollDraw;
        if Event.Command=cmDestroed then DestroyLinks;
        exit;
      end;
    evKeyDown:if Event.CharCode in [#9,#32..#255] then
                 AddChar(Event.CharCode, false);
    evCommand:
     case Event.Command of
          cmCharLeft    : CharLeft;
          cmCharRight   : CharRight;
          cmWordLeft    : WordLeft;
          cmWordRight   : WordRight;
          cmLineStart   : LineStart;
          cmLineEnd     : LineEnd;
          cmLineUp      : LineUp;
          cmLineDown    : LineDown;
          cmPageUp      : PageUp;
          cmPageDown    : PageDown;
          cmTextStart   : TextStart;
          cmTextEnd     : TextEnd;
          cmNewLine     : NewLine;
{          cmBreakLine   : BreakLine;}
          cmBackSpace   : BackSpace;
          cmDelChar     : DelChar;
          cmDelWord     : DelWord;
          cmDelStart    : DelStart;
          cmDelEnd      : DelEnd;
          cmDelLine     : DelLine;
          cmInsMode     : ToggleInsMode;
          cmHideSelect  : HideSelect;
          cmStartSelect : StartSelect;
          cmEndSelect   : EndSelect;
          cmDelSelect   : DelSelect;
          cmCopyBlock   : CopyBlock;
          cmMoveBlock   : MoveBlock;
          cmIndentBlock   : IndentBlock;
          cmUnindentBlock : UnindentBlock;
{          cmSelectStart    : JumpSelectStart;
          cmSelectEnd      : JumpSelectEnd;
          cmLastCursorPos : JumpToLastCursorPos;}
          cmJumpMark0..cmJumpMark9 : JumpMark(Event.Command-cmJumpMark0);
          cmSetMark0..cmSetMark9 : DefineMark(Event.Command-cmSetMark0);
          cmPaste       : ClipPaste;
          cmCut         : ClipCut;
          cmCopy        : ClipCopy;
{          cmSelectWord  : SelectWord;
          cmSelectLine  : SelectLine;}
          cmWriteBlock  : WriteBlock;
          cmReadBlock   : ReadBlock;
{          cmPrintBlock  : PrintBlock;}
          cmFind        : Find;
          cmReplace     : Replace;
          cmSearchAgain : DoSearchReplace;
{          cmUndo        : Undo;
          cmRedo        : Redo;}
          cmClear       : DelSelect;
          cmGotoLine    : GotoLine;
     else
       exit;
     end;
  else
    exit;
  end;
  ClearEvent(Event);
end;

procedure TExtEditor.GotoLine;
var Line:Longint;
begin
  Line:=CurPos.Y+1;
  if EditorDialog(edGotoLine, @Line) <> cmCancel then
  begin
    SetCurPos(0,Line-1);
    TrackCursor(true);
  end;
end;

procedure TExtEditor.ScrollDraw;
var
  D: TLargePoint;
begin
  if HScrollBar <> nil then D.X := HScrollBar^.Value
  else D.X := 0;
  if VScrollBar <> nil then D.Y := VScrollBar^.Value
  else D.Y := 0;
  if (D.X <> Delta.X) or (D.Y <> Delta.Y) then
  begin
    SetCursor(Cursor.X + Delta.X - D.X, Cursor.Y + Delta.Y - D.Y);
    Delta := D;
    DrawView;
  end;
end;

procedure TExtEditor.ScrollTo(X, Y: Longint);
begin
  Lock;
  if HScrollBar <> nil then HScrollBar^.SetValue(X)
  else Delta.X:=X;
  if VScrollBar <> nil then VScrollBar^.SetValue(Y) else Delta.Y:=Y;
  Unlock;
end;

procedure TExtEditor.DrawCur;
begin
  SetCursor(CurPos.X-Delta.X,CurPos.Y-Delta.Y);
end;

procedure TExtEditor.SetLimit(X, Y: Longint);
begin
  Limit.X := X;
  Limit.Y := Y;
  Lock;
  if HScrollBar <> nil then
     HScrollBar^.SetParams(HScrollBar^.Value, 0, X - Size.X, Size.X - 1,
      HScrollBar^.ArStep);
  if VScrollBar <> nil then
    VScrollBar^.SetParams(VScrollBar^.Value, 0, Y - Size.Y, Size.Y - 1,
      VScrollBar^.ArStep);
  UnLock;
end;

procedure TExtEditor.SetState(AState: Word; Enable: Boolean);

procedure ShowSBar(SBar: PView);
begin
  if (SBar <> nil) then
    if GetState(sfActive + sfSelected) then SBar^.Show
    else SBar^.Hide;
end;

begin
  TView.SetState(AState, Enable);
  if AState and (sfActive + sfSelected) <> 0 then
  begin
    ShowSBar(HScrollBar);
    ShowSBar(VScrollBar);
    ShowSBar(Indicator);
    UpdateCommands;
  end;
end;

procedure TExtEditor.LineDown;
begin
  if FileObject^.CountLine>CurPos.Y then
     SetCurPos(CurPos.X, CurPos.Y+1);
end;

procedure TExtEditor.LineUp;
begin
  if CurPos.Y>0 then SetCurPos(CurPos.X, CurPos.Y-1);
end;

procedure TExtEditor.CharLeft;
begin
  if CurPos.X>0 then SetCurPos(CurPos.X-1, CurPos.Y);
end;

procedure TExtEditor.CharRight;
begin
  if CurPos.X<MaxLineLength then SetCurPos(CurPos.X+1, CurPos.Y);
end;

procedure TExtEditor.LineStart;
begin
  SetCurPos(0, CurPos.Y);
end;

procedure TExtEditor.LineEnd;
begin
  SetCurPos(Length(TrimRight(FileObject^.Str(CurPos.Y))), CurPos.Y);
end;

procedure TExtEditor.PageUp;
begin
  ScrollTo(Delta.X,Max(Delta.Y-Size.Y+1,0));
  SetCurPos(CurPos.X,Max(0,CurPos.Y-Size.Y+1));
end;

procedure TExtEditor.PageDown;
begin
  ScrollTo(Delta.X,Min(Delta.Y+Size.Y-1,FileObject^.CountLine-1));
  SetCurPos(CurPos.X,Min(FileObject^.CountLine-1,CurPos.Y+Size.Y-1));
end;

procedure TExtEditor.TextStart;
begin
  SetCurPos(0,0);
end;

procedure TExtEditor.TextEnd;
begin
  SetCurPos(Length(TrimRight(FileObject^.Str(FileObject^.CountLine-1))), FileObject^.CountLine);
end;

procedure TExtEditor.WordRight;
var S:string;
    i:integer;
begin
  S:=FileObject^.Str(CurPos.Y);
  if CurPos.X<Length(S)-1 then
  begin
    i:=CurPos.X+1;
    while (i<Length(S)) and (S[i] in WordChars) do Inc(i);
    while (i<Length(S)) and not (S[i] in WordChars) do Inc(i);
    SetCurPos(i-1, CurPos.Y);
  end
  else
  if CurPos.Y<FileObject^.CountLine-1 then
  begin
    inc(CurPos.Y);
    CurPos.X:=0;
    SetCurPos(CurPos.X, CurPos.Y);
  end;
end;

procedure TExtEditor.WordLeft;
var S:string;
    i:integer;
begin
  if CurPos.X>0 then
  begin
    i:=CurPos.X-1;
    S:=FileObject^.Str(CurPos.Y);
    while (i>1) and not (S[i] in WordChars) do Dec(i);
    while (i>2) and (S[i-1] in WordChars) do Dec(i);
    SetCurPos(i-1, CurPos.Y);
  end
  else
  if CurPos.Y>0 then
  begin
    Dec(CurPos.Y);
    CurPos.X:=Length(FileObject^.Str(CurPos.Y));
    SetCurPos(CurPos.X, CurPos.Y);
  end;
end;

procedure TExtEditor.ToggleInsMode;
begin
  Overwrite := not Overwrite;
  SetState(sfCursorIns, not GetState(sfCursorIns));
end;

procedure TExtEditor.UpdateIndicator;
begin
  if Indicator<>nil then Indicator^.SetValue(CurPos,Modified);
end;

function TExtEditor.Modified:boolean;
begin
  Modified:=FileObject^.Modified;
end;

function TExtEditor.InsertLine(S: string; SelectText : Boolean): Boolean;
var j, Count: integer;
    S1:string;

procedure InsertL(f:boolean);
var S2:string;
begin
  S2:=FileObject^.Str(CurPos.Y);
  S1:=Copy(S2,1,CurPos.X)+S1;
  S2:=Copy(S2,CurPos.X+1,Length(S2));
  if f then
  begin
    FileObject^.ReplaseStr(CurPos.Y,S1);
    SetCurPos(CurPos.X+Length(S1),CurPos.Y);
    inc(Count);
    NewLine;
  end
  else S2:=S1+S2;
  FileObject^.ReplaseStr(CurPos.Y,S2);
  SetCurPos(CurPos.X+Length(S2),CurPos.Y);
end;

begin
  if ReadOnly then exit;
  InsertLine:=true;
  Count:=1;
{  Lock;}
  while S<>'' do
  begin
    J:=Pos(sEol, S);
    if J=0 then
    begin
      S1:=S;
      S:='';
    end
    else
    begin
      S1:=Copy(S,1,J-1);
      Delete(S,1,J+Length(sEol)-1);
    end;
    InsertL(J<>0);
  end;
  LimitsChanged;
  DrawLines(CurPos.Y-Count,Count,true);
{  UnLock;}
end;

procedure TExtEditor.AddChar(const C:char; SelectText : Boolean);
var S:string;
    TabS:string[20];
    Temp:boolean;
begin
  if ReadOnly then exit;
  s:=FileObject^.Str(CurPos.Y);
  if (C=#9) and (EditorFlags and efUseTabCharacters=0) then
  begin
    TabS:=DupChar(' ',TabSize-((CurPos.X) mod TabSize));
    if Length(S)<CurPos.X+1 then S:=S+DupChar(' ',CurPos.X-Length(S))+TabS
    else
    begin
      if Overwrite then
         S:=S+Copy(S,1,CurPos.X)+TabS+Copy(S,CurPos.X+Length(TabS),255)
      else
      begin
        S:=Copy(S,1,CurPos.X)+TabS+Copy(S,CurPos.X+1,255);
        ChangeSelected(Length(TabS));
      end;
    end;
    Temp:=NoSelect;
    NoSelect:=not SelectText;
    SetCurPos(CurPos.X+Length(TabS),CurPos.Y);
    NoSelect:=Temp;
  end
  else
  begin
    if Length(S)<CurPos.X+1 then
      S:=S+DupChar(' ',CurPos.X-Length(S))+C
    else
    begin
      if Overwrite then S[CurPos.X+1]:=C
      else
      begin
        S:=Copy(S,1,CurPos.X)+C+Copy(S,CurPos.X+1,255);
        ChangeSelected(1);
      end;
    end;
    Temp:=NoSelect;;
    NoSelect:=not SelectText;
    SetCurPos(CurPos.X+1,CurPos.Y);
    NoSelect:=Temp;
  end;
  FileObject^.ReplaseStr(CurPos.Y,S);
  LimitsChanged;
  DrawLines(CurPos.Y,1,true);
end;

procedure TExtEditor.ChangeSelected(Count:integer);
begin
  if not IsSelected then exit;
  if (CurPos.Y=SelectStart.Y) and (CurPos.X<SelectStart.X) then
     Inc(SelectStart.X, Count);
  if (CurPos.Y=SelectEnd.Y) and (CurPos.X<SelectEnd.X) then
     Inc(SelectEnd.X, Count);
end;

procedure TExtEditor.DelChar;
var S: string;
    Count: Longint;
begin
  if ReadOnly then exit;
  S:=FileObject^.Str(CurPos.Y);
  Count:=1;
  if CurPos.X=length(S) then
  begin
     if CurPos.Y<FileObject^.CountLine-1 then
      begin
        S:=S+FileObject^.Str(CurPos.Y+1);
        FileObject^.DeleteStr(CurPos.Y+1);
        Count:=Size.Y-CurPos.Y;
      end;
  end
  else
  begin
    Delete(S,CurPos.X+1,1);
    ChangeSelected(-1);
  end;
  FileObject^.ReplaseStr(CurPos.Y,S);
  SetCurPos(CurPos.X,CurPos.Y);
{  UpdateAttrs(CurPos.Y,attrAll);
  AdjustSelection(SDX,SDY);}
  LimitsChanged;
  DrawLines(CurPos.Y,Count,true);
  UpdateIndicator;
end;

procedure TExtEditor.BackSpace;
var S,PreS: string;
    CP,Y:Longint;
begin
  if ReadOnly then exit;
  if CurPos.X=0 then
  begin
    if CurPos.Y>0 then
    begin
      S:=FileObject^.Str(CurPos.Y-1);
      FileObject^.ReplaseStr(CurPos.Y-1,S+FileObject^.Str(CurPos.Y));
      FileObject^.DeleteStr(CurPos.Y);
      LimitsChanged;
      SetCurPos(length(S),CurPos.Y-1);
      DrawLines(CurPos.Y,Size.Y+Delta.Y-CurPos.Y,true);
    end;
  end
  else
  begin
    S:=FileObject^.Str(CurPos.Y);
    CP:=CurPos.X-1;
    if (EditorFlags and efBackspaceUnindents)<>0 then
      if Trim(copy(S,1,CP+1))='' then
      begin
        Y:=CurPos.Y;
        while (Y>0) do
        begin
           Dec(Y);
           PreS:=FileObject^.Str(Y);
           if Trim(copy(PreS,1,CP+1))<>'' then Break;
        end;
        if Y<0 then PreS:='';
        while (CP>0) and ((CP>Length(S)) or (S[CP]=' ')) and
              ((CP>Length(PreS)) or (PreS[CP]<>' ')) do Dec(CP);
      end;
    Delete(S,CP+1,CurPos.X-CP);
    ChangeSelected(-(CurPos.X-CP));
    FileObject^.ReplaseStr(CurPos.Y,S);
    SetCurPos(CP,CurPos.Y);
    LimitsChanged;
    DrawLines(CurPos.Y,1,true);
  end;
end;

procedure TExtEditor.JumpMark(MarkIdx: integer);
begin
  if (MarkIdx<Low(Bookmarks)) or (MarkIdx>High(Bookmarks)) then exit;;
  with Bookmarks[MarkIdx] do
  if Valid then SetCurPos(Pos.X,Pos.Y);
end;

function TExtEditor.IsClipboard:boolean;
begin
  IsClipboard := Clipboard = @Self;
end;

procedure TExtEditor.DefineMark(MarkIdx: integer);
begin
  if (MarkIdx<Low(Bookmarks)) or (MarkIdx>High(Bookmarks)) then exit;
  with Bookmarks[MarkIdx] do
   begin
     Pos:=CurPos;
     Valid:=true;
   end;
end;

procedure TExtEditor.CopyBlock;
var P:PStrings;
begin
  if ReadOnly then exit;
  if (SelectStart.X=SelectEnd.X) and (SelectStart.Y=SelectEnd.Y) then Exit;
  P:=GetSelectedText;
  SetSelectedText(P);
  Dispose(P, Done);
end;

procedure TExtEditor.MoveBlock;
var P:PStrings;
    Temp:TLargePoint;
begin
  if ReadOnly then exit;
  if (SelectStart.X=SelectEnd.X) and (SelectStart.Y=SelectEnd.Y) then Exit;
  Lock;
  P:=GetSelectedText;
  Temp:=CurPos;
  if CurPos.Y>SelectStart.Y then Dec(Temp.Y,P^.Count-1);
  DelSelect;
  SetCurPos(Temp.X, Temp.Y);
  SetSelectedText(P);
  UnLock;
  Dispose(P, Done);
end;

procedure TExtEditor.IndentBlock;
var
  ey,i : integer;
  S : String;
begin
  if ReadOnly then Exit;
  if (SelectStart.X=SelectEnd.X) and (SelectStart.Y=SelectEnd.Y) then Exit;
  Lock;
  ey:=SelectEnd.y;
  if SelectEnd.x=0 then
   dec(ey);
  for i:=SelectStart.y to ey do
   begin
     S:=FileObject^.Str(i);
     FileObject^.ReplaseStr(i,' '+S);
   end;
  SetCurPos(CurPos.X,CurPos.Y);
  DrawLines(SelectStart.y,EY,true);
  UpdateIndicator;
  UnLock;
end;

procedure TExtEditor.UnindentBlock;
var
  ey,i : integer;
  S : String;
begin
  if ReadOnly then Exit;
  if (SelectStart.X=SelectEnd.X) and (SelectStart.Y=SelectEnd.Y) then Exit;
  Lock;
  ey:=SelectEnd.y;
  if SelectEnd.x=0 then dec(ey);
  for i:=SelectStart.y to ey do
   begin
     S:=FileObject^.Str(i);
     if (length(s)>1) and (S[1]=' ') then Delete(s,1,1);
     FileObject^.ReplaseStr(i,S);
   end;
  SetCurPos(CurPos.X,CurPos.Y);
  DrawLines(SelectStart.y, EY,true);
  UpdateIndicator;
  UnLock;
end;

procedure TExtEditor.DelSelect;
begin
  if ReadOnly then Exit;
  if (SelectStart.X=SelectEnd.X) and (SelectStart.Y=SelectEnd.Y) then Exit;
  FileObject^.DeleteText(SelectStart,SelectEnd);
  SetCurPos(SelectStart.X,SelectStart.Y);
  HideSelect;
  DrawView;
  UpdateIndicator;
end;

function TExtEditor.NewLine: integer;
var Ind: integer;
    S,IndentStr: string;
procedure CalcIndent(LineOver: integer);
begin
  if (LineOver<0) or (LineOver>FileObject^.CountLine) or
     (EditorFlags and efAutoIndent = 0) then
     Ind:=0
  else
  begin
    IndentStr:=FileObject^.Str(LineOver);
    Ind:=0;
    while (Ind<length(IndentStr)) and (IndentStr[Ind+1]=' ') do Inc(Ind);
  end;
  IndentStr:=DupChar(' ',Ind);
end;
var i: integer;
begin
  NewLine:=0;
  if ReadOnly then exit;
  if CurPos.Y<FileObject^.CountLine then S:=FileObject^.Str(CurPos.Y)
  else S:='';
  if not Overwrite then
  begin
    FileObject^.ReplaseStr(CurPos.Y, copy(S,1,CurPos.X-1+1));
    CalcIndent(CurPos.Y);
    FileObject^.InsertStr(CurPos.Y+1,IndentStr+copy(S,CurPos.X+1,255));
    LimitsChanged;
    SetCurPos(Ind,CurPos.Y+1);
    DrawLines(CurPos.Y-1,Size.Y+Delta.Y-CurPos.Y,true);
  end
  else
  begin
    if CurPos.Y=FileObject^.CountLine-1 then
    begin
      CalcIndent(CurPos.Y);
      FileObject^.InsertStr(CurPos.Y+1,IndentStr);
      LimitsChanged;
      SetCurPos(Ind,CurPos.Y+1);
    end
    else
    begin
      S:=FileObject^.Str(CurPos.Y+1);
      i:=0;
      while (I<Length(S)) and (S[I]=#32) do inc(i);
      SetCurPos(I,CurPos.Y+1);
    end;
  end;
  UpdateIndicator;
end;


procedure TExtEditor.WriteBlock;
var Block:PStrings;
    AFileName:FNameStr;
begin
  Block:=GetSelectedText;
  if Block<>nil then
  begin
    AFileName:='*'+DefaultSaveExt;
    if EditorDialog(edWriteBlock, @AFileName)<>cmCancel then
       Block^.SaveToFile(AFileName);
    Dispose(Block, Done);
  end;
end;

procedure TExtEditor.ReadBlock;
var I:Longint;
    TempS,S:string;
    AFileName:FNameStr;
    f:text;
begin
  if ReadOnly then exit;
  AFileName:='*'+DefaultSaveExt;
  if EditorDialog(edReadBlock,@AFileName)<>cmCancel then
  begin
    Assign(f,AFileName);
    Reset(f);
    SelectStart:=CurPos;
    i:=0;
    if (CurPos.X<>0) and not Eof(f) then
    begin
      Readln(f,s);
      TempS:=FileObject^.Str(CurPos.Y);
      S:=Copy(TempS,1,CurPos.X)+S;
      Delete(TempS, 1, CurPos.X);
      FileObject^.ReplaseStr(CurPos.Y,S);
      inc(i);
    end else TempS:='';
    while not Eof(f) do
    begin
      Readln(f,s);
      FileObject^.InsertStr(CurPos.Y+i,S);
      inc(i);
    end;
    Dec(i);
    if TempS<>'' then
    begin
      S:=FileObject^.Str(CurPos.Y+I);
      FileObject^.ReplaseStr(CurPos.Y+i,S+TempS);
    end;
    SelectEnd.Y:=CurPos.Y+I;
    SelectEnd.X:=Length(S);
    Close(F);
    LimitsChanged;
    DrawView;
  end;
end;

procedure TExtEditor.HideSelect;
begin
  if (SelectStart.X=SelectEnd.X) and (SelectStart.Y=SelectEnd.Y) then exit;
  SetSelection(CurPos,CurPos);
  DrawView;
end;


function TExtEditor.GetSelectedText:PStrings;
var i{,a,b}:integer;
    R:PStrings;
begin
  if (SelectStart.X=SelectEnd.X) and (SelectStart.Y=SelectEnd.Y) then R:=nil
  else
  begin
    R:=New(PStrings, Init);
    if SelectStart.Y=SelectEnd.Y then
      R^.Add(Copy(FileObject^.Str(SelectStart.Y),SelectStart.X+1,SelectEnd.X-SelectStart.X))
    else
    begin
      R^.Add(Copy(FileObject^.Str(SelectStart.Y),SelectStart.X+1,MaxLineLength));
{      a:=1;b:=MaxLineLength;}
      for i:=SelectStart.Y+1 to SelectEnd.Y-1 do
        R^.Add(Copy(FileObject^.Str(i), 1, MaxLineLength));
      R^.Add(Copy(FileObject^.Str(SelectEnd.Y),1,SelectEnd.X))
    end;
  end;
  GetSelectedText:=R;
end;

procedure TExtEditor.SetSelectedText(P:PStrings);
var S, TempS:string;
    i:integer;
begin
  if ReadOnly then exit;
  if (P<>nil) then
  begin
    if (P^.Count>1) then
    begin
      SelectStart:=CurPos;
      s:=FileObject^.Str(CurPos.Y);
      TempS:=Copy(S,CurPos.X+1,255);
      S:=Copy(S,1,CurPos.X)+P^.Str(0);
      FileObject^.ReplaseStr(CurPos.Y,S);
      for i:=1 to P^.Count-2 do
        FileObject^.InsertStr(CurPos.Y+i,P^.Str(i));
      TempS:=P^.Str(P^.Count-1)+TempS;
      FileObject^.InsertStr(CurPos.Y+P^.Count-1,TempS);
      SelectEnd.X:=Length(P^.Str(P^.Count-1));
      SelectEnd.Y:=CurPos.Y+P^.Count-1;
    end
    else
    begin
      S:=FileObject^.Str(CurPos.Y);
      Insert(P^.Str(0), S, CurPos.X+1);
      FileObject^.ReplaseStr(CurPos.Y, S);
    end;
    LimitsChanged;
    DrawView;
  end
end;

procedure TExtEditor.EndSelect;
var P: TLargePoint;
    LS: integer;
begin
  P:=CurPos;
  LS:=length(FileObject^.Str(SelectEnd.Y));
  if LS<P.X then P.X:=LS;
{  CheckSels;
  SetSelection(SelectStart,P);}
  if (CompareLargePoint(SelectStart,P)>0) or not IsSelected then
     SetSelection(P,P)
  else SetSelection(SelectStart,P);
  DrawView;
end;

procedure TExtEditor.StartSelect;
begin
  Lock;
  if (CompareLargePoint(CurPos,SelectEnd)=1) or not IsSelected then
     SetSelection(CurPos,CurPos)
  else SetSelection(CurPos,SelectEnd);
  DrawView;
  UnLock;
end;

function TExtEditor.isSelected:boolean;
begin
  isSelected:=not ((SelectStart.X=SelectEnd.X) and (SelectStart.Y=SelectEnd.Y));
end;

function TExtEditor.InsertFrom(Editor: PExtEditor): Boolean;
var S,TempS:string;
    i:integer;
    Count:LOngint;
begin
  InsertFrom:=true;
  if ReadOnly then exit;
  if (Editor<>nil) and (Editor<>@Self) and Editor^.isSelected then
  begin
    SelectStart:=CurPos;
    Count:=Editor^.SelectEnd.Y-Editor^.SelectStart.Y+1;
    if Count=1 then
    begin
      with Editor^ do
      S:=Copy(FileObject^.Str(SelectStart.Y), SelectStart.X+1,SelectEnd.X-SelectStart.X);
      TempS:=FileObject^.Str(CurPos.Y);
      Insert(S,TempS,CurPos.X+1);
      FileObject^.ReplaseStr(CurPos.Y,TempS);
      SelectEnd.X:=CurPos.X+Length(S);
      SelectEnd.Y:=CurPos.Y;
    end
    else
    begin
      S:=FileObject^.Str(CurPos.Y);
      TempS:=Copy(S,CurPos.X+1,255);
      with Editor^ do
        S:=Copy(S,1,CurPos.X)+Copy(FileObject^.Str(SelectStart.Y), SelectStart.X+1,255);
      FileObject^.ReplaseStr(CurPos.Y,S);
      for i:=1 to Count-2 do
        FileObject^.InsertStr(CurPos.Y+i,
                      Editor^.FileObject^.Str(Editor^.SelectStart.Y+i));
      with Editor^ do
        TempS:=Copy(FileObject^.Str(SelectStart.Y+Count-1), 1, SelectEnd.X+1)+TempS;
      FileObject^.InsertStr(CurPos.Y+Count-1,TempS);
      SelectEnd.X:=Length(TempS)-1;
      SelectEnd.Y:=CurPos.Y+Count-1;
    end;
    LimitsChanged;
    if isClipboard then
    begin
      FileObject^.InsertStr(CurPos.Y+Count,'');
      SetCurPos(0,CurPos.Y+Count);
    end;
    DrawView;
    UpdateIndicator;
  end;
end;

procedure TExtEditor.ClipPaste;
begin
  if ReadOnly then exit;
  InsertFrom(Clipboard);
end;

procedure TExtEditor.ClipCut;
begin
  if ReadOnly then exit;
  if Clipboard<>nil then
  begin
     Clipboard^.InsertFrom(@Self);
     DelSelect;
  end;
end;

procedure TExtEditor.ClipCopy;
begin
  if Clipboard<>nil then Clipboard^.InsertFrom(@Self);
end;

procedure TExtEditor.DelWord;
var S:string;
    i:integer;
begin
  if ReadOnly then exit;
  s:=FileObject^.Str(CurPos.Y);
  if CurPos.Y>=Length(S)-1 then DelChar
  else
  begin
    I:=CurPos.X+1;
    while (i<Length(S)) and not (S[i] in WordChars) do Inc(i);
    while (i<Length(S)) and (S[i] in WordChars) do Inc(i);
    Delete(S,CurPos.X+1,I-CurPos.X-1);
    if isSelected then
    begin
      if (SelectStart.Y=CurPos.Y) and (SelectStart.X>CurPos.X) then
        SelectStart.X:=SelectStart.X-Min(I-CurPos.X-1, SelectStart.X-CurPos.X);
      if (SelectEnd.Y=CurPos.Y) and (SelectEnd.X>CurPos.X) then
        SelectEnd.X:=SelectEnd.X-Min(I-CurPos.X-1, SelectEnd.X-CurPos.X);
    end;
    FileObject^.ReplaseStr(CurPos.Y,S);
    DrawLines(CurPos.Y,1,true);
  end;
end;

function TExtEditor.GetCurWord:string;
var S:string;
    i,j:integer;
begin
  s:=FileObject^.Str(CurPos.Y);
  i:=CurPos.X+1;
  j:=i;
  while (i<Length(S)) and (S[i] in WordChars) do Inc(i);
  while (j>1) and (S[j] in WordChars) do Dec(j);
  GetCurWord:=trim(Copy(S,j,i-j));
end;

procedure TExtEditor.DelStart;
var S:String;
begin
  if ReadOnly then exit;
  s:=FileObject^.Str(CurPos.Y);
  Delete(S,1, CurPos.X+1);
{    if isSelected then
    begin
      if (SelectStart.Y=CurPos.Y) and (SelectStart.X>CurPos.X) then
        SelectStart.X:=SelectStart.X-Min(I-CurPos.X-1, SelectStart.X-CurPos.X);
      if (SelectEnd.Y=CurPos.Y) and (SelectEnd.X>CurPos.X) then
        SelectEnd.X:=SelectEnd.X-Min(I-CurPos.X-1, SelectEnd.X-CurPos.X);
    end;}
  FileObject^.ReplaseStr(CurPos.Y,S);
  SetCurPos(0,CurPos.Y);
  DrawLines(CurPos.Y,1,true);
end;

procedure TExtEditor.DelEnd;
var S:String;
begin
  if ReadOnly then exit;
  s:=FileObject^.Str(CurPos.Y);
  Delete(S,CurPos.X+1,255);
  FileObject^.ReplaseStr(CurPos.Y,S);
  DrawLines(CurPos.Y,1,true);
end;

procedure TExtEditor.DelLine;
begin
  if ReadOnly then exit;
  FileObject^.DeleteStr(CurPos.Y);
  if isSelected then
  begin
    if (SelectStart.Y>CurPos.Y) then Dec(SelectStart.Y);
    if SelectStart.Y=CurPos.Y then SelectStart.X:=0;
    if SelectEnd.Y=CurPos.Y then SelectEnd.X:=0;
    if (SelectEnd.Y>CurPos.Y) then Dec(SelectEnd.Y)
  end;
  SetCurPos(0,CurPos.Y);
  DrawView;
end;

procedure TExtEditor.Find;
var Rec:TStdEditorFindRec;
begin
  Rec.Options:=FindFlags and (ffCaseSensitive+
                              ffWholeWordsOnly);
  Rec.Scope:=Ord(FindFlags and ffSelectedText<>0);
  Rec.Direction:=Ord(FindFlags and ffBackward<>0);
  Rec.Origin:=Ord(FindFlags and ffEntireScope<>0);
  if EditorFlags and efFindTextAtCursor <> 0 then
     Rec.TextToFind:=GetCurWord
  else Rec.TextToFind:='';
  if Rec.TextToFind='' then Rec.TextToFind:=FindStr;
  if EditorDialog(edFind,@Rec)<>cmCancel then
  begin
    FindStr:=Rec.TextToFind;
    FindFlags := Rec.Options +
                 Rec.Scope * ffSelectedText +
                 Rec.Direction * ffBackward +
                 Rec.Origin * ffEntireScope;
    DoSearchReplace;
  end;
end;

procedure TExtEditor.Replace;
var Rec:TStdEditorReplaseRec;
    R:word;
begin
  Rec.Options:=FindFlags and (ffCaseSensitive+
                              ffWholeWordsOnly+
                              ffPromptOnReplace);
  Rec.Scope:=Ord(FindFlags and ffSelectedText<>0);
  Rec.Direction:=Ord(FindFlags and ffBackward<>0);
  Rec.Origin:=Ord(FindFlags and ffEntireScope<>0);
  if EditorFlags and efFindTextAtCursor <> 0 then
     Rec.TextToFind:=GetCurWord
  else Rec.TextToFind:='';
  Rec.NewText:=ReplaceStr;
  if Rec.TextToFind='' then Rec.TextToFind:=FindStr;
  R:=EditorDialog(edReplace,@Rec);
  if R<>cmCancel then
  begin
    FindStr:=Rec.TextToFind;
    ReplaceStr:=Rec.NewText;
    FindFlags := Rec.Options +
                 Rec.Scope * ffSelectedText +
                 Rec.Direction * ffBackward +
                 Rec.Origin * ffEntireScope +
                 ffReplace + Ord(R=cmYes)*ffReplaceAll;
    DoSearchReplace;
  end;
end;

procedure TExtEditor.DoSearchReplace;
var {R: Word;}
    C: TPoint;
    S:string;
    I:word;
begin
  repeat
    I := cmCancel;
    if not Search(FindStr) then
    begin
      if FindFlags and (ffReplaceAll + ffReplace) <>
          (ffReplaceAll + ffReplace) then
        EditorDialog(edSearchFailed, nil)
    end
    else
    if FindFlags and ffReplace <> 0 then
    begin
      I := cmYes;
      if FindFlags and ffPromptOnReplace <> 0 then
      begin
        MakeGlobal(Cursor, C);
        I := EditorDialog(edReplacePrompt, @C);
      end;
      if I = cmYes then
      begin
        S:=FileObject^.Str(CurPos.Y);
        Delete(S,SelectStart.X+1,SelectEnd.X-SelectStart.X);
        Insert(ReplaceStr,S,CurPos.X+1);
        FileObject^.ReplaseStr(CurPos.Y,S);
        DrawLines(CurPos.Y,1,false);
      end;
    end;
  until (I = cmCancel) or (FindFlags and ffReplaceAll = 0);
end;

function TExtEditor.Search(var FindStr:string):boolean;
var TempPos:TLargePoint;
begin
  TempPos:=CurPos;
  if FindTextSelected then Inc(CurPos.X);
  if not FindTextSelected then
  begin
    TempSelectStart:=SelectStart;
    TempSelectEnd:=SelectEnd;
  end;
  if (FindStr<>'') and FileObject^.FindString(FindStr, TempSelectStart, TempSelectEnd, CurPos) then
  begin
    SetCurPos(CurPos.X,CurPos.Y);
    SelectStart:=CurPos;
    SelectEnd:=SelectStart;
    SelectEnd.X:=SelectEnd.X+Length(FindStr);
    FindTextSelected:=true;
    DrawView;
    Search:=true;
  end
  else
  begin
{    EditorDialog(edSearchFailed,nil);}
    CurPos:=TempPos;
    Search:=false;
  end;
end;

procedure TExtEditor.UpdateCommands;
begin
{  SetCmdState(cmUndo, (DelCount <> 0) or (InsCount <> 0));}
  if not IsClipboard then
  begin
    SetCmdState([cmCopy,cmCut], isSelected);
    SetCmdState([cmPaste], (Clipboard <> nil) and (Clipboard^.isSelected));
  end;
  SetCmdState([cmClear], isSelected);
  SetCmdState([cmFind,cmReplace,cmSearchAgain,cmGotoLine], State and sfActive<>0);
end;

procedure RegisterExtEditor;
begin
{$IFNDEF DELPHI32}
  RegisterType(RExtEditor);
  RegisterType(RExtFileEditor);
  RegisterType(RExtIndicator);
  RegisterType(RExtEditWindow);
{$ENDIF}
end;

end.


