{*******************************************************}
{                                                       }
{   Turbo Pascal 7.0                                    }
{   FreePacal                                           }
{   Delphi 4                                            }
{   Extended visual components                          }
{                                                       }
{   Copyright (c) 1995-99 by Alexs                      }
{                                                       }
{*******************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}
{$I Define.inc}

unit Gauges;

interface

uses Drivers, Objects, Views, Register, Ext_Type;

type
  PPercentGauge = ^TPercentGauge;
  TPercentGauge = object(TView)
    MaxValue: Longint;
    CurValue: Longint;
    constructor Init(var Bounds: TRect; AMaxValue: Longint);
    procedure Draw; virtual;
    procedure Update(Progress: Longint); virtual;
    procedure AddProgress(Progress: Longint);
    procedure HandleEvent(var Event: TEvent); virtual;
    function SolveForX(Y, Z: Longint): Integer;
    function SolveForY(X, Z: Longint): Integer;
    function DataSize:Word; virtual;
    procedure SetData(var Rec);virtual;
  end;

  PBarGauge = ^TBarGauge;
  TBarGauge = object(TPercentGauge)
    constructor Init(var Bounds: TRect; AMaxValue: Longint);
    procedure Draw; virtual;
  end;

const
  RPercentGauge: TStreamRec = (
    ObjType: rcPercentGauge;
    VmtLink: Ofs(TypeOf(TPercentGauge)^);
    Load: @TPercentGauge.Load;
    Store: @TPercentGauge.Store
  );

const
  RBarGauge: TStreamRec = (
    ObjType: rcBarGauge;
    VmtLink: Ofs(TypeOf(TBarGauge)^);
    Load: @TBarGauge.Load;
    Store: @TBarGauge.Store
  );

procedure  RegisterGauges;

implementation

constructor TPercentGauge.Init(var Bounds: TRect; AMaxValue: Longint);
begin
  inherited Init(Bounds);
  EventMask := EventMask or evBroadcast;
  MaxValue := AMaxValue;
  CurValue := 0;
  ComponentType:=ctPercentGauge;
end;

procedure TPercentGauge.Draw;
var
  B: TDrawBuffer;
  C: Word;
  S: string[10];
  PercentDone: Longint;
begin
  C := GetColor(1);
  MoveChar(B, ' ', C, Size.X);
  PercentDone := SolveForY(CurValue, MaxValue);
  FormatStr(S, '%-3d%%', PercentDone);
  MoveStr(B, S, C);
  WriteLine(0,0,Size.X,Size.Y,B);
end;

procedure TPercentGauge.Update(Progress: Longint);
begin
  CurValue := Progress;
  DrawView;
end;

procedure TPercentGauge.AddProgress(Progress: Longint);
begin
  Update(Progress + CurValue);
end;

function TPercentGauge.DataSize:Word;
begin
  DataSize:=SizeOf(Longint)*2;
end;

procedure TPercentGauge.SetData(var Rec);
var
  R:TPersentRec absolute Rec;
begin
  MaxValue:=R.MaxValue;
  CurValue:=R.CurValue;
  DrawView;
end;

procedure TPercentGauge.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if Event.What = evBroadcast then
  begin
    case Event.Command of
      cmUpdateGauge :
        begin
          Update(Event.InfoLong);
        end;
      cmResetGauge:
        begin
          MaxValue := Event.InfoLong;
          Update(0);
        end;
      cmAddGauge:
        begin
          AddProgress(Event.InfoLong);
        end;
    end;
  end;
end;

{ This function solves for x in the equation "x is y% of z". }
function TPercentGauge.SolveForX(Y, Z: Longint): Integer;
begin
  SolveForX := Trunc( Z * (Y * 0.01) );
end;

{ This function solves for y in the equation "x is y% of z". }
function TPercentGauge.SolveForY(X, Z: Longint): Integer;
begin
  if Z = 0 then SolveForY := 0
  else SolveForY := Trunc( (X * 100) / Z );
end;


{ TBarGauge }
constructor TBarGauge.Init(var Bounds: TRect; AMaxValue: Longint);
begin
  inherited Init( Bounds, AMaxValue);
  ComponentType:=ctBarGauge;
end;

procedure TBarGauge.Draw;
var
  B: TDrawBuffer;
  C: Word;
  PercentDone: Longint;
  FillSize: Integer;
begin
  C := GetColor(1);
  MoveChar(B, #176, C, Size.X);
  PercentDone := SolveForY(CurValue, MaxValue);
  FillSize := SolveForX(PercentDone, Size.X);
  if FillSize > Size.X then FillSize := Size.X;
  MoveChar(B, #219, C, FillSize);
  WriteLine(0,0,Size.X,Size.Y,B);
end;

procedure RegisterGauges;
begin
  RegisterType( RPercentGauge);
  RegisterType( RBarGauge);
end;

end.
