{************************************************}
{                                                }
{   Tabbed notebook                              }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}

{$I Define.inc}

unit TabNotbk;

interface

uses Objects, Drivers, Views, Register;

type
  PTabPage = ^TTabPage;
  TTabPage = object(TCollection)
    Title:PString;
    CurSelect:PView;
    constructor Init(ATitle:string; ASize:integer);
    destructor Done;virtual;
  end;

type
  PTabbedNotebook = ^TTabbedNotebook;
  TTabbedNotebook = object(TView)
    CurPage:PTabPage;
    CurPageNum:integer;
    PageCollection:PCollection;
    constructor Init(var Bounds: TRect);
    destructor Done;virtual;
    constructor Load(var S:TStream);
    procedure Store(var S:TStream);
    procedure Draw;virtual;
    procedure HandleEvent(var Event:TEvent);virtual;
    procedure AddToPage(PageName:string; View:PView);
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    function GetPalette: PPalette; virtual;
    procedure SelectPage(PageNum:integer);
    function PageByName(const AName:string):PTabPage;
  end;

const
  RTabbedNotebook: TStreamRec = (
     ObjType: rcTabbedNotebook;
     VmtLink: Ofs(TypeOf(TTabbedNotebook)^);
     Load:    @TTabbedNotebook.Load;
     Store:   @TTabbedNotebook.Store
  );

procedure RegisterTabNotbk;
implementation
uses Ext_Str, Ext_Math, Dialogs, Ext_App;

  { TTabPage }
constructor TTabPage.Init(ATitle:string; ASize:integer);
begin
  inherited Init(ASize, 10);
  Title:=NewStr(ATitle);
end;

destructor TTabPage.Done;
begin
  DisposeStr(Title);
  DeleteAll;
  inherited Done;
end;


 { TTabbedNotebook }
constructor TTabbedNotebook.Init(var Bounds: TRect);
begin
  inherited Init(Bounds);
  GrowMode:=GrowMode or gfGrowHiX;
  Options := Options or (ofPreProcess + ofPostProcess);
  EventMask := EventMask or evBroadcast or evKeyDown;
  PageCollection:=New(PCollection, Init(10,10));
  ComponentType:=ctTabbedNotebook;
end;

destructor TTabbedNotebook.Done;
begin
  Dispose(PageCollection, Done);
  PageCollection:=nil;
  inherited Done;
end;

constructor TTabbedNotebook.Load(var S:TStream);
var C,i,l,j:integer;
    P:PTabPage;
begin
  inherited Load(S);
  PageCollection:=New(PCollection, Init(10,10));
  S.Read(C,SizeOf(Integer));
  for i:=1 to C do
  begin
    S.Read(L,SizeOf(Integer));
    P:=New(PTabPage, Init('', L));
    PageCollection^.Insert(P);
    P^.Title:=S.ReadStr;
    GetPeerViewPtr(S, P^.CurSelect);
    P^.Count:=L;
    for j:=0 to l-1 do GetPeerViewPtr(S, P^.Items^[j]);
  end;
  S.Read(CurPageNum, SizeOf(integer));
  if (CurPageNum>-1) and (CurPageNum<PageCollection^.Count) then
     CurPage:=PageCollection^.At(CurPageNum);
end;

procedure TTabbedNotebook.Store(var S:TStream);
var C,i,l,j:integer;
    P:PTabPage;
begin
  inherited Store(S);
  S.Write( PageCollection^.Count, SizeOf(Integer));
  for i:=0 to PageCollection^.Count-1 do
  begin
    P:=PageCollection^.At(i);
    S.Write(P^.Count,SizeOf(Integer));
    S.WriteStr(P^.Title);
    PutPeerViewPtr(S, P^.CurSelect);
    for j:=0 to P^.Count-1 do PutPeerViewPtr(S, P^.Items^[j]);
  end;
  S.Write(CurPageNum, SizeOf(integer));
end;

function TTabbedNotebook.GetPalette: PPalette;
const
  P: String[Length(CLabel)] = CLabel;
begin
  GetPalette := @P;
end;

procedure TTabbedNotebook.Draw;
var
  GrayColor,Color: Word;
  l,i,j,len,is1,ie: Integer;
  B: TDrawBuffer;
  PageSize:integer;
  PP:PTabPage;
  S,S1:string;
begin
  Color := GetColor($0402);
  GrayColor := GetColor($0301);
  if PageCollection^.Count<>0 then
     PageSize:=Size.X div PageCollection^.Count
  else
  begin
    WriteStr(0, 0, DupChar(#196,Size.X), Byte(Color));
    MoveChar(B[0], ' ', Byte(Color), Size.X);
    WriteLine(0, 1, Size.X, Size.Y-1, B);
    WriteStr(0, Size.Y-1, DupChar(#196,Size.X), Byte(Color));
    exit;
  end;
  S:=#218+DupChar(#196,Size.X-2)+#191;
  if CurPageNum=0 then S[1]:=#201;
  if CurPageNum=PageCollection^.Count-1 then
  begin
    for j:=(CurPageNum)*PageSize+2 to Length(S)-1 do S[j]:=#205;
    S[Length(S)]:=#187;
  end;
  for i:=1 to PageCollection^.Count-1 do
  begin
    if i-1 = CurPageNum then
    begin
      for j:=(i-1)*PageSize+2 to i*PageSize do S[j]:=#205;
      S[i*PageSize+1]:=#187
    end
    else if i = CurPageNum then S[i*PageSize+1]:=#201
    else S[i*PageSize+1]:=#218;{#194;}
  end;
  MoveStr(B, S, byte(Color));
  WriteLine(0, 0, Size.X, 1, B);
  MoveChar(B[0], ' ', Byte(Color), Size.X);
  for i:=0 to PageCollection^.Count-1 do
  begin
    PP:=PageCollection^.At(i);
    Len:=PageSize;
    Len:=(Len-Length(PP^.Title^)) div 2 +Length(PP^.Title^);
    FormatStr(S1,'%'+IntToStr(Len)+'s',PP^.Title);
    if (i=CurPageNum) or (i = CurPageNum+1) then
    begin
       MoveCStr(B[i*PageSize+1], S1, Color);
       S:=#186;
    end
    else
    begin
      MoveCStr(B[i*PageSize+1], S1, GrayColor);
      S:=#179;
    end;
    MoveCStr(B[i*PageSize], S, Color);
  end;
  if (CurPageNum=PageCollection^.Count-1) then  MoveCStr(B[Size.X-1], #186, Color)
  else MoveCStr(B[Size.X-1], #179, Color);
  WriteLine(0, 1, Size.X, Size.Y-1, B);
  S:=#207+DupChar(#205,Size.X-2)+#207;
  if CurPageNum=0 then
  begin
    S[1]:=#188;
    is1:=1;
  end;
  if CurPageNum=PageCollection^.Count-1 then
  begin
    S[Length(S)]:=#200;
    ie:=Length(S);
  end;
  for i:=1 to PageCollection^.Count-1 do
  begin
    if i-1 = CurPageNum then
    begin
      S[i*PageSize+1]:=#200;
      ie:=i*PageSize+1;
    end
    else if i = CurPageNum then
    begin
      S[i*PageSize+1]:=#188;
      is1:=i*PageSize+1;
    end
    else S[i*PageSize+1]:=#207;
  end;
  for i:=is1+1 to ie-1 do S[i]:=' ';
  MoveStr(B, S, Color);
  WriteLine(0, Size.Y-1, Size.X, Size.Y, B);
end;

procedure TTabbedNotebook.SelectPage(PageNum:integer);
procedure DoHide(P:PView);far;
begin
  if P<>nil then P^.Hide;
end;
procedure DoShow(P:PView);far;
begin
  if P<>nil then P^.Show;
end;
begin
  if PageCollection^.Count=0 then exit;
  if (PageNum<0) or (PageNum>PageCollection^.Count-1) then exit;
  if (CurPageNum=PageNum) then exit;
  Owner^.Lock;
  if CurPage<>nil then
  begin
    CurPage^.CurSelect:=Owner^.Current;
    CurPage^.ForEach(@DoHide);
  end;
  CurPage:=PageCollection^.At(PageNum);
  CurPage^.ForEach(@DoShow);
  if CurPage^.CurSelect<>nil then CurPage^.CurSelect^.Select;
  CurPageNum:=PageNum;
  PWindow(Owner)^.Frame^.DrawView;
  DrawView;
  Owner^.UnLock;
end;

procedure TTabbedNotebook.HandleEvent(var Event:TEvent);
var
  Mouse: TPoint;
  PageSize,i:integer;
  P:PTabPage;
  C:Char;
begin
  inherited HandleEvent(Event);
  case Event.What of
   evMouseDown:begin
                 if PageCollection^.Count<>0 then
                    PageSize:=Size.X div PageCollection^.Count
                 else PageSize:=Size.X;
                 MakeLocal(Event.Where, Mouse);
                 SelectPage(Mouse.X div PageSize);
               end;
   evKeyDown  :begin
                 if (Event.KeyCode=kbGrayPlus) or
                    (Event.KeyCode=kbCtrlTab) then
                 begin
                    if CurPageNum<PageCollection^.Count-1 then
                      SelectPage(CurPageNum+1)
                    else SelectPage(0);
                    ClearEvent(Event);
                 end
                 else
                 if Event.KeyCode=kbGrayMinus then
                 begin
                    if CurPageNum>0 then
                      SelectPage(CurPageNum-1)
                    else SelectPage(PageCollection^.Count-1);
                    ClearEvent(Event);
                 end
                 else
                 for i:=0 to PageCollection^.Count-1 do
                 begin
                    P:=PageCollection^.At(i);
                    C:=HotKey(P^.Title^);
                    if ((GetAltCode(C) = Event.KeyCode) and (P<>CurPage)) or
                       ((C <> #0) and (Owner^.Phase = phPostProcess) and
                       (UpCase(Event.CharCode) = C)) then
                    begin
                      SelectPage(i);
                      break;
                      ClearEvent(Event);
                    end
                 end;
                 exit;
               end;
  else
    exit;
  end;
  ClearEvent(Event);
end;

procedure TTabbedNotebook.AddToPage(PageName:string; View:PView);
var P,PP:PTabPage;
function DoFind(Item:PTabPage):boolean;far;
begin
  DoFind:=Item^.Title^=PageName;
end;
function DoCheck(Item:PTabPage):boolean;far;
begin
  DoCheck:=Item^.IndexOf(View)<>-1;
end;
begin
  P:=PageCollection^.FirstThat(@DoFind);
  if P=nil then
  begin
    P:=New(PTabPage, Init(PageName, 10));
    PageCollection^.Insert(P);
  end;
  PP:=PageCollection^.FirstThat(@DoCheck);
  if PP<>nil then PP^.Delete(View);
  P^.Insert(View);
  CurPage:=P;
  CurPageNum:=PageCollection^.IndexOf(P);
end;

function TTabbedNotebook.PageByName(const AName:string):PTabPage;
function DoFind(Item:PTabPage):boolean;far;
begin
  DoFind:=stUpCase(Item^.Title^)=stUpCase(AName);
end;
var P:PTabPage;
begin
  P:=PageCollection^.FirstThat(@DoFind);
  if P=nil then Abstract;
  PageByName:=P;
end;


procedure TTabbedNotebook.SetState(AState: Word; Enable: Boolean);
procedure DoPageHide(PP:PTabPage);far;
procedure DoHide(P:PView);far;
begin
  if P<>nil then P^.Hide;
end;
begin
  if PP<>CurPage then PP^.ForEach(@DoHide);
end;

procedure DoShow(P:PView);far;
begin
  if P<>nil then P^.Show;
end;

begin
  if (AState and (sfVisible+sfExposed)<>0) and Enable then
  begin
    PageCollection^.ForEach(@DoPageHide);
    if CurPage<>nil then CurPage^.ForEach(@DoShow);
  end;
  inherited SetState(AState, Enable)
end;

procedure RegisterTabNotbk;
begin
  RegisterType(RTabbedNotebook);
end;
end.
