{************************************************}
{                                                }
{   Turbo Vision Unit                            }
{   Copyright (c) 1992 Borland International     }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}

{$I Define.inc}

unit Validate;

interface

uses Register, Objects, Ext_Type;

const
{ TValidator Status constants }
  vsOk     =  0;
  vsSyntax =  1;      { Error in the syntax of either a TPXPictureValidator
                        or a TDBPictureValidator }

  { Validator option flags }
  voFill     =  $0001;
  voTransfer =  $0002;
  voOnAppend =  $0004;
  voReserved =  $00F8;

type
  TVTransfer = (vtDataSize, vtSetData, vtGetData);

{ Abstract TValidator object }

  PValidator = ^TValidator;
  TValidator = object(TObject)
    Status: Word;
    Options: Word;
    constructor Init;
    constructor Load(var S: TStream);
    procedure Error; virtual;
    function IsValidInput(var S: string;
      SuppressFill: Boolean): Boolean; virtual;
    function IsValid(const S: string): Boolean; virtual;
    procedure Store(var S: TStream);
    function Transfer(var S: String; Buffer: Pointer;
      Flag: TVTransfer): Word; virtual;
    function Valid(const S: string): Boolean;
  end;

{ TPXPictureValidator result type }

  TPicResult = (prComplete, prIncomplete, prEmpty, prError, prSyntax,
    prAmbiguous, prIncompNoFill);

{ TPXPictureValidator }

  PPXPictureValidator = ^TPXPictureValidator;
  TPXPictureValidator = object(TValidator)
    _Pic: PString;
    constructor Init(const APic: string; AutoFill: Boolean);
    constructor Load(var S: TStream);
    destructor Done; virtual;
    procedure Error; virtual;
    function IsValidInput(var S: string;
      SuppressFill: Boolean): Boolean; virtual;
    function IsValid(const S: string): Boolean; virtual;
    function Picture(var Input: string;
      AutoFill: Boolean): TPicResult; virtual;
    procedure Store(var S: TStream);
    function GetPic:String; virtual;
  end;

{ TFilterValidator }

  PFilterValidator = ^TFilterValidator;
  TFilterValidator = object(TValidator)
    ValidChars: TCharSet;
    constructor Init(AValidChars: TCharSet);
    constructor Load(var S: TStream);
    procedure Error; virtual;
    function IsValid(const S: string): Boolean; virtual;
    function IsValidInput(var S: string;
      SuppressFill: Boolean): Boolean; virtual;
    procedure Store(var S: TStream);
  end;

{ TRangeValidator }

  PRangeValidator = ^TRangeValidator;
  TRangeValidator = object(TFilterValidator)
    Min, Max: LongInt;
    constructor Init(AMin, AMax: LongInt);
    constructor Load(var S: TStream);
    procedure Error; virtual;
    function IsValid(const S: string): Boolean; virtual;
    procedure Store(var S: TStream);
    function Transfer(var S: String; Buffer: Pointer;
      Flag: TVTransfer): Word; virtual;
  end;

{ TLookupValidator }

  PLookupValidator = ^TLookupValidator;
  TLookupValidator = object(TValidator)
    function IsValid(const S: string): Boolean; virtual;
    function Lookup(const S: string): Boolean; virtual;
  end;

{ TStringLookupValidator }

  PStringLookupValidator = ^TStringLookupValidator;
  TStringLookupValidator = object(TLookupValidator)
    Strings: PStringCollection;
    constructor Init(AStrings: PStringCollection);
    constructor Load(var S: TStream);
    destructor Done; virtual;
    procedure Error; virtual;
    function Lookup(const S: string): Boolean; virtual;
    procedure NewStringList(AStrings: PStringCollection);
    procedure Store(var S: TStream);
  end;

  {#Z+}
  PByteValidator = ^TByteValidator;
  {#Z-}
  TByteValidator = Object(TRangeValidator)
    { A TByteValidator provides the same functionality of a TRangeValidator,
      but the data is transfered as a byte rather than a LongInt.  This saves
      3 bytes of space in the record used to set/get data from a dialog. }
    constructor Init (AMin, AMax : Byte);
    function Transfer (var S : String; Buffer : Pointer;
                       Flag : TVTransfer) : Word; virtual;
  end;  { of TByteValidator }


  PIntegerValidator = ^TIntegerValidator;
  TIntegerValidator = Object(TRangeValidator)
    constructor Init (AMin, AMax : Integer);
    function Transfer (var S : String; Buffer : Pointer;
                       Flag : TVTransfer) : Word; virtual;
  end;


  PRealValidator = ^TRealValidator;
  TRealValidator = object(TPXPictureValidator)
    Min : Real;
    Max : Real;
    constructor Init (APic: string; AutoFill: Boolean;
                      AMin, AMax : Real);
    constructor Load (var S: TStream);
    procedure Error; virtual;
    function IsValid (const S: String): Boolean; virtual;
    function IsValidInput (var S: string; SuppressFill: Boolean): Boolean;
                           virtual;
    procedure Store (var S: TStream);
    function Transfer (var S: String; Buffer: Pointer;
                       Flag: TVTransfer): Word; virtual;
  end;


  PSingleValidator = ^TSingleValidator;
  TSingleValidator = Object(TRealValidator)
    constructor Init (APic: string; AutoFill: Boolean; AMin, AMax : Single);
    function Transfer (var S : String; Buffer : Pointer;
                       Flag : TVTransfer) : Word; virtual;
  end;


  PWordValidator = ^TWordValidator;
  TWordValidator = Object(TRangeValidator)
    constructor Init (AMin, AMax : Word);
    function Transfer (var S : String; Buffer : Pointer;
                       Flag : TVTransfer) : Word; virtual;
  end;  { of TWordValidator }


  PDateRec = ^TDateRec;
  TDateRec = record
    Month: Byte;
    Day: Byte;
    Year: Word;
  end;  { of TDateRec }

  TDateFormat = (dfMonth, dfDay, dfYear, dfMMDDYY, dfYYMMDD, df4DigitYear,
                 dfFill);
  TDateFormats = set of TDateFormat;

const
  WesternDate = [dfMonth,dfDay,dfYear,dfMMDDYY,dfFill];
  EuropeanDate = [dfMonth,dfDay,dfYear,dfYYMMDD,dfFill];

type
  PDateValidator = ^TDateValidator;
  TDateValidator = object(TPXPictureValidator)
    constructor Init (AFormat: TDateFormats; ASeparator: Char;
                      AutoFill: Boolean);
    constructor Load (var S: TStream);
    procedure Error; virtual;
    procedure Format (var AFormat: TDateFormats); virtual;
    function IsValid (const S: String): Boolean; virtual;
    function IsValidInput (var S: string; SuppressFill: Boolean): Boolean;virtual;
    function LeapYear (AYear: Word): Boolean;
    function Separator: Char;
    procedure SetFormat (AFormat: TDateFormats);
    procedure SetSeparator (ASeparator: Char);
    procedure Store (var S: TStream);
    function Transfer (var S : String; Buffer : Pointer;
                       Flag : TVTransfer) : Word; virtual;
    private
      FFormat: TDateFormats;
      FSeparator: Char;
  end;


  PTimeRec = ^TTimeRec;
  TTimeRec = record
    Hours: Byte;
    Minutes: Byte;
    Seconds: Byte;
    Hundredths: Byte;
  end;  { of TTimeRec }

  TTimeFormat = (tfHours, tfMinutes, tfSeconds, tfHundredths, tf24Hour,
                 tfFill);
  TTimeFormats = set of TTimeFormat;

  PTimeValidator = ^TTimeValidator;
  TTimeValidator = object(TPXPictureValidator)
    constructor Init (ATimeFormat: TTimeFormats; AutoFill: Boolean);
    constructor Load (var S: TStream);
    procedure Format (var ATimeFormat: TTimeFormats);
    function Separator: Char;
    procedure SetFormat (ATimeFormat: TTimeFormats);
    procedure SetSeparator (ASeparator: Char);
    procedure Store (var S: TStream);
    function Transfer (var S : String; Buffer : Pointer;
                       Flag : TVTransfer) : Word; virtual;
  private
    FFormat: TTimeFormats;
    FSeparator: Char;
  end;  { of TTimeValidator }

type
  PDateValidatorEx = ^TDateValidatorEx;
  TDateValidatorEx = object(TPXPictureValidator)
    function GetPic:String; virtual;
  end;



procedure RegisterValidate;

const
  RDateValidator: TStreamRec = (
     ObjType: rcRDateValidator;
     VmtLink: Ofs(TypeOf(TDateValidatorEx)^);
     Load:    @TDateValidatorEx.Load;
     Store:   @TDateValidatorEx.Store
  );

const
  RPXPictureValidator: TStreamRec = (
    ObjType: rcPXPictureValidator;
    VmtLink: Ofs(TypeOf(TPXPictureValidator)^);
    Load: @TPXPictureValidator.Load;
    Store: @TPXPictureValidator.Store
  );

const
  RFilterValidator: TStreamRec = (
    ObjType: rcFilterValidator;
    VmtLink: Ofs(TypeOf(TFilterValidator)^);
    Load: @TFilterValidator.Load;
    Store: @TFilterValidator.Store
  );

const
  RRangeValidator: TStreamRec = (
    ObjType: rcRangeValidator;
    VmtLink: Ofs(TypeOf(TRangeValidator)^);
    Load: @TRangeValidator.Load;
    Store: @TRangeValidator.Store
  );

const
  RStringLookupValidator: TStreamRec = (
    ObjType: rcStringLookupValidator;
    VmtLink: Ofs(TypeOf(TStringLookupValidator)^);
    Load: @TStringLookupValidator.Load;
    Store: @TStringLookupValidator.Store
  );

implementation
uses Ext_Dos, MsgBox, Ext_Sys, Ext_Str;

 { TByteValidator }
constructor TByteValidator.Init (AMin, AMax : Byte);
begin
  if not TRangeValidator.Init(AMin,AMax) then
    Fail;
  Options := Options or voTransfer;
end;

function TByteValidator.Transfer (var S : String; Buffer : Pointer;
                                  Flag : TVTransfer) : Word;
var
  Value : Byte;
  Code : Integer;
begin
  if (Options and voTransfer) = voTransfer then
  begin
    Transfer := SizeOf(Value);
    case Flag of
      vtGetData :
        begin
          Val(S,Value,Code);
          Byte(Buffer^) := Value;
        end;
      vtSetData : Str(Byte(Buffer^),S);
    end;
  end
  else Transfer := 0;
end;

  { TDateValidator }
constructor TDateValidator.Init (AFormat: TDateFormats; ASeparator: Char;
                                 AutoFill: Boolean);
begin
  if (not TPXPictureValidator.Init('',AutoFill)) then
    Fail;
  Options := Options or voTransfer;
  Status := vsOk;
  FSeparator := ASeparator;
  SetFormat(AFormat);
end;

constructor TDateValidator.Load (var S: TStream);
begin
  if not TPXPictureValidator.Load(S) then Fail;
  S.Read(FFormat,SizeOf(FFormat));
  S.Read(FSeparator,SizeOf(FSeparator));
  if (S.Status <> stOk) then
  begin
    TPXPictureValidator.Done;
    Fail;
  end;
end;

procedure TDateValidator.Error;
type
  Str4 = string[4];
var
  S: string;
  Rec: record
    S: PString;
  end;
  function Year: Str4;
  begin
    if dfYear in FFormat then
      if df4DigitYear in FFormat then
        Year := 'YYYY'
      else Year := 'YY'
    else Year := '';
  end;
begin
  if dfYYMMDD in FFormat then
    S := Year
  else S := '';
  if dfMonth in FFormat then
    S := S + Separator + 'MM';
  if dfDay in FFormat then
    S := S + Separator + 'DD';
  if not (dfYYMMDD in FFormat) then
    S := S + Separator + Year;
  while (Length(S) > 0) and (S[1] = FSeparator) do
    Delete(S,1,SizeOf(FSeparator));
  while (Length(S) > 0) and (S[Length(S)] = FSeparator) do
    Delete(S,Length(S),SizeOf(FSeparator));
  Rec.S := PString(@S);
  MessageBox(^C + 'Invalid date format.  The date must be as follows:'#13#10 +
             #13#10 +
             ^C + S,@Rec,mfError or mfOkButton);
end;

procedure TDateValidator.Format (var AFormat: TDateFormats);
begin
  AFormat := FFormat;
end;

function TDateValidator.IsValid (const S: String): Boolean;
begin
  IsValid := (S <> '') and TPXPictureValidator.IsValid(S);
end;

function TDateValidator.IsValidInput (var S: string;
                                      SuppressFill: Boolean): Boolean;
begin
  IsValidInput := False;
  if TPXPictureValidator.IsValidInput(S,SuppressFill) then
  begin
{    Transfer(S,@Date,vtGetData); }
(*
    if (Length(S) = Length(_Pic^)) then  { check total date }
    begin
      if (dfMonth in FFormat) then
      begin
        if (Date.Month < 1) or (Date.Month > 12) then
          Exit;
        if (dfDay in FFormat) then
      end;
    end
    else begin { check what we can of the day number }
    end;
    if (dfMonth in FFormat) then
    begin
      if (Date.Month = 0) or (Date.Month > 12) then
        Exit;  { bad month number }
      if (dfDay in FFormat) then
    end;
*)
    IsValidInput := True;
  end;
end;

function TDateValidator.LeapYear (AYear: Word): Boolean;
begin
  LeapYear := (not (dfYear in FFormat)) or IsLeapYear(AYear);
end;

function TDateValidator.Separator: Char;
begin
  Separator := FSeparator;
end;

procedure TDateValidator.SetFormat (AFormat: TDateFormats);
var
  S: string[10];
  i: TDateFormat;
begin
  DisposeStr(_Pic);
  FFormat := AFormat;
  S := '';
  for i := dfMonth to dfYear do
    if TDateFormat(i) in FFormat then
      S := S + Separator + '##';
  while (Length(S) > 0) and (S[1] = Separator) do
    Delete(S,1,SizeOf(FSeparator));
  if df4DigitYear in FFormat then
    if dfYYMMDD in FFormat then
      S := '##' + S
    else S := S + '##';
  _Pic := NewStr(S);
end;

procedure TDateValidator.SetSeparator (ASeparator: Char);
var
  i: Byte;
begin
  for i := 1 to Length(_Pic^) do
    if _Pic^[i] = FSeparator then
      _Pic^[i] := ASeparator;
  FSeparator := ASeparator;
end;

procedure TDateValidator.Store (var S: TStream);
begin
  TPXPictureValidator.Store(S);
  S.Write(FFormat,SizeOf(FFormat));
  S.Write(FSeparator,SizeOf(FSeparator));
end;

function TDateValidator.Transfer (var S : String; Buffer : Pointer;
                                  Flag : TVTransfer) : Word;
  procedure BuildRec;
  var
    S1: string[11];
    function GetPart: Word;
    var
      SepPos: Byte;  { position of first separator in S }
      Part: string[4];
      N: Word;
      Code: Integer;
    begin
      SepPos := Pos(Separator,S1);
      if SepPos > 0 then
      begin
        Part := Copy(S1,1,Pred(SepPos));
        Delete(S1,1,Succ(Length(Part)));  { delete separator too }
      end
      else begin
        Part := S1;
        Delete(S1,1,Length(Part));
      end;
      Val(Part,N,Code);
      GetPart := N;
    end;
  begin
    S1 := S;
    FillChar(Buffer^,SizeOf(TDateRec),#0);
    if (dfYear in FFormat) and (dfYYMMDD in FFormat) then
        TDateRec(Buffer^).Year := GetPart;  { european format string }
    if (dfMonth in FFormat) then
      TDateRec(Buffer^).Month := GetPart;
    if (dfDay in FFormat) then
      TDateRec(Buffer^).Day := GetPart;
    if (dfYear in FFormat) and (not (dfYYMMDD in FFormat)) then
      TDateRec(Buffer^).Year := GetPart;  { english format string }
  end;
  procedure BuildStr;
  type
    Str4 = string[4];
  var
    SubStr: Str4;
    function AddNumber (N: Word): str4;
    begin
      Str(N,SubStr);
      if (dfFill in FFormat) and (Length(SubStr) < 2) then
        AddNumber := '0' + SubStr
      else AddNumber := SubStr;
    end;
    function AddPart (Part: TDateFormat): str4;
    begin
      case Part of
        dfYear:
          begin
            SubStr := AddNumber(TDateRec(Buffer^).Year);
            if not (df4DigitYear in FFormat) then
              Delete(SubStr,1,Length(SubStr)-2);
          end;
        dfMonth: SubStr := AddNumber(TDateRec(Buffer^).Month);
        dfDay: SubStr := AddNumber(TDateRec(Buffer^).Day);
      end;
      AddPart := FSeparator + SubStr;
    end;
  begin
    S := '';
    if (dfYYMMDD in FFormat) and (dfYear in FFormat) then
      S := AddPart(dfYear);  { European format }
    if (dfMonth in FFormat) then
      S := S + AddPart(dfMonth);
    if (dfDay in FFormat) then
      S := S + AddPart(dfDay);
    if (not (dfYYMMDD in FFormat)) and (dfYear in FFormat) then
      S := S + AddPart(dfYear);   { American format }
    while (S[1] = FSeparator) do
      Delete(S,1,SizeOf(FSeparator));
  end;
begin
  if Options and voTransfer <> 0 then
  begin
    Transfer := SizeOf(TDateRec);
    case Flag of
      vtGetData: BuildRec;
      vtSetData: BuildStr;
    end;
  end
  else Transfer := 0;
end;


  { TFilterValidator }

constructor TFilterValidator.Init(AValidChars: TCharSet);
begin
  inherited Init;
  ValidChars := AValidChars;
end;

constructor TFilterValidator.Load(var S: TStream);
begin
  inherited Load(S);
  S.Read(ValidChars, SizeOf(TCharSet));
end;

function TFilterValidator.IsValid(const S: string): Boolean;
var
  I: Integer;
begin
  I := 1;
  while S[I] in ValidChars do Inc(I);
  IsValid := (I > Length(S));
end;

function TFilterValidator.IsValidInput(var S: string; SuppressFill: Boolean): Boolean;
var
  I: Integer;
begin
  I := 1;
  while S[I] in ValidChars do Inc(I);
  IsValidInput := I > Length(S);
end;

procedure TFilterValidator.Store(var S: TStream);
begin
  inherited Store(S);
  S.Write(ValidChars, SizeOf(TCharSet));
end;

procedure TFilterValidator.Error;
begin
  ErrorMessage(mcInvalidCharacter, nil, mfError + mfOKButton);
end;

 { TIntegerValidator }
constructor TIntegerValidator.Init (AMin, AMax : Integer);
begin
  if not TRangeValidator.Init(AMin,AMax) then
    Fail;
  Options := Options or voTransfer;
end;

function TIntegerValidator.Transfer (var S : String; Buffer : Pointer;
                                     Flag : TVTransfer) : Word;
var
  Value : Integer;
  Code : Integer;
begin
  if (Options and voTransfer) = voTransfer then
  begin
    Transfer := SizeOf(Value);
    case Flag of
      vtGetData :
        begin
          Val(S,Value,Code);
          Integer(Buffer^) := Value;
        end;
      vtSetData : Str(Integer(Buffer^),S);
    end;
  end
  else Transfer := 0;
end;

{ TLookupValidator }

function TLookupValidator.IsValid(const S: string): Boolean;
begin
  IsValid := Lookup(S);
end;

function TLookupValidator.Lookup(const S: string): Boolean;
begin
  Lookup := True;
end;

{ TPXPictureValidator }

constructor TPXPictureValidator.Init(const APic: string;
  AutoFill: Boolean);
var
  S: String;
begin
  inherited Init;
  _Pic := NewStr(APic);
  Options := voOnAppend;
  if AutoFill then Options := Options or voFill;
  S := '';
  if Picture(S, False) <> prEmpty then
    Status := vsSyntax;
end;

constructor TPXPictureValidator.Load(var S: TStream);
begin
  inherited Load(S);
  _Pic := S.ReadStr;
end;

function TPXPictureValidator.GetPic:String;
begin
  if _Pic<>nil then GetPic:=_Pic^
  else GetPic:='';
end;

destructor TPXPictureValidator.Done;
begin
  DisposeStr(_Pic);
  inherited Done;
end;

procedure TPXPictureValidator.Error;
var S:string;
    P:PString;
begin
  S:=GetPic;
  P:=@S;
  ErrorMessage(mcInvalidPicture, @P, mfError + mfOKButton);
end;

function TPXPictureValidator.IsValidInput(var S: string;
  SuppressFill: Boolean): Boolean;
begin
  IsValidInput := (GetPic = '') or
     (Picture(S, (Options and voFill <> 0)  and not SuppressFill) <> prError);
end;

function TPXPictureValidator.IsValid(const S: string): Boolean;
var
  Str: String;
  Rslt: TPicResult;
begin
  Str := S;
  Rslt := Picture(Str, False);
  IsValid := (GetPic = '') or (Rslt = prComplete) or (Rslt = prEmpty);
end;

{$IFDEF VER70}
function IsNumber(Chr: Char): Boolean; near; assembler;
asm
        XOR     AL,AL
        MOV     Ch,Chr
        CMP     Ch,'0'
        JB      @@1
        CMP     Ch,'9'
        JA      @@1
        INC     AL
@@1:
end;

function IsSpecial(Chr: Char; const Special: string): Boolean; near;
  assembler;
asm
        XOR     AH,AH
        LES     DI,Special
        MOV     AL,ES:[DI]
        INC     DI
        MOV     CH,AH
        MOV     CL,AL
        MOV     AL,Chr
        REPNE   SCASB
        JCXZ    @@1
        INC     AH
@@1:    MOV     AL,AH
end;
{$ELSE}

function IsNumber(Chr: Char): Boolean;
const Numbers = ['0'..'9'];
begin
  IsNumber := (Chr in Numbers);
end;

function IsSpecial(Chr: Char; const Special: string): Boolean;
var
  Len, i: Byte;
begin
  Len := Length(Special);
  i := 1;
  while (i < Len) and (Chr <> Special[i]) do
    Inc(i);
  IsSpecial := (i <= Len);
end;
{$ENDIF}

function IsComplete(Rslt: TPicResult): Boolean;
begin
  IsComplete := Rslt in [prComplete, prAmbiguous];
end;


function IsIncomplete(Rslt: TPicResult): Boolean;
begin
  IsIncomplete := Rslt in [prIncomplete, prIncompNoFill];
end;

function TPXPictureValidator.Picture(var Input: string;
  AutoFill: Boolean): TPicResult;
var
  I, J: Byte;
  Rslt: TPicResult;
  Reprocess: Boolean;
  Pic:string;

  function Process(TermCh: Byte): TPicResult;
  var
    Rslt: TPicResult;
    Incomp: Boolean;
    OldI, OldJ, IncompJ, IncompI: Byte;

    { Consume input }

    procedure Consume(Ch: Char);
    begin
      Input[J] := Ch;
      Inc(J);
      Inc(I);
    end;

    { Skip a character or a picture group }

    procedure ToGroupEnd(var I: Byte);
    var
      BrkLevel, BrcLevel: Integer;
    begin
      BrkLevel := 0;
      BrcLevel := 0;
      repeat
        if I = TermCh then Exit;
        case Pic[I] of
          '[': Inc(BrkLevel);
          ']': Dec(BrkLevel);
          '{': Inc(BrcLevel);
          '}': Dec(BrcLevel);
          ';': Inc(I);
          '*':
            begin
              Inc(I);
              while IsNumber(Pic[I]) do Inc(I);
              ToGroupEnd(I);
              Continue;
            end;
        end;
        Inc(I);
      until (BrkLevel = 0) and (BrcLevel = 0);
    end;

    { Find the a comma separator }

    function SkipToComma: Boolean;
    begin
      repeat ToGroupEnd(I) until (I = TermCh) or (Pic[I] = ',');
      if Pic[I] = ',' then Inc(I);
      SkipToComma := I < TermCh;
    end;

    { Calclate the end of a group }

    function CalcTerm: Byte;
    var
      K: Byte;
    begin
      K := I;
      ToGroupEnd(K);
      CalcTerm := K;
    end;

    { The next group is repeated X times }

    function Iteration: TPicResult;
    var
      Itr, K, L: Byte;
      Rslt: TPicResult;
      NewTermCh: Byte;
    begin
      Itr := 0;
      Iteration := prError;

      Inc(I);  { Skip '*' }

      { Retrieve number }

      while IsNumber(Pic[I]) do
      begin
        Itr := Itr * 10 + Byte(Pic[I]) - Byte('0');
        Inc(I);
      end;

      if I > TermCh then
      begin
        Iteration := prSyntax;
        Exit;
      end;

      K := I;
      NewTermCh := CalcTerm;

      { If Itr is 0 allow any number, otherwise enforce the number }
      if Itr <> 0 then
      begin
        for L := 1 to Itr do
        begin
          I := K;
          Rslt := Process(NewTermCh);
          if not IsComplete(Rslt) then
          begin
            { Empty means incomplete since all are required }
            if Rslt = prEmpty then Rslt := prIncomplete;
            Iteration := Rslt;
            Exit;
          end;
        end;
      end
      else
      begin
        repeat
          I := K;
          Rslt := Process(NewTermCh);
        until not IsComplete(Rslt);
        if (Rslt = prEmpty) or (Rslt = prError) then
        begin
          Inc(I);
          Rslt := prAmbiguous;
        end;
      end;
      I := NewTermCh;
      Iteration := Rslt;
    end;

    { Process a picture group }

    function Group: TPicResult;
    var
      Rslt: TPicResult;
      TermCh: Byte;
    begin
      TermCh := CalcTerm;
      Inc(I);
      Rslt := Process(TermCh - 1);
      if not IsIncomplete(Rslt) then I := TermCh;
      Group := Rslt;
    end;

    function CheckComplete(Rslt: TPicResult): TPicResult;
    var
      J: Byte;
    begin
      J := I;
      if IsIncomplete(Rslt) then
      begin
        { Skip optional pieces }
        while True do
          case Pic[J] of
            '[': ToGroupEnd(J);
            '*':
              if not IsNumber(Pic[J + 1]) then
              begin
                Inc(J);
                ToGroupEnd(J);
              end
              else
                Break;
          else
            Break;
          end;

        if J = TermCh then Rslt := prAmbiguous;
      end;
      CheckComplete := Rslt;
    end;

    function Scan: TPicResult;
    var
      Ch: Char;
      Rslt: TPicResult;
    begin
      Scan := prError;
      Rslt := prEmpty;
      while (I <> TermCh) and (Pic[I] <> ',') do
      begin
        if J > Length(Input) then
        begin
          Scan := CheckComplete(Rslt);
          Exit;
        end;

        Ch := Input[J];
        case Pic[I] of
          '#': if not IsNumber(Ch) then Exit
               else Consume(Ch);
          '?': if not (chUpCase(ch) in ['A'..'Z',''..'','']) then Exit
               else Consume(Ch);
          '&': if not (chUpCase(ch) in ['A'..'Z',''..'','']) then Exit
               else Consume(chUpCase(Ch));
          '>': if not (chUpCase(ch) in ['A'..'Z',''..'','']) then Exit
               else Consume(chLoCase(Ch));
          '!': Consume(chUpCase(Ch));
          '@': Consume(Ch);
          '*':
            begin
              Rslt := Iteration;
              if not IsComplete(Rslt) then
              begin
                Scan := Rslt;
                Exit;
              end;
              if Rslt = prError then Rslt := prAmbiguous;
            end;
          '{':
            begin
              Rslt := Group;
              if not IsComplete(Rslt) then
              begin
                Scan := Rslt;
                Exit;
              end;
            end;
          '[':
            begin
              Rslt := Group;
              if IsIncomplete(Rslt) then
              begin
                Scan := Rslt;
                Exit;
              end;
              if Rslt = prError then Rslt := prAmbiguous;
            end;
        else
          if Pic[I] = ';' then Inc(I);
          if chUpCase(Pic[I]) <> chUpCase(Ch) then
            if Ch = ' ' then Ch := Pic[I]
            else Exit;
          Consume(Pic[I]);
        end;

        if Rslt = prAmbiguous then
          Rslt := prIncompNoFill
        else
          Rslt := prIncomplete;
      end;

      if Rslt = prIncompNoFill then
        Scan := prAmbiguous
      else
        Scan := prComplete;
    end;

  begin
    Incomp := False;
    OldI := I;
    OldJ := J;
    repeat
      Rslt := Scan;

      { Only accept completes if they make it farther in the input
        stream from the last incomplete }
      if (Rslt in [prComplete, prAmbiguous]) and Incomp and (J < IncompJ) then
      begin
        Rslt := prIncomplete;
        J := IncompJ;
      end;

      if (Rslt = prError) or (Rslt = prIncomplete) then
      begin
        Process := Rslt;
        if not Incomp and (Rslt = prIncomplete) then
        begin
          Incomp := True;
          IncompI := I;
          IncompJ := J;
        end;
        I := OldI;
        J := OldJ;
        if not SkipToComma then
        begin
          if Incomp then
          begin
            Process := prIncomplete;
            I := IncompI;
            J := IncompJ;
          end;
          Exit;
        end;
        OldI := I;
      end;
    until (Rslt <> prError) and (Rslt <> prIncomplete);

    if (Rslt = prComplete) and Incomp then
      Process := prAmbiguous
    else
      Process := Rslt;
  end;

  function SyntaxCheck: Boolean;
  var
    I: Integer;
    BrkLevel, BrcLevel: Integer;
  begin
    SyntaxCheck := False;

    if Pic = '' then Exit;

    if Pic[Length(Pic)] = ';' then Exit;
    if (Pic[Length(Pic)] = '*') and (Pic[Length(Pic) - 1] <> ';') then
      Exit;

    I := 1;
    BrkLevel := 0;
    BrcLevel := 0;
    while I <= Length(Pic) do
    begin
      case Pic[I] of
        '[': Inc(BrkLevel);
        ']': Dec(BrkLevel);
        '{': Inc(BrcLevel);
        '}': Dec(BrcLevel);
        ';': Inc(I);
      end;
      Inc(I);
    end;
    if (BrkLevel <> 0) or (BrcLevel <> 0) then Exit;

    SyntaxCheck := True;
  end;


begin
  Pic:=GetPic;
  Picture := prSyntax;
  if not SyntaxCheck then Exit;

  Picture := prEmpty;
  if Input = '' then Exit;

  J := 1;
  I := 1;

  Rslt := Process(Length(Pic) + 1);
  if (Rslt <> prError) and (Rslt <> prSyntax) and (J <= Length(Input)) then
    Rslt := prError;

  if (Rslt = prIncomplete) and AutoFill then
  begin
    Reprocess := False;
    while (I <= Length(Pic)) and
      not IsSpecial(Pic[I], '#?&>!@*{}[],'#0) do
    begin
      if Pic[I] = ';' then Inc(I);
      Input := Input + Pic[I];
      Inc(I);
      Reprocess := True;
    end;
    J := 1;
    I := 1;
    if Reprocess then
      Rslt := Process(Length(Pic) + 1)
  end;

  if Rslt = prAmbiguous then
    Picture := prComplete
  else if Rslt = prIncompNoFill then
    Picture := prIncomplete
  else
    Picture := Rslt;
end;

procedure TPXPictureValidator.Store(var S: TStream);
begin
  inherited Store(S);
  S.WriteStr(_Pic);
end;

{ TRangeValidator }

constructor TRangeValidator.Init(AMin, AMax: LongInt);
begin
  inherited Init(['0'..'9','+','-']);
  if AMin >= 0 then ValidChars := ValidChars - ['-'];
  Min := AMin;
  Max := AMax;
end;

constructor TRangeValidator.Load(var S: TStream);
begin
  inherited Load(S);
  S.Read(Min, SizeOf(Max) + SizeOf(Min));
end;

procedure TRangeValidator.Error;
var
  Params: array[0..1] of Longint;
begin
  Params[0] := Min;
  Params[1] := Max;
  ErrorMessage(mcInvalidValue, @Params, mfError + mfOKButton);
end;

function TRangeValidator.IsValid(const S: string): Boolean;
var
  Value: LongInt;
  Code: Integer;
begin
  IsValid := False;
  if inherited IsValid(S) then
  begin
    Val(S, Value, Code);
    if (Code = 0) and (Value >= Min) and (Value <= Max) then
      IsValid := True;
  end;
end;

procedure TRangeValidator.Store(var S: TStream);
begin
  inherited Store(S);
  S.Write(Min, SizeOf(Max) + SizeOf(Min));
end;

function TRangeValidator.Transfer(var S: String; Buffer: Pointer;
  Flag: TVTransfer): Word;
var
  Value: LongInt;
  Code: Integer;
begin
  if Options and voTransfer <> 0 then
  begin
    Transfer := SizeOf(Value);
    case Flag of
     vtGetData:
       begin
         Val(S, Value, Code);
         LongInt(Buffer^) := Value;
       end;
     vtSetData:
       Str(LongInt(Buffer^), S);
    end;
  end
  else
    Transfer := 0;
end;

 { TRealValidator }
constructor TRealValidator.Init (APic: string; AutoFill: Boolean;
                                 AMin, AMax : Real);
begin
  if not TPXPictureValidator.Init(APic,AutoFill) then
    Fail;
  Min := AMin;
  Max := AMax;
end;

constructor TRealValidator.Load(var S: TStream);
begin
  if not TPXPictureValidator.Load(S) then
    Fail;
  S.Read(Min,SizeOf(Min));
  S.Read(Max,SizeOf(Max));
  if S.Status <> stOk then
  begin
    TPXPictureValidator.Done;
    Fail;
  end;
end;


procedure TRealValidator.Error;
var
  Params: array[0..1] of PString;
  MinStr, MaxStr : string;
begin
  Transfer(MinStr,@Min,vtSetData);
  Transfer(MaxStr,@Max,vtSetData);
  Params[0] := PString(@MinStr);
  Params[1] := PString(@MaxStr);
  ErrorMessage(mcInvalidFValue, @Params, mfError + mfOKButton);
end;


function TRealValidator.IsValid (const S: String): Boolean;
var
  MinStr, MaxStr: string;
  Width, Decimals: Integer;
  P, S1: string;
begin
  IsValid := False;
  if GetPic = '' then
  begin
    IsValid := TPXPictureValidator.IsValid(S);
    Exit;
  end;
  P := GetPic;
  Width := Pos('.',P);
  Decimals := Length(P) - Width - 1;
  S1 := S;
  if (Pos('.',S1) = 0) then
    S1 := S1 + '.';
  if Pos('.',S1) > Width then
    Exit;  { too many numbers prior to decimal point }
  if ((Length(S1) - Width - 1) > Decimals) then
    Exit;  { too mnay numbers after decimal point }
  while ((Length(S1) - Pos('.',S1) - 1) < Decimals) do
    S1 := S1 + '0';
  Transfer(MinStr,@Min,vtSetData);
  Transfer(MaxStr,@Max,vtSetData);
  IsValid := (S1 >= MinStr) and (S1 <= MaxStr);
end;

function TRealValidator.IsValidInput (var S: string;
                                      SuppressFill: Boolean): Boolean;
const
  ValidChars = ['0'..'9','+','-','.','E','e'];
var
  I: Integer;
begin
  I := 1;
  while S[I] in ValidChars do
    Inc(I);
  if (I > Length(S)) then
    if TPXPictureValidator.IsValidInput(S,SuppressFill) then
    begin
      IsValidInput := True;
      Exit;
    end
    else begin
      { catch simple formatting that doesn't pass picture format }
      IsValidInput := (S[Length(S)] = '.');
    end
  else IsValidInput := False;
end;

procedure TRealValidator.Store(var S: TStream);
begin
  TPXPictureValidator.Store(S);
  S.Write(Min, SizeOf(Max) + SizeOf(Min));
end;

function TRealValidator.Transfer (var S : String; Buffer : Pointer;
  Flag: TVTransfer): Word;
type
  Preal=^real;
var
  Value: Real;
  Width, Decimals, Code: Integer;
begin
  if Options and voTransfer <> 0 then
  begin
    Transfer := SizeOf(Value);
    case Flag of
     vtGetData:
       begin
         Val(S, Value, Code);
         PReal(Buffer)^:=Value;
       end;
     vtSetData:
       begin
         Width := Pred(Pos('.',GetPic));
         if (Width = 0) then
         begin
           Width := Length(GetPic);
           Decimals := 0;
         end
         else Decimals := Length(GetPic) - Width - 1 { for period };
         Str(PReal(Buffer)^:Width:Decimals, S);
       end;
    end;
  end
  else
    Transfer := 0;
end;

 { TSingleValidator }
constructor TSingleValidator.Init (APic: string; AutoFill: Boolean;
                                   AMin, AMax: Single);
begin
  if not TRealValidator.Init(APic,AutoFill,AMin,AMax) then
    Fail;
end;

function TSingleValidator.Transfer (var S : String; Buffer : Pointer;
  Flag: TVTransfer): Word;
type
  PSingle=^single;
var
  Value: Single;
  Code: Integer;
begin
  if Options and voTransfer <> 0 then
  begin
    Transfer := SizeOf(Value);
    case Flag of
     vtGetData:
       begin
         Val(S, Value, Code);
         PSingle(Buffer)^:= Value;
       end;
     vtSetData:
       Str(PSingle(Buffer)^, S);
    end;
  end
  else
    Transfer := 0;
end;

{ TStringLookupValidator }

constructor TStringLookupValidator.Init(AStrings: PStringCollection);
begin
  inherited Init;
  Strings := AStrings;
end;

constructor TStringLookupValidator.Load(var S: TStream);
begin
  inherited Load(S);
  Strings := PStringCollection(S.Get);
end;

destructor TStringLookupValidator.Done;
begin
  NewStringList(nil);
  inherited Done;
end;

procedure TStringLookupValidator.Error;
begin
  ErrorMessage(mcNotInList, nil, mfError + mfOKButton);
end;

function TStringLookupValidator.Lookup(const S: string): Boolean;
var
  Index: Integer;
  Str: PString;
begin
  Str:=@s;
  Lookup := False;
  if Strings <> nil then
    Lookup := Strings^.Search(Str, Index);
end;

procedure TStringLookupValidator.NewStringList(AStrings: PStringCollection);
begin
  if Strings <> nil then Dispose(Strings, Done);
  Strings := AStrings;
end;

procedure TStringLookupValidator.Store(var S: TStream);
begin
  inherited Store(S);
  S.Put(Strings);
end;

 { TTimeValidator }
constructor TTimeValidator.Init (ATimeFormat: TTimeFormats;
                                 AutoFill: Boolean);
begin
  if (not inherited Init('',AutoFill)) then
    Fail;
  FSeparator := ':';
  SetFormat(ATimeFormat);
  Options := Options or voTransfer;
end;

constructor TTimeValidator.Load (var S: TStream);
begin
  if not TPXPictureValidator.Load(S) then
    Fail;
  S.Read(FFormat,SizeOf(FFormat));
  S.Read(FSeparator,SizeOf(FSeparator));
  if (S.Status <> stOk) then
  begin
    TPXPictureValidator.Done;
    Fail;
  end;
end;

procedure TTimeValidator.Format (var ATimeFormat: TTimeFormats);
begin
  ATimeFormat := FFormat;
end;

function TTimeValidator.Separator: Char;
begin
  Separator := FSeparator;
end;

procedure TTimeValidator.SetFormat (ATimeFormat: TTimeFormats);
var
  P: string[10];
  Part: TTimeFormat;
begin
    { build new picture }
  P := '';
  for Part := tfHours to tfHundredths do
    if (Part in ATimeFormat) then
      P := P + FSeparator + '##';
  if (Pos(Separator,P) = 1) then
     Delete(P,1,SizeOf(FSeparator));
  DisposeStr(_Pic);
  _Pic := NewStr(P);
  FFormat := ATimeFormat;
end;

procedure TTimeValidator.Store (var S: TStream);
begin
  TPXPictureValidator.Store(S);
  S.Write(FFormat,SizeOf(FFormat));
  S.Write(FSeparator,SizeOf(FSeparator));
end;

procedure TTimeValidator.SetSeparator (ASeparator: Char);
begin
  FSeparator := ASeparator;
  SetFormat(FFormat);
end;

function TTimeValidator.Transfer (var S : String; Buffer : Pointer;
                                  Flag : TVTransfer) : Word;
  procedure BuildRec;
  var
    S1: string[11];
    function GetPart: Word;
    var
      SubStr: string[2];
      Width: Byte;
      N: Byte;
      Code: Integer;
    begin
      Width := Pos(FSeparator,S1);
      SubStr := Copy(S1,1,Width);
      Delete(S1,1,Succ(Width));
      Val(SubStr,N,Code);
      GetPart := N;
    end;
  begin
    with TTimeRec(Buffer^) do
    begin
      if (tfHours in FFormat) then
        Hours := GetPart;
      if (tfMinutes in FFormat) then
        Minutes := GetPart;
      if (tfSeconds in FFormat) then
        Seconds := GetPart;
      if (tfHundredths in FFormat) then
        Hundredths := GetPart;
    end;
  end;
  procedure BuildStr;
  type
    str2 = string[2];
    function AddPart (Part: TTimeFormat): str2;
      function AddNumber (N: Word): str2;
      var
        S2: str2;
      begin
        Str(N,S2);
        if (tfFill in FFormat) and (Length(S2) < 2) then
          S2 := '0' + S2;
        AddNumber := S2;
      end;
    var
      SubStr: str2;
    begin
      with TTimeRec(Buffer^) do
        case Part of
          tfHours: SubStr := AddNumber(Hours);
          tfMinutes: SubStr := AddNumber(Minutes);
          tfSeconds: SubStr := AddNumber(Seconds);
          tfHundredths: SubStr := AddNumber(Hundredths);
      end;
      AddPart := FSeparator + SubStr;
    end;
  var
    i: TTimeFormat;
  begin
    for i := tfHours to tfHundredths do
      if (i in FFormat) then
        S := S + AddPart(i);
    while (S[1] = FSeparator) do
      Delete(S,1,SizeOf(FSeparator));
  end;
begin
  if Options and voTransfer <> 0 then
  begin
    Transfer := SizeOf(TTimeRec);
    case Flag of
      vtGetData: BuildRec;
      vtSetData: BuildStr;
    end;
  end
  else Transfer := 0;
end;

{ TValidator }
constructor TValidator.Init;
begin
  if not TObject.Init then
    Fail;
  Status := 0;
  Options := 0;
end;

constructor TValidator.Load (var S : TStream);
begin
  if not TObject.Init then
    Fail;
  Status := 0;
  S.Read(Options, SizeOf(Options));
  if S.Status <> stOk then
  begin
    TObject.Done;
    Fail;
  end;
end;

procedure TValidator.Error;
begin
end;

function TValidator.IsValidInput (var S : String;
                                  SuppressFill : Boolean) : Boolean;
begin
  IsValidInput := True;
end;

function TValidator.IsValid (const S : String) : Boolean;
begin
  IsValid := True;
end;

procedure TValidator.Store (var S : TStream);
begin
  S.Write(Options,SizeOf(Options));
end;

function TValidator.Transfer (var S : String; Buffer : Pointer;
                              Flag : TVTransfer) : Word;
begin
  Transfer := 0;
end;

function TValidator.Valid (const S : String) : Boolean;
begin
  Valid := False;
  if not IsValid(S) then
  begin
    Error;
    Exit;
  end;
  Valid := True;
end;

 { TWordValidator }
constructor TWordValidator.Init (AMin, AMax : Word);
begin
  if not TRangeValidator.Init(AMin,AMax) then
    Fail;
  Options := Options or voTransfer;
end;

function TWordValidator.Transfer (var S : String; Buffer : Pointer;
                                  Flag : TVTransfer) : Word;
var
  Value : Word;
  Code : Integer;
begin
  if (Options and voTransfer) = voTransfer then
  begin
    Transfer := SizeOf(Word);
    case Flag of
      vtGetData :
        begin
          Val(S,Value,Code);
          Word(Buffer^) := Value;
        end;
      vtSetData : Str(Word(Buffer^),S);
    end;
  end
  else Transfer := 0;
end;

  { TDateValidatorEx }
function TDateValidatorEx.GetPic:String;
{$IFDEF VER70}
var Result:string;
{$ENDIF}
begin
  if ExpandDate then
     Result:='{#[#]}'+DateSeparator+'{#[#]}'+DateSeparator+'{#[#[#[#]]]}'
  else Result:='{#[#]}'+DateSeparator+'{#[#]}'+DateSeparator+'{#[#]}';
  Result:=Result+'[ {[#]#}'+TimeSeparator+'{[#]#}'+TimeSeparator+'{[#]#}]';
{$IFDEF VER70}
  GetPic:=Result;
{$ENDIF}
end;

{ Validate registration procedure }
procedure RegisterValidate;
begin
  RegisterType(RPXPictureValidator);
  RegisterType(RFilterValidator);
  RegisterType(RRangeValidator);
  RegisterType(RStringLookupValidator);
  RegisterType(RDateValidator);
end;

end.