{*******************************************************}
{                                                       }
{       Turbo Pascal 7.0                                }
{       FreePacal                                       }
{       Delphi 4                                        }
{       New index object                                }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O-}
{$ENDIF}
{$I Define.inc}
unit IDX_Unit;
interface
uses Objects, DB;
const
  nfLeaf = 2;
  nfRoot = 1;
  IDXPageSize  = 512;
type
  PIDXHeader = ^TIDXHeader;
  TIDXHeader = packed record
    RootOffset:LongInt;{⥫  ୥ 設}
    FreeRecord:LongInt;{⥫  ᢮  ᯨ᪥ 設}
    FileLen   :LongInt;{⥫   䠩}
    KeyLen    :Word;
    Flags     :Byte;
    ISign     :Byte;
    KeyExp    :array[0..219] of char;
    ForExp    :array[0..219] of char;
    Reserved  :array[0..55] of Char;
  end;

type
  PIDXStream = ^TIDXStream;
  TIDXStream = object(TIndexStream)
     Header:TIDXHeader;
     constructor Init(AFileName:FNameStr; Mode:word);
     destructor Done;virtual;
     function NewPage:longint;virtual;
     procedure FreePage(L:Longint);virtual;
     function MaxItems:integer;
  end;

type
  PIDXNode = ^TIDXNode;
  TIDXNode = object(TNode)
    procedure Store;virtual;
    procedure Load(ARef:Longint);virtual;
    function Count:word;virtual;
    function Key(Num:integer):string;virtual;
    function PageRef(Num:integer):longint;virtual;
    function RecordNumber(Num:integer):longint;virtual;
    function Search(S:string; PtrRecord:Longint; var Num:integer):boolean;virtual;
    function LeftNode:longint;
    function RightNode:longint;
    function NodeAttr:word;
    function Insert(var S:string;var PtrRecord:longint):PNode;virtual;
    function Delete(var S:string;var PtrRecord:longint):boolean;virtual;
  end;

type
  PIDXIndex = ^TIDXIndex;
  TIDXIndex = object(TIndexObject)
     constructor Init(ALink:PDataSet; AFileName:FNameStr);
     function KeySize:word;virtual;
     function NextKey:Longint;virtual;
     function PrevKey:Longint;virtual;
     function Unique:Boolean;virtual;
     function FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;virtual;
     function FirstKey:Longint;virtual;
     function LastKey:Longint;virtual;
     function IndexExpresion:string;virtual;
     function GetNode(ARef:Longint):PNode;virtual;
     function Curentkey:longint;virtual;
     procedure EmptyIndex; virtual;
     procedure AddKey( AKey:String; ALink:LongInt);virtual;
     procedure DeleteKey( AKey:String; ALinkPos:longint);virtual;
  end;

implementation
uses Strings, Ext_Type, Ext_Str;

  { TIDXStream }
constructor TIDXStream.Init(AFileName:FNameStr; Mode:word);
begin
  inherited Init(AFileName, Mode);
  Read(Header, SizeOf(TIDXHeader));
end;

destructor TIDXStream.Done;
var P:Pointer;
begin
  if Modified then
  begin
    Seek(0);
    Write(Header, IDXPageSize);
  end;
  inherited Done;
end;

function TIDXStream.NewPage:longint;
var L:Longint;
    P:Pointer;
begin
  if Header.FreeRecord=-1 then L:=GetSize
  else
  begin
    L:=Header.FreeRecord;
    Seek(Header.FreeRecord);
    Read(Header.FreeRecord, SizeOf(Longint));
  end;
  NewPage:=L;
  GetMem(P,IDXPageSize);
  FillChar(P,IDXPageSize,0);
  Seek(L);
  Write(P,IDXPageSize);
  FreeMem(P,IDXPageSize);
end;

procedure TIDXStream.FreePage(L:Longint);
begin
  Seek(L);
  write(Header.FreeRecord, SizeOf(Longint));
  Header.FreeRecord:=L;
end;

function TIDXStream.MaxItems:integer;
begin
  MaxItems:=(IDXPageSize-12) div Header.KeyLen;
end;

  { TIDXNode }
procedure TIDXNode.Store;
begin
  inherited Store;
  Stream^.Seek(Handle);
  Stream^.Write(Buffer^, IDXPageSize)
end;

procedure TIDXNode.Load(ARef:Longint);
begin
  inherited Load(ARef);
  Stream^.Seek(ARef);
  Stream^.Read(Buffer^, IDXPageSize);
end;
{$IFDEF VER70}
function TIDXNode.Count:word;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     AX,ES:[DI+2]
end;

function TIDXNode.Key(Num:integer):string;assembler;
asm
        PUSH    DS
        LDS     SI,Self
        LES     DI,DS:[SI].TNode.Stream
        MOV     AX,WORD PTR ES:[DI].TIDXStream.Header.KeyLen
        MOV     CX,AX
        LES     DI,@Result
        STOSB
        ADD     AX,4
        MUL     NUM
        ADD     AX,12
        LDS     SI,DS:[SI].TNode.Buffer
        ADD     SI,AX
        CLD
    REP MOVSB
        POP     DS
end;

function TIDXNode.PageRef(Num:integer):longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     AX,WORD PTR ES:[DI].TIDXStream.Header.KeyLen
        MOV     BX,AX
        ADD     AX,4
        MUL     NUM
        ADD     AX,12
        ADD     BX,AX
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     DX,ES:[DI+BX]
        MOV     AX,ES:[DI+BX+2]
        XCHG    DH,DL
        XCHG    AH,AL
end;


function TIDXNode.RecordNumber(Num:integer):longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     AX,WORD PTR ES:[DI].TIDXStream.Header.KeyLen
        MOV     BX,AX
        ADD     AX,4
        MUL     NUM
        ADD     AX,12
        ADD     BX,AX
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     DX,ES:[DI+BX]
        MOV     AX,ES:[DI+BX+2]
        XCHG    DH,DL
        XCHG    AH,AL
end;

function TIDXNode.LeftNode:longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     AX,ES:[DI+4]
        MOV     DX,ES:[DI+6]
end;

function TIDXNode.RightNode:longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     AX,ES:[DI+8]
        MOV     DX,ES:[DI+10]
end;

function TIDXNode.NodeAttr:word;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     AX,ES:[DI]
end;

function TIDXNode.Insert(var S:string;var PtrRecord:longint):PNode;
var c:integer;
    Node:PIdxNode;
procedure _Insert;assembler;
asm
        PUSH    DS
        LDS     SI,Self
        LES     DI,DS:[SI].TNode.Stream
        MOV     AX,WORD PTR ES:[DI].TIDXStream.Header.KeyLen
        MOV     BX,AX
        LES     DI,DS:[SI].TNode.Buffer {  }
        MOV     CX,ES:[DI]
        MUL     CX
        ADD     AX,12
        XCHG    AX,BX
        ADD     DI,BX
        ADD     DI,AX
        LDS     SI,DS:[SI].TNode.Buffer
        ADD     SI,BX
        MOV     BX,AX
        INC     CX
        SUB     CX,C
        MUL     CX
        MOV     CX,AX
        STD
    REP MOVSB
        LES     DI,Self                       {㢥稢 ⢮ 祩}
        LES     DI,ES:[DI].TNode.Buffer
        INC     WORD PTR ES:[DI].TNode.Buffer

        LDS     SI,PtrRecord
        MOV     AX,BX                         {                  }
        MOV     CX,C
        DEC     CX
        MUL     CX
        ADD     AX,12                        {塞 ᬥ襭  ⠢}
        ADD     DI,AX
        CLD
        LODSW
        XCHG    AL,AH                       { 祭 㪠⥫}
        MOV     DX,AX                       {८ࠧ㥬   Fox ଠ}
        LODSW                               {࠭塞 祭}
        XCHG    AL,AH
        STOSW
        MOV     AX,DX
        STOSW

        PUSH    DI                           {塞       }
        MOV     CX,BX                        {ﬨ  }
        XOR     AX,AX                        {}
    REP STOSB
        POP     DI

        LDS     SI,S                          {㦠  ப}
        XOR     AX,AX
        LODSB
        CMP     AX,BX                         {    }
        JB      @1                            {  ப }
        MOV     AX,BX                         {     }
@1:     MOV     CX,AX
    REP MOVSB                                 {࠭塞   }
        POP     DS
end;
begin
  abstract;
{  if Count=PIDXStream(Stream)^.MaxItems then
  begin
    Node:=New(PIdxNode, Init(Stream, Stream^.NewPage,IDXPageSize));
    PLongArray(Node^.Buffer)^[2]:=LeftNode;
    PLongArray(Node^.Buffer)^[1]:=Handle;
    PLongArray(Buffer)^[2]:=Node^.Handle;
    PWordArray(Node^.Buffer)^[1]:=Count div 2;
    PWordArray(Node^.Buffer)^[0]:=NodeAttr;
    NewPage(Node);
    move(Buffer, Node^.Buffer,(count-(count div 2))*PIDXStream(Stream)^.Header.KeyLen)
  end;
  Search(S, -1, C);
  _Insert;}
end;
{$ELSE}
function TIDXNode.Count:word;
begin
end;
function TIDXNode.Key(Num:integer):string;
begin
end;
function TIDXNode.PageRef(Num:integer):longint;
begin
end;
function TIDXNode.RecordNumber(Num:integer):longint;
begin
end;
function TIDXNode.LeftNode:longint;
begin
end;
function TIDXNode.RightNode:longint;
begin
end;
function TIDXNode.NodeAttr:word;
begin
end;
function TIDXNode.Insert(var S:string;var PtrRecord:longint):PNode;
begin
end;
{$ENDIF}
function TIDXNode.Search(S:string; PtrRecord:Longint; var Num:integer):boolean;
var
  L,H,C:longint;
begin
  Search := False;
  L := 0;
  H := Count - 1;
  while L <= H do
  begin
    Num := (L + H) shr 1;
    C := Compare(Key(Num),S);
    if (C = 0) and (PtrRecord<>-1) then C:=RecordNumber(Num)-PtrRecord;
    if C < 0 then L := Num + 1 else
    begin
      H := Num - 1;
      Search:=C=0;
    end;
  end;
  Num:=L;
end;

function TIDXNode.Delete(var S:string;var PtrRecord:longint):boolean;
begin
end;

  { TIDXIndex }
constructor TIDXIndex.Init(ALink:PDataSet; AFileName:FNameStr);
begin
  inherited Init(ALink);
  IndexStream:=PIDXStream(Session^.GetDBStream(AFileName));
  if IndexStream=nil then
     IndexStream:=New(PIDXStream, Init(AFileName, stOpen));
  FileName:=IndexStream^.FileName;
  IndexField:=Owner^.FieldByName(IndexExpresion);
end;

function TIDXIndex.KeySize:word;
begin
  KeySize:=PIDXStream(IndexStream)^.Header.KeyLen;
end;

function TIDXIndex.NextKey:Longint;
var
  LL:integer;
procedure _NextKey(_PP:Integer;var L:integer);
var Node:PIdxNode;
begin
  L:=-1;
  if _PP=0 then exit;
  PNode(Node):=GetPage(@Self, PagePtrArray[_PP].Ref);
  if PagePtrArray[_PP].CurItem<Node^.Count-1 then
  begin
     Inc(PagePtrArray[_PP].CurItem);
     L:=Node^.RecordNumber(PagePtrArray[_PP].CurItem);
  end
  else
  if Node^.RightNode<>-1 then
  begin
    PagePtrArray[_PP].Ref:=Node^.RightNode;
    PNode(Node):=GetPage(@Self, PagePtrArray[_PP].Ref);
    PagePtrArray[_PP].CurItem:=0;
    L:=Node^.RecordNumber(0);
  end
  else if PP=_PP then iStatus:=iStatus+[iEofFind];
end;
begin
  if iEofFind in iStatus then LL:=-1
  else
  if PP=0 then LL:=FirstKey
  else
  begin
     iStatus:=iOk;
    _NextKey(PP,LL);
  end;
   NextKey:=LL;
end;

function TIDXIndex.PrevKey:Longint;
var
  LL:integer;
procedure _PrevKey(_PP:Integer;var L:integer);
var Node:PIdxNode;
begin
  L:=-1;
  if _PP=0 then exit;
  PNode(Node):=GetPage(@Self, PagePtrArray[_PP].Ref);
  if PagePtrArray[_PP].CurItem>0 then
  begin
     Dec(PagePtrArray[_PP].CurItem);
     L:=Node^.RecordNumber(PagePtrArray[_PP].CurItem);
  end
  else
  if Node^.LeftNode<>-1 then
  begin
    PagePtrArray[_PP].Ref:=Node^.LeftNode;
    PNode(Node):=GetPage(@Self, PagePtrArray[_PP].Ref);
    PagePtrArray[_PP].CurItem:=Node^.Count-1;
    L:=Node^.RecordNumber(PagePtrArray[_PP].CurItem);
  end
  else if PP=_PP then iStatus:=iStatus+[iBofFind];
end;
begin
  if iBofFind in iStatus then LL:=-1
  else if PP=0 then LL:=LastKey
  else
  begin
    iStatus:=iOk;
    _PrevKey(PP,LL);
  end;
   PrevKey:=LL;
end;

function TIDXIndex.Unique:Boolean;
begin
  Unique:=PIDXStream(IndexStream)^.Header.Flags and 1<>0;
end;

function TIDXIndex.FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;
var
  Node:PIdxNode;
  L:integer;
  Ref:longint;
  LL:longint;
procedure DoFind(LNode:PIdxNode);
var RNode:PIdxNode;
begin
   if LNode^.RightNode<>-1 then
   begin
      RNode:=PIdxNode(GetPage(@Self, LNode^.RightNode));
      if RNode^.Search(Key, PtrRec, L) then
      begin
        FindKey:=RNode^.RecordNumber(L);
        PagePtrArray[PP].CurItem:=L;
        PagePtrArray[PP].Ref:=LNode^.RightNode;
        exit;
      end
      else
      if L>=RNode^.Count then DoFind(RNode)
      else LL:=-1;
   end
   else LL:=-1;
end;
begin
  if PP<>0 then
  begin
    Node:=PIdxNode(GetPage(@Self, PagePtrArray[PP].Ref));
    if PtrRec=Node^.PageRef(PagePtrArray[PP].CurItem) then
       FindKey:=PtrRec;
  end;
  ClearKey;
  FindKey:=0;
  Ref:=PIDXStream(IndexStream)^.Header.RootOffset;
  repeat
    Node:=PIdxNode(GetPage(@Self, Ref));
    inc(pp);
    PagePtrArray[PP].Ref:=Ref;
    if PIdxNode(Node)^.NodeAttr and nfLeaf<>0 then LL:=PtrRec else LL:=-1;
    if Node^.Search(Key, LL, L) and (PIdxNode(Node)^.NodeAttr and nfLeaf<>0) then
    begin
      FindKey:=Node^.RecordNumber(L);
      PagePtrArray[PP].CurItem:=L;
      exit;
    end
    else
    begin
      PagePtrArray[PP].CurItem:=L;
      Ref:=Node^.PageRef(L);
    end;
  until PIdxNode(Node)^.NodeAttr and nfLeaf<>0;
  if (L>=Node^.Count) and (PtrRec>=0) then DoFind(Node);
  if LL=-1 then
  begin
    FindKey:=PagePtrArray[PP].CurItem;
    iStatus:=iStatus+[iKeyNotFound];
  end;
end;

function TIDXIndex.FirstKey:Longint;
var
  Node:PIdxNode;
  Ref:longint;
begin
  ClearKey;
  Ref:=PIDXStream(IndexStream)^.Header.RootOffset;
  Node:=GetPage(@Self, Ref);
  inc(pp);
  PagePtrArray[PP].Ref:=Ref;
  PagePtrArray[PP].CurItem:=0;
  Ref:=Node^.PageRef(0);
  while PIdxNode(Node)^.NodeAttr and nfLeaf = 0 do
  begin
    Node:=GetPage(@Self, Ref);
    inc(pp);
    PagePtrArray[PP].Ref:=Ref;
    PagePtrArray[PP].CurItem:=0;
    Ref:=Node^.PageRef(0);
  end;
  FirstKey:=Node^.RecordNumber(0);
end;

function TIDXIndex.LastKey:Longint;
var
  Node:PIdxNode;
  Ref:longint;
begin
  ClearKey;
  Ref:=PIDXStream(IndexStream)^.Header.RootOffset;
  Node:=GetPage(@Self, Ref);
  inc(pp);
  PagePtrArray[PP].Ref:=Ref;
  PagePtrArray[PP].CurItem:=Node^.Count-1;
  Ref:=Node^.PageRef(PagePtrArray[PP].CurItem);
  while PIdxNode(Node)^.NodeAttr and nfLeaf = 0 do
  begin
    Node:=GetPage(@Self, Ref);
    inc(pp);
    PagePtrArray[PP].Ref:=Ref;
    PagePtrArray[PP].CurItem:=Node^.Count-1;
    Ref:=Node^.PageRef(PagePtrArray[PP].CurItem);
  end;
  LastKey:=Node^.RecordNumber(PagePtrArray[PP].CurItem);
end;

function TIDXIndex.IndexExpresion:string;
begin
  IndexExpresion:=StrPas(@PIDXStream(IndexStream)^.Header.KeyExp)
end;

function TIDXIndex.GetNode(ARef:Longint):PNode;
begin
  GetNode:=New(PIdxNode, Init(IndexStream, ARef, IdxPageSize));
end;

function TIDXIndex.Curentkey:longint;
var Node:PIdxNode;
begin
  Node:=GetPage(@Self, PagePtrArray[PP].Ref);
  Curentkey:=Node^.PageRef(PagePtrArray[PP].CurItem);
end;

procedure TIDXIndex.EmptyIndex;
begin
end;

procedure TIDXIndex.AddKey( AKey:String; ALink:LongInt);
var Node:PIdxNode;
    LeafNode:boolean;
procedure DoFind;
var
  L:integer;
  Ref:longint;
  LL:longint;
procedure _DoFind(LNode:PIdxNode);
var RNode:PIdxNode;
procedure ShiftReferehs(APP:integer);
var TopNode:PIdxNode;
begin
  if APP=0 then exit;
  TopNode:=PIdxNode(GetPage(@Self,PagePtrArray[APP].Ref));
  inc(PagePtrArray[APP].CurItem);
  if (PagePtrArray[APP].CurItem<TopNode^.Count) or
     (TopNode^.RightNode=-1 ) then exit;
  PagePtrArray[APP].Ref:=TopNode^.RightNode;
  TopNode:=PIdxNode(GetPage(@Self,TopNode^.RightNode));
  PagePtrArray[APP].CurItem:=0;
  ShiftReferehs(APP-1);
end;
begin
   if LNode^.RightNode<>-1 then
   begin
      RNode:=PIdxNode(GetPage(@Self, LNode^.RightNode));
      ShiftReferehs(PP-1);
      if RNode^.Search(AKey, ALink, L) then
      begin
        PagePtrArray[PP].CurItem:=L;
        PagePtrArray[PP].Ref:=LNode^.RightNode;
      end
      else
      if L>=RNode^.Count then _DoFind(RNode)
      else LL:=-1;
   end
   else LL:=-1;
end;
begin
  ClearKey;
  Ref:=PIDXStream(IndexStream)^.Header.RootOffset;
  repeat
    Node:=PIdxNode(GetPage(@Self, Ref));
    LeafNode:=Node^.NodeAttr and nfLeaf<>0;
    inc(pp);
    PagePtrArray[PP].Ref:=Ref;
    if LeafNode then LL:=ALink else LL:=-1;
    if Node^.Search(AKey, LL, L) and LeafNode then
    begin
      PagePtrArray[PP].CurItem:=L;
      exit;
    end
    else
    begin
      PagePtrArray[PP].CurItem:=L;
      Ref:=Node^.PageRef(L);
    end;
  until LeafNode;
  if (L>=Node^.Count) and (ALink>=0) then _DoFind(Node);
  if LL=-1 then iStatus:=iStatus + [iKeyNotFound];
end;

begin
  DoFind;
end;

procedure TIDXIndex.DeleteKey( AKey:String; ALinkPos:longint);
begin
end;
end.

