{*******************************************************}
{                                                       }
{       Turbo Pascal 7.0                                }
{       FreePacal                                       }
{       Delphi 4                                        }
{       New index object                                }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O-}
{$ENDIF}
{$I Define.inc}
unit IND_Unit;
interface
uses Objects, DAE, Dos, DBFTable, DAE_Type, Ext_Type;
const
  defIndExt = '.IND';
type
  PTaItem = ^TaItem;
  TaItem = packed record
    DataRef : LongInt;
    PageRef : LongInt;
    Key:array[1..255] of char;
  end;

type
  PNodeBuffer = ^TNodeBuffer;
  TNodeBuffer = packed record
    Count:Word;
    BckwPageRef:Longint;
    Items:array [0..0] of byte;
  end;

type
  PINDHeader = ^TINDHeader;
  TINDHeader = packed record
    FirstFree:Longint;
    FistNode:Longint;
    NodeSize:word;
    ItemSize:word;
    KeySize:word;
    PageSize:word;
    HalfPage:word;
    Descending:Boolean;
    Unique:boolean;
    CaseSensitive:Boolean;
    KeyExp:string;
    FileName:NameStr;
  end;

type
  PINDStream = ^TINDStream;
  TINDStream = object(TIndexStream)
     Header:TINDHeader;
     constructor Init(AFileName:FNameStr; Mode:word);
     function NewPage:longint;virtual;
     procedure FreePage(L:Longint);virtual;
     procedure WriteHeader;virtual;
     procedure CopyStructure(S:PStream);virtual;
  end;

type
  PIndNode = ^TIndNode;
  TIndNode = object(TNode)
     procedure Store;virtual;
     procedure Load(ARef:Longint);virtual;
     function ItemsOnPage:word;
     function Key(Num:integer):string;
     function PageRef(Num:integer):longint;
     function DataRef(Num:integer):longint;
     function BckwPageRef:longint;
     function Search(S:string; PtrRecord:longint; var Num:integer):boolean;

     procedure SetBckwPageRef(L:longint);
     procedure SetCount(ACount:word);

     procedure GetItemArray(Num:integer; var Rec:TaItem);
     procedure SetItemArray(Num:integer; var Rec:TaItem);
  end;

type
  PINDIndex = ^TINDIndex;
  TINDIndex = object(TIndexObject)
     Header:PINDHeader;
     constructor Init(ALink:PDataSet; AFileName:FNameStr);
     constructor Create(ALink:PDataSet; AFileName:FNameStr; AField:string;
                        AOptions:TIndexOptions);
     function KeySize:word;virtual;
     function NextKey:Longint;virtual;
     function PrevKey:Longint;virtual;
     function Unique:Boolean;virtual;
     function FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;virtual;
     function FirstKey:Longint;virtual;
     function LastKey:Longint;virtual;
     function IndexExpresion:string;virtual;
     function GetNode(ARef:Longint):PNode;virtual;
     procedure EmptyIndex; virtual;
     procedure AddKey( AKey:String; ALink:LongInt);virtual;
     procedure DeleteKey( AKey:String; ALink:longint);virtual;
     function Open:Boolean;virtual;
     procedure RenameIndex;virtual;
  end;

implementation
uses Ext_Str, Ext_Dos, Ext_Math;

constructor TINDStream.Init(AFileName:FNameStr; Mode:word);
begin
  inherited Init(AFileName, Mode);
  if Mode<>stCreate then Read(Header, SizeOf(TINDHeader));
end;

procedure TINDStream.WriteHeader;
begin
  if Modified then
  begin
    Seek(0);
    Write(Header, SizeOf(TINDHeader));
  end;
end;

function TINDStream.NewPage:longint;
var P:Pointer;
begin
  Modified:=true;
  if Header.FirstFree<>0 then
  begin
    NewPage:=Header.FirstFree;
    Seek(Header.FirstFree);
    Read(Header.FirstFree, SizeOf(int32));
  end
  else
  begin
    NewPage:=GetSize;
    Seek(GetSize);
    GetMem(P, Header.NodeSize);
    FillChar(P^, Header.NodeSize, 0);
    Write(P^, Header.NodeSize);
    FreeMem(P, Header.NodeSize);
  end;
end;

procedure TINDStream.FreePage(L:Longint);
begin
  Seek(L);
  Write(Header.FirstFree, SizeOf(int32));
  Header.FirstFree:=L;
end;

procedure TINDStream.CopyStructure(S:PStream);
begin
  S^.Write(Header, SizeOf(TINDHeader));
end;

  { TIndNode }
procedure TIndNode.Store;
begin
  inherited Store;
  Stream^.Seek(Handle);
  Stream^.Write(Buffer^, PINDStream(Stream)^.Header.NodeSize)
end;

procedure TIndNode.Load(ARef:Longint);
begin
  inherited Load(ARef);
  Stream^.Seek(ARef);
  Stream^.Read(Buffer^, PINDStream(Stream)^.Header.NodeSize);
end;
{$IFDEF VER70}
function TIndNode.ItemsOnPage:word;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     AX,ES:[DI]
end;
{$ELSE}
function TIndNode.ItemsOnPage:word;
begin
  Result:=PNodeBuffer(Buffer)^.Count;
end;
{$ENDIF}

{$IFDEF VER70}
function TIndNode.BckwPageRef:longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     AX,ES:[DI+2]
        MOV     DX,ES:[DI+4]
end;
{$ELSE}
function TIndNode.BckwPageRef:longint;
begin
  Result:=PNodeBuffer(Buffer)^.BckwPageRef;
end;
{$ENDIF}

{$IFDEF VER70}
procedure TIndNode.SetBckwPageRef(L:longint);assembler;
asm
        MOV     AX,L.WORD[0]
        MOV     DX,L.WORD[2]
        LES     DI,Self
        MOV     ES:[DI].TNode.Modified,1
        LES     DI,ES:[DI].TNode.Buffer
        MOV     ES:[DI+2],AX
        MOV     ES:[DI+4],DX
end;
{$ELSE}
procedure TIndNode.SetBckwPageRef(L:longint);
begin
  Modified:=true;
  PNodeBuffer(Buffer)^.BckwPageRef:=L;
end;
{$ENDIF}

{$IFDEF VER70}
procedure TIndNode.SetCount(ACount:word);assembler;
asm
        MOV     AX,ACount
        LES     DI,Self
        MOV     ES:[DI].TNode.Modified,1
        LES     DI,ES:[DI].TNode.Buffer
        MOV     ES:[DI],AX
end;
{$ELSE}
procedure TIndNode.SetCount(ACount:word);
begin
  Modified:=true;
  PNodeBuffer(Buffer)^.Count:=ACount;
end;
{$ENDIF}

{$IFDEF VER70}
function TIndNode.Key(Num:integer):string;assembler;
asm
        PUSH    DS
        LES     DI,Self
        LDS     SI,ES:[DI].TNode.Buffer
        LES     DI,ES:[DI].TNode.Stream
        MOV     BX,ES:[DI].TINDStream.Header.ItemSize
        MOV     AX,ES:[DI].TINDStream.Header.KeySize
        MOV     CX,AX
        LES     DI,@Result
        STOSB
        MOV     AX,Num
        DEC     AX
        MUL     BX
        ADD     AX,14
        ADD     SI,AX
    REP MOVSB
        POP     DS
end;
{$ELSE}
function TIndNode.Key(Num:integer):string;
var P:PTaItem;
begin
  P:=@(PNodeBuffer(Buffer)^.Items[PINDStream(Stream)^.Header.ItemSize*(Num-1)]);
  SetLength(Result,PINDStream(Stream)^.Header.KeySize);
  Move(P^.Key, Result[1], PINDStream(Stream)^.Header.KeySize);
end;
{$ENDIF}

{$IFDEF VER70}
function TIndNode.PageRef(Num:integer):longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     CX,ES:[DI].TINDStream.Header.ItemSize
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     AX,Num
        DEC     AX
        MUL     CX
        MOV     BX,AX
        ADD     BX,10
        MOV     AX,ES:[DI+BX]
        MOV     DX,ES:[DI+BX+2]
end;
{$ELSE}
function TIndNode.PageRef(Num:integer):longint;
var P:PTaItem;
begin
  P:=@(PNodeBuffer(Buffer)^.Items[PINDStream(Stream)^.Header.ItemSize*(Num-1)]);
  PageRef:=P^.PageRef;
end;
{$ENDIF}

{$IFDEF VER70}
function TIndNode.DataRef(Num:integer):longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     CX,ES:[DI].TINDStream.Header.ItemSize
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     AX,Num
        DEC     AX
        MUL     CX
        MOV     BX,AX
        ADD     BX,6
        MOV     AX,ES:[DI+BX]
        MOV     DX,ES:[DI+BX+2]
end;
{$ELSE}
function TIndNode.DataRef(Num:integer):longint;
var P:PTaItem;
begin
  P:=@(PNodeBuffer(Buffer)^.Items[PINDStream(Stream)^.Header.ItemSize*(Num-1)]);
  DataRef:=P^.DataRef;
end;
{$ENDIF}

function TIndNode.Search(S:string; PtrRecord:longint; var Num:integer):boolean;
var
  L,R:integer;
  C:longint;
begin
  L:=1;
  R:=ItemsOnPage;
  repeat
    Num:=(L + R) div 2;
    C:=ASCIICompare(S,Key(Num));
    {if PINDStream(Stream)^.Header.Descending then C:=C*-1;}
    if (C=0) and (PtrRecord<>-1) then C:=PtrRecord-DataRef(Num);
    if C<=0 then R := Num - 1;
    if C>=0 then L := Num + 1;
  until R < L;
  if L-R>1 then Search:=true
  else
  begin
    Num:=R;
    Search:=false;
  end;
end;


{$IFDEF VER70}
procedure TIndNode.GetItemArray(Num:integer; var Rec:TaItem);assembler;
asm
        MOV     BX,DS
        LES     DI,Self
        LDS     SI,ES:[DI].TNode.Buffer
        LES     DI,ES:[DI].TNode.Stream
        MOV     CX,ES:[DI].TINDStream.Header.ItemSize
        MOV     AX,Num
        DEC     AX
        MUL     CX
        ADD     AX,6
        ADD     SI,AX
        LES     DI,Rec
    REP MOVSB
        MOV     DS,BX
end;
{$ELSE}
procedure TIndNode.GetItemArray(Num:integer; var Rec:TaItem);
var P:Pointer;
begin
  P:=@(PNodeBuffer(Buffer)^.Items[PINDStream(Stream)^.Header.ItemSize*(Num-1)]);
  Move(P^, Rec, PINDStream(Stream)^.Header.ItemSize);
end;
{$ENDIF}

{$IFDEF VER70}
procedure TIndNode.SetItemArray(Num:integer; var Rec:TaItem);assembler;
asm
        MOV     BX,DS
        LDS     SI,Self
        MOV     ES:[DI].TNode.Modified,1
        LES     DI,DS:[SI].TNode.Buffer
        LDS     SI,DS:[SI].TNode.Stream
        MOV     CX,DS:[SI].TINDStream.Header.ItemSize
        MOV     AX,Num
        DEC     AX
        MUL     CX
        ADD     AX,6
        ADD     DI,AX
        LDS     SI,Rec
    REP MOVSB
        MOV     DS,BX
end;
{$ELSE}
procedure TIndNode.SetItemArray(Num:integer; var Rec:TaItem);
var P:Pointer;
begin
  P:=@(PNodeBuffer(Buffer)^.Items[PINDStream(Stream)^.Header.ItemSize*(Num-1)]);
  Move(Rec,P^, PINDStream(Stream)^.Header.ItemSize);
  Modified:=true;
end;
{$ENDIF}

  { TINDIndex }
constructor TINDIndex.Init(ALink:PDataSet; AFileName:FNameStr);
begin
  inherited Init(ALink);
  FileName:=NewStr(ReplaceExt(AFileName, defIndExt,false));
  if not Open then
  begin
    DisposeStr(FileName);
    fail;
  end;
end;

function TINDIndex.Open:boolean;
begin
  inherited Open;
  if IndexStream=nil then
     IndexStream:=New(PIndStream, Init(FileName^, stOpen));
  if (IndexStream<>nil) and (IndexStream^.Status=stOk) then
  begin
    Header:=@PIndStream(IndexStream)^.Header;
    if ExtractName(Owner^.TableName^)=Header^.FileName then
    begin
      DisposeStr(FileName);
      FileName:=PIndexStream(IndexStream)^.FileName;
      IndexField:=GetIndexsField(Owner, IndexExpresion, Options);
      if Header^.Unique then Options:=Options+[ixUnique];
      if Header^.CaseSensitive then Options:=Options+[ixCaseInsensitive];
      Open:=true;
      exit;
    end;
    Session^.FreeDBStream(PIndexStream(IndexStream));
    DBError(edbIndexFileMistMath, FileName^);
  end;
  Open:=false;
end;

constructor TINDIndex.Create(ALink:PDataSet; AFileName:FNameStr; AField:string;
                        AOptions:TIndexOptions);
begin
  AFileName:=ReplaceExt(AFileName, defIndExt, false);
  inherited Create(ALink, {AFileName, AField, }AOptions);
  IndexField:=GetIndexsField(Owner, AField, AOptions);
  Options:=Options+AOptions;
  IndexStream:=PIndStream(Session^.GetDBStream(AFileName));
  if (IndexStream<>nil) or (IndexField=nil) then
  begin
     Session^.FreeDBStream(PIndexStream(IndexStream));
     if ixExpression in Options then IndexField^.Free;
     if IndexStream<>nil then DBError(edbFileAlreadyOpen, AFileName);
     fail;
   end;
   IndexStream:=New(PIndStream, Init(AFileName, stCreate));
   Header:=@PIndStream(IndexStream)^.Header;
   Header^.FileName:=ExtractName(ALink^.TableName^);
   FileName:=PIndexStream(IndexStream)^.FileName;
   with Header^ do
   begin
     KeySize:=IndexField^.Size;
     ItemSize:=KeySize+8;
     NodeSize:=((2048 div ItemSize) and $FFFE) * ItemSize;
     PageSize:=NodeSize div ItemSize;
     NodeSize:=NodeSize+6;
     HalfPage:=PageSize div 2;
     KeyExp:=AField;
     Unique:=ixUnique in Options;
     CaseSensitive:=ixCaseInsensitive in Options;
     Descending:=ixDescending in Options;
   end;
   PIndexStream(IndexStream)^.Modified:=true;
   PIndStream(IndexStream)^.WriteHeader;
   Reindexs;
end;

procedure TINDIndex.RenameIndex;
var NewName:string;
begin
  IndexStream:=New(PIndStream, Init(FileName^, stCreate));
  Header:=@PIndStream(IndexStream)^.Header;
  Header^.FileName:=ExtractName(Owner^.TableName^);
  PIndexStream(IndexStream)^.Modified:=true;
  PIndStream(IndexStream)^.WriteHeader;
  IndexStream^.Free;
  NewName:=ReplaceExt(Owner^.TableName^, defIndExt, true);
  if FileExists(NewName) then DeleteFile(NewName);
  RenameFile(FullName(FileName^), FullName(NewName));
  AssignStr(FileName, NewName);
end;

function TINDIndex.KeySize:word;
begin
  KeySize:=Header^.KeySize
end;

function TINDIndex.NextKey:Longint;
var
  R:LongInt;
  Node:PIndNode;
begin
  if iEofFind in iStatus then exit;
  Move(PagePtrArray,OldPagePtrArray,SizeOf(TPagePtrArray));
  OldPP:=PP;
  iStatus:=iStatus - [iEofFind];
    if PP=0 then R:=Header^.FistNode
    else
    begin
      Node:=PIndNode(GetPage(@Self,PagePtrArray[PP].Ref));
      R:=Node^.PageRef(PagePtrArray[PP].CurItem);
    end;
    while R <> 0 do
    begin
      Inc(PP);
      PagePtrArray[PP].CurItem:=0;
      PagePtrArray[PP].Ref:=R;
      Node:=PIndNode(GetPage(@Self,PagePtrArray[PP].Ref));
      R:=Node^.BckwPageRef;
    end;
    if PP <> 0 then
    begin
      while (PP>1) and (PagePtrArray[PP].CurItem=Node^.ItemsOnPage) do
      begin
        Dec(PP);
        Node:=PIndNode(GetPage(@Self,PagePtrArray[PP].Ref));
      end;
      if PagePtrArray[PP].CurItem < Node^.ItemsOnPage then
      begin
          Inc(PagePtrArray[PP].CurItem);
          NextKey:=Node^.DataRef(PagePtrArray[PP].CurItem);
          if EndRange(Node^.Key(PagePtrArray[PP].CurItem)) then PP:=0;
      end
      else PP := 0;
    end;
  if PP=0 then
  begin
    iStatus:=iStatus + [iEofFind];
    Move(OldPagePtrArray,PagePtrArray,SizeOf(TPagePtrArray));
    PP:=OldPP;
  end
  else iStatus:=iStatus - [iBofFind];
end;

function TINDIndex.PrevKey:Longint;
var
  R:LongInt;
  Node:PIndNode;
begin
  if iBofFind in iStatus then exit;
  Move(PagePtrArray,OldPagePtrArray,SizeOf(TPagePtrArray));
  OldPP:=PP;
  iStatus:=iStatus - [iBofFind];
  if PP = 0 then R:=Header^.FistNode
  else
  begin
    Node:=PIndNode(GetPage(@Self,PagePtrArray[PP].Ref));
    Dec(PagePtrArray[PP].CurItem);
    if PagePtrArray[PP].CurItem=0 then R:=Node^.BckwPageRef
    else R:=Node^.PageRef(PagePtrArray[PP].CurItem);
  end;
  while R <> 0 do
  begin
      Node:=PIndNode(GetPage(@Self,R));
      Inc(PP);
      PagePtrArray[PP].CurItem:=Node^.ItemsOnPage;
      PagePtrArray[PP].Ref:=R;
      R:=Node^.PageRef(Node^.ItemsOnPage);
  end;
  if PP <> 0 then
  begin
      while (PP > 1) and (PagePtrArray[PP].CurItem = 0) do
      begin
         Dec(PP);
         Node:=PIndNode(GetPage(@Self,PagePtrArray[PP].Ref));
      end;
      if PagePtrArray[PP].CurItem > 0 then
      begin
         PrevKey:=Node^.DataRef(PagePtrArray[PP].CurItem);
         if StartRange(Node^.Key(PagePtrArray[PP].CurItem)) then PP:=0;
      end
      else PP := 0;
  end;
  if PP=0 then
  begin
    iStatus:=iStatus + [iBofFind];
    Move(OldPagePtrArray,PagePtrArray,SizeOf(TPagePtrArray));
    PP:=OldPP;
  end
  else iStatus:=iStatus - [iEofFind];
end;

function TINDIndex.Unique:Boolean;
begin
  Unique:=Header^.Unique;
end;

function TINDIndex.FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;
var
  PrPgRef:Longint;
  Node:PINDNode;
  R:integer;
{  S:String;}
begin
  if Header^.CaseSensitive then Key:=stUpCase(Key);
  ClearKey;
  PrPgRef:=Header^.FistNode;
  Node:=nil;
  while PrPgRef<>0 do
  begin
    inc(PP);
    PagePtrArray[PP].Ref := PrPgRef;
    Node:=PIndNode(GetPage(@Self,PrPgRef));
    if Node^.Search(Key, PtrRec, R) then
    begin
      FindKey:=Node^.DataRef(R);
      PagePtrArray[PP].CurItem:=R;
      if EndRange(Node^.Key(PagePtrArray[PP].CurItem)) then iStatus:=iStatus+[iBofFind];
      if StartRange(Node^.Key(PagePtrArray[PP].CurItem)) then iStatus:=iStatus+[iEofFind];
      exit;
    end;
    if R = 0 then PrPgRef:=Node^.BckwPageRef
    else PrPgRef:=Node^.PageRef(R);
    PagePtrArray[PP].CurItem:=R;
  end;
  if PP > 0 then
  begin
     while (PP > 1) and (PagePtrArray[PP].CurItem=0) do  Dec(PP);
     if PagePtrArray[PP].CurItem=0 then
     begin
       PP:=0;
       Node:=nil;
     end;
  end;
  if SoftFind then
  begin
    FindKey:=NextKey;
    if PP<>0 then
    begin
       Node:=PIndNode(GetPage(@Self,PagePtrArray[PP].Ref));
       if iEofFind in iStatus then
          if R = 0 then FindKey:=Node^.BckwPageRef
          else FindKey:=Node^.PageRef(R);
    end;
  end
  else FindKey:=-1;
  if (node=nil) or StartRange(Node^.Key(PagePtrArray[PP].CurItem)) then iStatus:=iStatus+[iEofFind];
  if (node=nil) or EndRange(Node^.Key(PagePtrArray[PP].CurItem)) then iStatus:=iStatus+[iBofFind];
end;

function TINDIndex.FirstKey:Longint;
var S:string;
begin
  ClearKey;
  S:=StartRangeValue;
  if S<>'' then FirstKey:=FindKey(S,0, true)
  else FirstKey:=NextKey;
end;

function TINDIndex.LastKey:Longint;
var S:string;
    L:Longint;
    Node:PIndNode;
begin
  ClearKey;
  S:=EndRangeValue;
  if S<>'' then
  begin
    L:=FindKey(S, MaxLongint, false);
    if L<>-1 then LastKey:=L
    else
    begin
      if PP=0 then
      begin
        iStatus:=[iEofFind, iBofFind];
        LastKey:=-1;
      end
      else
      begin
        Node:=PIndNode(GetPage(@Self,PagePtrArray[PP].Ref));
        LastKey:=Node^.DataRef(PagePtrArray[PP].CurItem);
      end;
    end;
  end
  else LastKey:=PrevKey;
end;

function TINDIndex.IndexExpresion:string;
begin
  IndexExpresion:=Header^.KeyExp;
end;

function TINDIndex.GetNode(ARef:Longint):PNode;
begin
  GetNode:=New(PIndNode, Init(IndexStream, ARef, Header^.NodeSize));
end;

procedure TINDIndex.EmptyIndex;
begin
  if PIndexStream(IndexStream)^.CountUse>1 then DBError(edbNonExclus, Owner^.TableName^)
  else
  begin
    ClearKey;
    Header^.FirstFree:=0;
    Header^.FistNode:=0;
    PIndexStream(IndexStream)^.WriteHeader;
    PIndexStream(IndexStream)^.Truncate;
  end;
end;

procedure TINDIndex.AddKey( AKey:String; ALink:LongInt);
var
  Node1,
  Node2:PIndNode;
  PassUp    : Boolean;

{  I:Integer;}
  Left,C,num:Longint;
  ProcItem1,
  ProcItem2,
  ProcItem3 : TaItem;


procedure Search(PrPgRef1 : LongInt);
var
  R : Integer;
procedure Insert;
var
  i : integer;
begin
  Node1:=PIndNode(GetPage(@Self, PrPgRef1));
  if Node1^.ItemsOnPage < Header^.PageSize then
  begin
     Node1^.SetCount(Node1^.ItemsOnPage+1);
     for I := Node1^.ItemsOnPage downto R + 2 do
     begin
        Node1^.GetItemArray(I-1,ProcItem3);
        Node1^.SetItemArray(I,ProcItem3);
     end;
     Node1^.SetItemArray(R+1, ProcItem1);
     PassUp := false;
  end
  else
  begin
      Node2:=PIndNode(GetPage(@Self, PIndexStream(IndexStream)^.NewPage));
      Node1:=PIndNode(GetPage(@Self, PrPgRef1));
      if R <= Header^.HalfPage then
      begin
        if R = Header^.HalfPage then ProcItem2 := ProcItem1
        else
        begin
          Node1^.GetItemArray(Header^.HalfPage,ProcItem2);
          for I := Header^.HalfPage downto R + 2 do
          begin
            Node1^.GetItemArray(I-1,ProcItem3);
            Node1^.SetItemArray(I,ProcItem3);
          end;
          Node1^.SetItemArray(R+1,ProcItem1);
        end;
        for I := 1 to Header^.HalfPage do
        begin
          Node1^.GetItemArray(I+Header^.HalfPage,ProcItem3);
          Node2^.SetItemArray(I,ProcItem3);
        end;
      end
      else
      begin
        R := R - Header^.HalfPage;
        Node1^.GetItemArray(Header^.HalfPage+1,ProcItem2);
        for I := 1 to R - 1 do
        begin
          Node1^.GetItemArray(I + Header^.HalfPage + 1,ProcItem3);
          Node2^.SetItemArray(I,ProcItem3);
        end;
        Node2^.SetItemArray(R,ProcItem1);
        for I := R + 1 to Header^.HalfPage do
        begin
          Node1^.GetItemArray(I + Header^.HalfPage,ProcItem3);
          Node2^.SetItemArray(I,ProcItem3);
        end;
      end;
      Node1^.SetCount(Header^.HalfPage);
      Node2^.SetCount(Header^.HalfPage);
      Node2^.SetBckwPageRef(ProcItem2.PageRef);
      ProcItem2.PageRef := Node2^.Handle;
      ProcItem1 := ProcItem2;
      Node2^.Modified:=true;
  end;
  Node1^.Modified:=true;
end; { Insert }

begin { Search }
  if PrPgRef1 = 0 then
  begin
    PassUp := true;
    FillChar(ProcItem1, SizeOf(TaItem),0);
    Move(AKey[1],ProcItem1.Key,Min(Length(AKey),Header^.KeySize));
    ProcItem1.DataRef:=ALink;
  end
  else
  begin
    Node1:=PIndNode(GetPage(@Self, PrPgRef1));
    with Node1^ do
    begin
       Left := 1;
       R := ItemsOnPage;
       repeat
         Num:=(Left + R) div 2;
         C:=ASCIICompare(AKey,Key(Num));
         if C=0 then C:=ALink-DataRef(Num);
         if C<=0 then R := Num - 1;
         if C>=0 then Left := Num + 1;
       until R < Left;
    end;
    if R = 0 then Search(Node1^.BckwPageRef)
    else Search(Node1^.PageRef(R));
    if PassUp then  Insert;
  end;
end; { Search }

begin { AddKey }
  if Header^.CaseSensitive then AKey:=stUpCase(AKey);
  Search(Header^.FistNode);
  if PassUp then
  begin
    Node1:=PIndNode(GetPage(@Self, PIndexStream(IndexStream)^.NewPage));
    Node1^.SetBckwPageRef(Header^.FistNode);
    Node1^.SetCount(1);
    Node1^.SetItemArray(1,ProcItem1);
    Node1^.Modified:=true;
    Header^.FistNode:=Node1^.Handle;
  end;
  ClearKey;
end;

procedure TINDIndex.DeleteKey( AKey:String; ALink:longint);
var
  PageTooSmall:Boolean;
  Node:PIndNode;
  ProcItem3 : TaItem;

procedure DelB(PrPgRef : LongInt);
var
  XPageRef  : LongInt;
  R,I:integer;
  L,K:integer;
  C:longint;
  Node:PIndNode;

procedure UnderFlow(PrPgRef,
                    PrPgRef2 : LongInt;
                    R        : Integer);
var
  I,K,
  LItem : Integer;
  LPageRef : LongInt;
  Node,
  Node2,
  LNode:PIndNode;
begin
  Node:=PIndNode(GetPage(@Self, PrPgRef));
  Node2:=PIndNode(GetPage(@Self, PrPgRef2));
  if R < Node^.ItemsOnPage then
  begin
    R:=R+1;
    LPageRef := Node^.PageRef(R);
    LNode:=PIndNode(GetPage(@Self, LPageRef));
    K := (LNode^.ItemsOnPage - Header^.HalfPage + 1) div 2;
    Node^.GetItemArray(R, ProcItem3);
    ProcItem3.PageRef:=LNode^.BckwPageRef;
    Node2^.SetItemArray(Header^.HalfPage, ProcItem3);
    if K > 0 then
    begin
      for I := 1 to K - 1 do
      begin
        LNode^.GetItemArray(I,ProcItem3);
        Node2^.SetItemArray(I+Header^.HalfPage,ProcItem3);
      end;
      LNode^.GetItemArray(K,ProcItem3);
      ProcItem3.PageRef:=LPageRef;
      Node^.SetItemArray(R,ProcItem3);
      LNode^.SetBckwPageRef(LNode^.PageRef(K));
      LNode^.SetCount(LNode^.ItemsOnPage-K);
      for I := 1 to LNode^.ItemsOnPage do
      begin
        LNode^.GetItemArray(I+K, ProcItem3);
        LNode^.SetItemArray(I, ProcItem3);
      end;
      Node2^.SetCount(Header^.HalfPage - 1 + K);
      PageTooSmall:=false;
    end
    else
    begin
      for I:= 1 to Header^.HalfPage do
      begin
        LNode^.GetItemArray(I, ProcItem3);
        Node2^.SetItemArray(I+Header^.HalfPage, ProcItem3);
      end;
      for I:=R to Node^.ItemsOnPage - 1 do
      begin
        Node^.GetItemArray(I+1, ProcItem3);
        Node^.SetItemArray(I, ProcItem3);
      end;
      Node2^.SetCount(Header^.PageSize);
      Node^.SetCount(Node^.ItemsOnPage-1);
      FreePage(LNode);
      PageTooSmall := Node^.ItemsOnPage<Header^.HalfPage;
    end;
  end
  else
  begin
    if R = 1 then LPageRef:=Node^.BckwPageRef
    else LPageRef:=Node^.PageRef(R-1);
    LNode:=PIndNode(GetPage(@Self, LPageRef));
    LItem := LNode^.ItemsOnPage+1;
    K := (LItem - Header^.HalfPage) div 2;
    if K > 0 then
    begin
      for I := Header^.HalfPage - 1 downto 1 do
      begin
        Node2^.GetItemArray(I, ProcItem3);
        Node2^.SetItemArray(I+K, ProcItem3);
      end;
      Node^.GetItemArray(R, ProcItem3);
      ProcItem3.PageRef:=Node2^.BckwPageRef;
      Node2^.SetItemArray(K, ProcItem3);
      LItem := LItem - K;
      for I := K - 1 downto 1 do
      begin
        LNode^.GetItemArray(I+LItem, ProcItem3);
        Node2^.SetItemArray(I, ProcItem3);
      end;
      Node2^.SetBckwPageRef(LNode^.PageRef(LItem));
      LNode^.GetItemArray(LItem, ProcItem3);
      ProcItem3.PageRef:=PrPgRef2;
      Node^.SetItemArray(R, ProcItem3);
      LNode^.SetCount(LItem-1);
      Node2^.SetCount(Header^.HalfPage-1+K);
      PageTooSmall := false;
    end
    else
    begin
      Node^.GetItemArray(R, ProcItem3);
      ProcItem3.PageRef:=Node2^.BckwPageRef;
      LNode^.SetItemArray(LItem, ProcItem3);
      for I:=1 to Header^.HalfPage-1 do
      begin
        Node2^.GetItemArray(I, ProcItem3);
        LNode^.SetItemArray(I+LItem, ProcItem3);
      end;
      LNode^.SetCount(Header^.PageSize);
      Node^.SetCount(Node^.ItemsOnPage-1);
      FreePage(Node2);
      PageTooSmall := Node^.ItemsOnPage < Header^.HalfPage;
    end;
  end;
end; { UnderFlow }


procedure DelA(PrPgRef2 : LongInt);
var
  C : Integer;
  XPageRef : LongInt;
  Node2:PIndNode;
begin
  Node2:=PIndNode(GetPage(@Self, PrPgRef2));
  XPageRef := Node2^.PageRef(Node2^.ItemsOnPage);
  if XPageRef <> 0 then
  begin
     C := Node2^.ItemsOnPage;
     DelA(XPageRef);
     if PageTooSmall then UnderFlow(PrPgRef2,XPageRef,C);
  end
  else
  begin
     Node:=PIndNode(GetPage(@Self, PrPgRef));
     Node2^.GetItemArray(Node2^.ItemsOnPage, ProcItem3);
     ProcItem3.PageRef:=Node^.PageRef(K);
     Node^.SetItemArray(K, ProcItem3);
     Node2^.SetCount(Node2^.ItemsOnPage-1);
     PageTooSmall:=Node2^.ItemsOnPage<Header^.HalfPage;
  end;
end; { DelA }

begin { DelB }
  if PrPgRef = 0 then PageTooSmall := false
  else
  begin
    Node:=PIndNode(GetPage(@Self, PrPgRef));
      L := 1;
      R := Node^.ItemsOnPage;
      repeat
        K := (L + R) div 2;
        C:=ASCIICompare(AKey,Node^.Key(K));
        if (C=0) then C:=ALink-Node^.DataRef(K);
        if C <= 0 then
          R := K - 1;
        if C >= 0 then
          L := K + 1;
      until L > R;
    if R = 0 then XPageRef:=Node^.BckwPageRef
    else XPageRef:=Node^.PageRef(R);
    if L-R>1 then
    begin
      if XPageRef = 0 then
      begin
        Node^.SetCount(Node^.ItemsOnPage-1);
        PageTooSmall:=Node^.ItemsOnPage<Header^.HalfPage;
        for I:=K to Node^.ItemsOnPage do
        begin
          Node^.GetItemArray(I+1,ProcItem3);
          Node^.SetItemArray(I,ProcItem3);
        end;
      end
      else
      begin
        DelA(XPageRef);
        if PageTooSmall then
            UnderFlow(PrPgRef,XPageRef,R);
      end;
    end
    else
    begin
      DelB(XPageRef);
      if PageTooSmall then
         UnderFlow(PrPgRef,XPageRef,R);
    end;
  end;
end; { DelB }

begin { DeleteKey }
  if Header^.CaseSensitive then AKey:=stUpCase(AKey);
  DelB(Header^.FistNode);
  if PageTooSmall then
  begin
    Node:=PIndNode(GetPage(@Self, Header^.FistNode));
    if Node^.ItemsOnPage=0 then
    begin
       Header^.FistNode:=Node^.BckwPageRef;
       FreePage(Node);
    end;
  end;
  ClearKey;
end;

const
   OldOnOpenIndex:TIndexOpen=nil;
   OldOnCreateIndex:TIndexCreate=nil;

function INDOpen(FileName:FNameStr; var Ext:ExtStr; DS:PDataSet; OnOpenTable:boolean):PIndexObject;far;
begin
  if Ext=defIndExt then
     IndOpen:=New(PIndIndex, Init(DS, FileName))
  else
  if Assigned(OldOnOpenIndex) then
     IndOpen:=OldOnOpenIndex(FileName,Ext,DS, OnOpenTable)
  else IndOpen:=nil;
end;

function IndCreate( FileName:FNameStr; var Ext:ExtStr; DS:PDataSet;
                    AField:string; AOptions:TIndexOptions):PIndexObject;far;
begin
  if Ext=defIndExt then IndCreate:=New(PIndIndex, Create(DS, FileName, AField, AOptions))
  else
  if Assigned(OldOnCreateIndex) then
     IndCreate:=OldOnCreateIndex(FileName,Ext, DS, AField, AOptions)
  else IndCreate:=nil;
end;

begin
   OldOnOpenIndex:=OnOpenIndex;
   OldOnCreateIndex:=OnCreateIndex;
   {$IFNDEF FPC}
   OnCreateIndex:=IndCreate;
   OnOpenIndex:=IndOpen;
   {$ELSE}
   OnCreateIndex:=@IndCreate;
   OnOpenIndex:=@IndOpen;
   {$ENDIF}
end.

{ ࠭                                                      }
{  ᬥ饭      祭                                          }
{  0         2         - ⮢  ࠭                       }
{  2         4         뫪  ᫥ ࠭  訬 - }
{  //  㧫                                                         }
{  n+0       4         뫪                                    }
{  n+4       4         易 㧥                                    }
{  n+8       KeySize   祢 ࠦ                                }
{  ( 易 㧫  ⥪饣                     }
{     ᫥饣)                                                }

