{*******************************************************}
{                                                       }
{       Turbo Pascal 7.0                                }
{       FreePacal                                       }
{       Delphi 4                                        }
{       New index object                                }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O-}
{$ENDIF}
{$I Define.inc}
unit NDX_Unit;
interface
uses Objects, DB;
const
  NDXPageSize = 512;
type
  PNDXHeader = ^TNDXHeader;
  TNDXHeader = record
    StartKeyPage:Longint;
    TotalPages:Longint;
    Reserved:Longint;
    IndexKeyLen:word;
    MaxKeyPage:word;
    NdxKeyType:word;
    SizeKeyRecord:word;
    Reserved1:word;
    Reserved2:byte;
    Unique:boolean;
    KeyName:array [1..488] of char;
  end;
type
  PNDXStream = ^TNDXStream;
  TNDXStream = object(TIndexStream)
    Header:TNDXHeader;
    constructor Init(AFileName:FNameStr; Mode:word);
    destructor Done;virtual;
  end;

type
  PNdxNode = ^TNdxNode;
  TNdxNode = object(TNode)
    procedure Store;virtual;
    procedure Load(ARef:Longint);virtual;
    function Count:word;virtual;
    function Key(Num:integer):string;virtual;
    function PageRef(Num:integer):longint;virtual;
    function RecordNumber(Num:integer):longint;virtual;
    function LastRef:longint;virtual;
    function Search(S:string; PtrRecord:Longint; var Num:integer):boolean;virtual;
  end;

type
  PNdX_Index = ^TNdX_Index;
  TNdX_Index = object(TIndexObject)
     constructor Init(ALink:PDataSet; AFileName:FNameStr);
     destructor Done;virtual;
     function KeySize:word;virtual;
     function NextKey:Longint;virtual;
     function PrevKey:Longint;virtual;
     function Unique:Boolean;virtual;
     function FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;virtual;
     function FirstKey:Longint;virtual;
     function LastKey:Longint;virtual;
     function IndexExpresion:string;virtual;
     function GetNode(ARef:Longint):PNode;virtual;
     function Curentkey:longint;virtual;
     procedure EmptyIndex; virtual;
  end;

implementation
uses Strings;

  { TNDXStream }
constructor TNDXStream.Init(AFileName:FNameStr; Mode:word);
begin
  inherited Init(AFileName, Mode);
  Read(Header, SizeOf(TNDXHeader));
end;

destructor TNDXStream.Done;
begin
  if Modified then
  begin
    Seek(0);
    Write(Header, NDXPageSize);
  end;
  inherited Done;
end;

    { TNdxNode }
procedure TNdxNode.Store;
begin
  inherited Store;
  Stream^.Seek(Handle shl 9);
  Stream^.Write(Buffer^, NDXPageSize)
end;

procedure TNdxNode.Load(ARef:Longint);
begin
  if Modified then Store;
  inherited Load(ARef);
  Stream^.Seek(ARef shl 9);
  Stream^.Read(Buffer^, NDXPageSize);
end;
{$IFDEF VER70}
function TNdxNode.Count:word;assembler;
asm
  LES DI,Self
  LES DI,ES:[DI].TNode.Buffer
  MOV AX,ES:[DI]
end;

function TNdxNode.Key(Num:integer):string;assembler;
asm
        PUSH    DS
        LDS     SI,Self
        LES     DI,DS:[SI].TNode.Stream
        MOV     AX,WORD PTR ES:[DI].TNDXStream.Header.SizeKeyRecord
        MOV     CX,WORD PTR ES:[DI].TNDXStream.Header.IndexKeyLen
        MUL     NUM
        LDS     SI,DS:[SI].TNode.Buffer
        MOV     BX,AX
        ADD     BX,12
        ADD     SI,BX
        LES     DI,@Result
        CLD
        MOV     AL,CL
        STOSB
    REP MOVSB
        POP     DS
end;

function TNdxNode.PageRef(Num:integer):longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     AX,WORD PTR ES:[DI].TNDXStream.Header.SizeKeyRecord
        MUL     NUM
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     BX,AX
        ADD     BX,4
        MOV     AX,ES:[DI+BX]
        MOV     DX,ES:[DI+BX+2]
end;

function TNdxNode.RecordNumber(Num:integer):longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     AX,WORD PTR ES:[DI].TNDXStream.Header.SizeKeyRecord
        MUL     NUM
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     BX,AX
        ADD     BX,8
        MOV     AX,ES:[DI+BX]
        MOV     DX,ES:[DI+BX+2]
end;

function TNdxNode.LastRef:longint;assembler;
asm
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Stream
        MOV     AX,WORD PTR ES:[DI].TNDXStream.Header.SizeKeyRecord
        LES     DI,Self
        LES     DI,ES:[DI].TNode.Buffer
        MOV     BX,ES:[DI]
        MUL     BX
        MOV     BX,AX
        ADD     BX,4
        MOV     AX,ES:[DI+BX]
        MOV     DX,ES:[DI+BX+2]
end;
{$ELSE}
function TNdxNode.Count:word;
begin
end;
function TNdxNode.Key(Num:integer):string;
begin
end;
function TNdxNode.PageRef(Num:integer):longint;
begin
end;
function TNdxNode.RecordNumber(Num:integer):longint;
begin
end;
function TNdxNode.LastRef:longint;
begin
end;
{$ENDIF}
function TNdxNode.Search(S:string; PtrRecord:Longint; var Num:integer):boolean;

var
  L,H,C:integer;
  S1:string;
begin
{  Search := False;
  L := 0;
  H := Count - 1;
  while L <= H do
  begin
    Num := (L + H) shr 1;
    C := Compare(Key(Num),S);
    if C < 0 then L := Num + 1 else
    begin
      H := Num - 1;
      if C = 0 then Search := True;
    end;
  end;
  Num:=L;}
end;

  {TNDX_Index}
constructor TNDX_Index.Init(ALink:PDataSet; AFileName:FNameStr);
begin
  inherited Init(ALink);
  IndexStream:=PNDXStream(Session^.GetDBStream(AFileName));
  if IndexStream=nil then
     IndexStream:=New(PNDXStream, Init(AFileName, stOpen));
  FileName:=IndexStream^.FileName;
  IndexField:=Owner^.FieldByName(IndexExpresion);
end;

destructor TNDX_Index.Done;
begin
  inherited Done;
end;

function TNDX_Index.FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;
var
  Node:PNode;
  L:integer;
  Ref:longint;
begin
  ClearKey;
  FindKey:=0;
{  Ref:=PNDXStream(IndexStream)^.Header.RootNode;
  while Ref<>0 do
  begin
    Node:=GetPage(@Self, Ref);
    inc(pp);
    PagePtrArray[PP].Ref:=Ref;
    if Node^.Search(Key, L) then
    begin
      FindKey:=Node^.RecordNumber(L);
      PagePtrArray[PP].CurItem:=L;
      exit;
    end
    else
    begin
      PagePtrArray[PP].CurItem:=L;
      if L>=Node^.Count then Ref:=Node^.LastRef
      else Ref:=Node^.PageRef(L);
    end;
  end;}
end;

function TNDX_Index.FirstKey:Longint;
var
  Node:PNdxNode;
  Ref:longint;
  S:string;
begin
  ClearKey;
  Ref:=PNDXStream(IndexStream)^.Header.StartKeyPage;
  while Ref<>0 do
  begin
    Node:=GetPage(@Self, Ref);
    S:=Node^.Key(Node^.Count-1);
{    writeln(S);}
    inc(pp);
    PagePtrArray[PP].Ref:=Ref;
    PagePtrArray[PP].CurItem:=0;
    Ref:=Node^.PageRef(0);
  end;
  FirstKey:=Node^.RecordNumber(0);
end;

function TNDX_Index.LastKey:Longint;
var
  Node:PNode;
  L:integer;
  Ref:longint;
begin
  ClearKey;
{  Ref:=PNDXStream(IndexStream)^.Header.RootNode;
  while Ref<>0 do
  begin
    Node:=GetPage(@Self, Ref);
    inc(pp);
    PagePtrArray[PP].Ref:=Ref;
    PagePtrArray[PP].CurItem:=Node^.Count;
    Ref:=Node^.LastRef;
  end;
  LastKey:=Node^.RecordNumber(Node^.Count-1);}
end;

function TNDX_Index.IndexExpresion:string;
begin
  IndexExpresion:=StrPas(@PNDXStream(IndexStream)^.Header.KeyName)
end;

function TNDX_Index.KeySize:word;
begin
  KeySize:=PNDXStream(IndexStream)^.Header.SizeKeyRecord;
end;

function TNDX_Index.NextKey:Longint;
var Node:PNode;
    L:longint;
begin
{  if PP = 0 then L:=PNDXStream(IndexStream)^.Header.RootNode
  else
  begin
    Node:=GetPage(@Self, PagePtrArray[PP].Ref);
    if PagePtrArray[PP].CurItem<Node^.Count then
       L:=Node^.PageRef(PagePtrArray[PP].CurItem)
    else L:=Node^.LastRef;
  end;
  while L<>0 do
  begin
    Node:=GetPage(@Self, L);
    inc(PP);
    PagePtrArray[PP].Ref:=L;
    PagePtrArray[PP].CurItem:=-1;
    L:=Node^.PageRef(0);
  end;
  if PP<>0 then
  begin
    while (PP>1) and (PagePtrArray[PP].CurItem=Node^.Count-1) do
    begin
      Dec(PP);
      Node:=GetPage(@Self, PagePtrArray[PP].Ref);
    end;
    if PagePtrArray[PP].CurItem<Node^.Count-1 then
    begin
       inc(PagePtrArray[PP].CurItem);
       NextKey:=Node^.RecordNumber(PagePtrArray[PP].CurItem)
    end else PP:=0;
  end;
  if PP=0 then iStatus:=iEofFind;}
end;

function TNDX_Index.PrevKey:Longint;
begin
end;

function TNDX_Index.Unique:Boolean;
begin
  Unique:=PNDXStream(IndexStream)^.Header.Unique;
end;

function TNDX_Index.GetNode(ARef:Longint):PNode;
begin
  GetNode:=New(PNDXNode, Init(IndexStream, ARef, NDXPageSize));
end;

function TNDX_Index.CurentKey:longint;
var Node:PNdxNode;
begin
  Node:=GetPage(@Self, PagePtrArray[PP].Ref);
  CurentKey:=Node^.RecordNumber(PagePtrArray[PP].CurItem);
end;

procedure TNDX_Index.EmptyIndex;
begin
end;

end.
