{************************************************}
{                                                }
{   Extended objects unit                        }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}
{$I Define.inc}

unit Clases;

interface

uses Objects, Ext_Type;

type
{ TSItem }
  PSItem = ^TSItem;
  TSItem = record
    Value: PString;
    Next: PSItem;
  end;


type
  PProcedureRec = ^TProcedureRec;
  TProcedureRec = record
    Name:PString;
    Ptr:Pointer;
    Code:word;
    Next:PProcedureRec;
  end;

type
  PStrings = ^TStrings;
  TStrings = object(TCollection)
    constructor Init;
    function GetItem(var S: TStream): Pointer; virtual;
    procedure PutItem(var S: TStream; Item: Pointer); virtual;
    procedure FreeItem(Item: Pointer); virtual;
    procedure SaveToFile(AFileName:FNameStr);
    procedure LoadFromFile(AFileName:FNameStr);
    procedure SaveToStream(var S:TStream);
    procedure LoadFromStream(var S:TStream);
    function Str(AItem:integer):string;
    procedure SetStr(AItem:integer; S:string);
    procedure Add(const S:string);
    procedure Assign(P:PStrings);
    procedure Clear;
  end;

type
  PBLOBObject = ^TBLOBObject;
  TBLOBObject = object(TObject)
    Ptr:Pointer;
    Size:word;
    constructor Init(ASize:word);
    destructor Done;virtual;
    constructor Load(var S:TStream);
    procedure Store(var S:TStream);
    procedure LoadFromStream(P:PStream);
    procedure SaveToStream(P:PStream);
  end;

type
  TCreateObject = function(S:PStream):PObject;
  PObjectRec = ^TObjectRec;
  TObjectRec = record
    Code:word;
    VmtOfs:word;
    CreateProc:TCreateObject;
    Next:PObjectRec;
  end;

type
  PIOProperty = ^TIOProperty;
  PIOPropertyRec = ^TIOPropertyRec;
  TCreateIOProperty=function(S:PStream):PIOProperty;
  TIOPropertyRec = record
    Code:word;
    VmtOfs:word;
    CreateProc:TCreateIOProperty;
    Next:PIOPropertyRec;
  end;

  PWriter = ^TWriter;
  TWriter = object(TCollection)
    Stream:PStream;
    constructor Init(S:PStream);
    destructor Done;virtual;
    procedure WriteString(const Name,Value:string);
    procedure WriteInteger(const Name:string; Value:int32);
    procedure WriteFloat(const Name:string; Value:Double);
    procedure WriteBoolean(const Name:string; Value:boolean);
    procedure WriteObject(const Name:string; Value:PObject);
  end;

  PReader = ^TReader;
  TReader = object(TCollection)
    constructor Init(S:PStream);
    function ReadString(const Name, DefValue:string):string;
    function ReadInteger(const Name:string; DefValue:int32):int32;
    function ReadBoolean(const Name:string; DefValue:boolean):boolean;
    function ReadFloat(const Name:string; DefValue:Double):Double;
    procedure ReadObject(const Name:string; Value:PObject);
  private
    function PropByName(Name:string):Pointer;
  end;

  TIOProperty = object(TObject)
    Owner:PWriter;
    Name:PString;
    constructor Init(const AName:string; AOwner:PWriter);
    constructor LoadBin(S:PStream);
    constructor LoadText(S:PStream);
    destructor Done;virtual;
    procedure SaveToBin(S:PStream);virtual;
    procedure SaveToText(S:PStream);virtual;
    function DataSize:int32;virtual;
  end;

  PStringIOProperty = ^TStringIOProperty;
  TStringIOProperty = object(TIOProperty)
    Value:PString;
    constructor Init(const AName:string; AOwner:PWriter; const AValue:string);
    destructor Done;virtual;
    constructor LoadBin(S:PStream);
    constructor LoadText(S:PStream);
    procedure SaveToBin(S:PStream);virtual;
    procedure SaveToText(S:PStream);virtual;
    function DataSize:int32;virtual;
  end;

  PIntegerIOProperty = ^TIntegerIOProperty;
  TIntegerIOProperty = object(TIOProperty)
    Value:int32;
    constructor Init(const AName:string; AOwner:PWriter; const AValue:int32);
    constructor LoadBin(S:PStream);
    constructor LoadText(S:PStream);
    procedure SaveToBin(S:PStream);virtual;
    procedure SaveToText(S:PStream);virtual;
    function DataSize:int32;virtual;
  end;

  PFloatIOProperty = ^TFloatIOProperty;
  TFloatIOProperty = object(TIOProperty)
    Value:Double;
    constructor Init(const AName:string; AOwner:PWriter; const AValue:Double);
    constructor LoadBin(S:PStream);
    constructor LoadText(S:PStream);
    procedure SaveToBin(S:PStream);virtual;
    procedure SaveToText(S:PStream);virtual;
    function DataSize:int32;virtual;
  end;

  PBlobIOProperty = ^TBlobIOProperty;
  TBlobIOProperty = object(TIOProperty)
    Value:Pointer;
    Size:word;
    constructor Init(const AName:string; AOwner:PWriter; const AValue:Pointer; const ASize:word);
    destructor Done;virtual;
    constructor LoadBin(S:PStream);
    constructor LoadText(S:PStream);
    procedure SaveToBin(S:PStream);virtual;
    procedure SaveToText(S:PStream);virtual;
    function DataSize:int32;virtual;
  end;

const
  RStrings: TStreamRec = (
    ObjType: 10024;
    {$IFDEF BP_VMTLink}
    VmtLink: Ofs(TypeOf(TStrings)^);
    {$ELSE}
    VmtLink: TypeOf(TStrings);
    {$ENDIF}
    Load: @TStrings.Load;
    Store: @TStrings.Store);

const
  RBLOBObject: TStreamRec = (
    ObjType: 10030;
    {$IFDEF BP_VMTLink}
    VmtLink: Ofs(TypeOf(TBLOBObject)^);
    {$ELSE}
    VmtLink: TypeOf(TBLOBObject);
    {$ENDIF}
    Load: @TBLOBObject.Load;
    Store: @TBLOBObject.Store);

{ SItem routines }
function NewSItem(const Str: String; ANext: PSItem): PSItem;

function GetObject(const FName:FNameStr):PObject;
procedure RegisterClases;

procedure RegisterIOProperty(ACode:word; VmtOfs:word; CreateProc:TCreateIOProperty);
function FindIOProperty(Code:word; S:PStream):PIOProperty;
procedure RegisterProcedure(P:Pointer; ANum:word; AName:string);
procedure StoreProcedure(var S:TStream;{$IFDEF VER70}var{$ENDIF} P:Pointer);
procedure LoadProcedure(var S:TStream;var P:Pointer);
function NameProcedure({$IFDEF VER70}var{$ENDIF} P:Pointer):string;
procedure RegisterError;

implementation
uses Ext_Dos, {Ext_Math,} Ext_Str;

function NewSItem(const Str: String; ANext: PSItem): PSItem;
var
  Item: PSItem;
begin
  New(Item);
  Item^.Value := NewStr(Str);
  Item^.Next := ANext;
  NewSItem := Item;
end;

procedure RegisterError;
begin
  RunError(212);
end;

const
   First:PProcedureRec = nil;
   Registred:Boolean = false;
   FirstProp:PIOPropertyRec = nil;
   FirstObj:PObjectRec = nil;


function FindNumberObject(VmtOfs:word):word;
var P:PObjectRec;
begin
  FindNumberObject:=0;
  P:=FirstObj;
  while P<>nil do
  begin
    if P^.VmtOfs=VmtOfs then
    begin
      FindNumberObject:=P^.Code;
      exit;
    end;
    P:=P^.Next;
  end;
  RegisterError;
end;

function FindObject(Code:word; S:PStream):PObject;
var P:PObjectRec;
begin
  FindObject:=nil;
  P:=FirstObj;
  while P<>nil do
  begin
    if P^.Code=Code then
    begin
      if Assigned(P^.CreateProc) then FindObject:=P^.CreateProc(S);
      exit;
    end;
    P:=P^.Next;
  end;
end;

procedure RegisterObject(ACode:word; VmtOfs:word; CreateProc:TCreateObject);
var P:PObjectRec;
begin
  New(P);
  P^.Code:=ACode;
  P^.VmtOfs:=VmtOfs;
  P^.CreateProc:=CreateProc;
  P^.Next:=FirstObj;
  FirstObj:=P;
end;

function FindNumberProp(VmtOfs:word):word;
var P:PIOPropertyRec;
begin
  FindNumberProp:=0;
  P:=FirstProp;
  while P<>nil do
  begin
    if P^.VmtOfs=VmtOfs then
    begin
      FindNumberProp:=P^.Code;
      exit;
    end;
    P:=P^.Next;
  end;
{  RegisterError;}
end;

function FindIOProperty(Code:word; S:PStream):PIOProperty;
var P:PIOPropertyRec;
begin
  FindIOProperty:=nil;
  P:=FirstProp;
  while P<>nil do
  begin
    if P^.Code=Code then
    begin
      if Assigned(P^.CreateProc) then FindIOProperty:=P^.CreateProc(S);
      exit;
    end;
    P:=P^.Next;
  end;
end;

procedure RegisterIOProperty(ACode:word; VmtOfs:word; CreateProc:TCreateIOProperty);
var P:PIOPropertyRec;
begin
  New(P);
  P^.Code:=ACode;
  P^.VmtOfs:=VmtOfs;
  P^.CreateProc:=CreateProc;
  P^.Next:=FirstProp;
  FirstProp:=P;
end;

procedure RegisterProcedure(P:Pointer; ANum:word; AName:string);
var Rec:PProcedureRec;
begin
  if not Registred then RegisterError;
  Rec:=First;
  while Rec<>nil do
    if Rec^.Code=ANum then RegisterError else Rec:=Rec^.Next;
  New(Rec);
  Rec^.Ptr:=P;
  Rec^.Name:=NewStr(AName);
  Rec^.Code:=ANum;
  Rec^.Next:=First;
  First:=Rec;
end;

procedure StoreProcedure(var S:TStream;{$IFDEF VER70}var{$ENDIF} P:Pointer);
var Rec:PProcedureRec;
    W:Word;
begin
  if not Registred then RegisterError;
  if P<>nil then
  begin
    Rec:=First;
    while Rec<>nil do
      if Rec^.Ptr=P then break else Rec:=Rec^.Next;
    if Rec=nil then RegisterError;
    S.Write(Rec^.Code, SizeOf(Word));
    S.WriteStr(Rec^.Name);
  end
  else
  begin
    w:=0;
    S.Write(W, SizeOf(word));
  end;
end;

procedure LoadProcedure(var S:TStream;var P:Pointer);
var Rec:PProcedureRec;
    W:Word;
    PStr:PString;
begin
  if not Registred then RegisterError;
  S.Read(W, SizeOf(W));
  if W<>0 then
  begin
    PStr:=S.ReadStr;
    Rec:=First;
    while Rec<>nil do
      if Rec^.Code=W then break else Rec:=Rec^.Next;
    if Rec=nil then RegisterError;
    P:=Rec^.Ptr;
    if Rec^.Name=nil then Rec^.Name:=PStr else DisposeStr(PStr);
  end
  else P:=nil;
end;

function NameProcedure({$IFDEF VER70}var{$ENDIF} P:Pointer):string;
var Rec:PProcedureRec;
begin
  if not Registred then RegisterError;
  if P<>nil then
  begin
    Rec:=First;
    while Rec<>nil do
      if Rec^.Ptr=P then break else Rec:=Rec^.Next;
    if (Rec<>nil) and (Rec^.Name<>nil) then NameProcedure:=Rec^.Name^
    else NameProcedure:='';
  end
  else NameProcedure:='';
end;

  {TStrings}
constructor TStrings.Init;
begin
  inherited Init(100,50)
end;

procedure TStrings.Assign(P:PStrings);
var i:integer;
begin
  FreeAll;
  if P<>nil then
  for i:=0 to P^.Count-1 do Add(P^.Str(i));
end;

procedure TStrings.Clear;
begin
  FreeAll;
end;

function TStrings.GetItem(var S: TStream): Pointer;
begin
  GetItem:=S.ReadStr;
end;

procedure TStrings.PutItem(var S: TStream; Item: Pointer);
begin
  S.WriteStr(Item);
end;

procedure TStrings.FreeItem(Item: Pointer);
begin
  DisposeStr(PString(Item));
end;

procedure TStrings.SaveToStream(var S:TStream);
var i:integer;
begin
  for i:=0 to Count-1 do
  begin
    if At(I)<>nil then WriteText(S, Str(I))
    else WriteText(S, '')
  end;
end;

procedure TStrings.LoadFromStream(var S:TStream);
var St:string;
begin
  while S.Status=stOk do
  begin
    St:=ReadText(S);
    if St<>'' then Add(St);
  end;
end;

procedure TStrings.SaveToFile( AFileName:FNameStr);
var S:TDosStream;
begin
  S.Init(AFileName, stCreate);
  SaveToStream(S);
  S.Done;
end;

procedure TStrings.LoadFromFile(AFileName:FNameStr);
var f:text;
    S:String;
begin
  FreeAll;
  System.Assign(f, AFileName);
  Reset(f);
  while not eof(f) do
  begin
    readln(f,s);
    Insert(NewStr(S));
  end;
  close(f);
end;

function TStrings.Str(AItem:integer):string;
begin
  if At(AItem)<>nil then Str:=PString(At(AItem))^ else Str:='';
end;

procedure TStrings.SetStr(AItem:integer; S:string);
var P:PString;
begin
  P:=At(AItem);
  DisposeStr(P);
  AtPut(AItem, NewStr(S));
end;

procedure TStrings.Add(const S:string);
begin
  Insert(NewStr(S));
end;

  { TBLOBObject }

constructor TBLOBObject.Init(ASize:word);
begin
  inherited Init;
  if ASize>65520 then Size:=65520 else Size:=ASize;
{  Size:=MinWord(ASize, 65520);}
  GetMem(Ptr, Size);
  FillChar(Ptr^, Size, 0);
end;

destructor TBLOBObject.Done;
begin
  FreeMem(Ptr, Size);
  inherited Done;
end;

constructor TBLOBObject.Load(var S:TStream);
begin
  inherited Init;
  S.Read(Size, SizeOf(word));
  GetMem(Ptr, Size);
  S.Read(Ptr^, Size);
end;

procedure TBLOBObject.Store(var S:TStream);
begin
  S.Write(Size, SizeOf(word));
  S.Write(Ptr^, Size);
end;

procedure TBLOBObject.LoadFromStream(P:PStream);
begin
  P^.Read(Ptr^, Size);
end;

procedure TBLOBObject.SaveToStream(P:PStream);
begin
  P^.Write(Ptr^, Size);
end;

function GetObject(const FName:FNameStr):PObject;
var S:TDosStream;
begin
  S.Init(FName,stOpenRead);
  GetObject:=S.Get;
  S.Done;
end;

var
  SaveExitProc:Pointer;

procedure DoExit;{$IFNDEF FPC}far;{$ENDIF}
var P:PProcedureRec;
    Prop:PIOPropertyRec;
begin
  P:=First;
  while P<>nil do
  begin
    First:=P^.Next;
    DisposeStr(P^.Name);
    Dispose(P);
    P:=First;
  end;
  Prop:=FirstProp;
  while Prop<>nil do
  begin
    FirstProp:=Prop^.Next;
    Dispose(Prop);
    Prop:=FirstProp;
  end;
  ExitProc:=SaveExitProc;
end;


  { TWriter }
constructor TWriter.Init(S:PStream);
begin
  inherited Init(20,10);
  Stream:=S;
end;

destructor TWriter.Done;
var i:integer;
    W:word;
    P:PIOProperty;
    Size:int32;
begin
  if Stream<>nil then
  begin
    Stream^.Write(Count, SizeOf(Integer));
    for i:=0 to Count-1 do
    begin
      P:=At(i);
      W:=FindNumberProp(Ofs(TypeOf(P^)^));
      Stream^.Write(W, SizeOf(W));
      Size:=P^.DataSize;
      Stream^.Write(Size, SizeOf(int32));
      P^.SaveToBin(Stream);
    end;
  end;
  inherited Done;
end;

procedure TWriter.WriteString(const Name,Value:string);
begin
  Insert(New(PStringIOProperty, Init(Name,@Self,Value)));
end;

procedure TWriter.WriteObject(const Name:string;Value:PObject);
begin
end;

procedure TWriter.WriteInteger(const Name:string; Value:int32);
begin
  Insert(New(PIntegerIOProperty, Init(Name, @Self, Value)));
end;

procedure TWriter.WriteFloat(const Name:string; Value:Double);
begin
end;

procedure TWriter.WriteBoolean(const Name:string; Value:boolean);
begin
  Insert(New(PIntegerIOProperty, Init(Name, @Self, ord(Value))));
end;

  { TReader }
constructor TReader.Init(S:PStream);
var i, W:integer;
    A:word;
    P:PIOProperty;
    L:int32;
begin
  S^.Read(W, SizeOf(integer));
  inherited Init(W, 10);
  for i:=0 to W-1 do
  begin
    S^.Read(A, SizeOf(word));
    S^.Read(L,SizeOf(int32));
    P:=FindIOProperty(A,S);
    if P<>nil then Insert(P)
    else S^.Seek(S^.GetPos+L);
  end;
end;

function TReader.PropByName(Name:string):Pointer;
var i:integer;
begin
  Name:=stUpCase(Name);
  for i:=0 to Count-1 do
  begin
    if PIOProperty(At(i))^.Name^=Name then
    begin
      PropByName:=At(i);
      exit;
    end;
  end;
  PropByName:=nil;
end;

function TReader.ReadString(const Name, DefValue:string):string;
var P:PStringIOProperty;
begin
  P:=PropByName(Name);
  if P=nil then ReadString:=DefValue else ReadString:=DinamicStr(P^.Value);
end;

function TReader.ReadInteger(const Name:string; DefValue:int32):int32;
var P:PIntegerIOProperty;
begin
  P:=PropByName(Name);
  if P=nil then ReadInteger:=DefValue else ReadInteger:=P^.Value;
end;

function TReader.ReadFloat(const Name:string; DefValue:Double):Double;
begin
end;

function TReader.ReadBoolean(const Name:string; DefValue:boolean):boolean;
var P:PIntegerIOProperty;
begin
  P:=PropByName(Name);
  if P=nil then
    ReadBoolean:=DefValue else ReadBoolean:=(P^.Value<>0);
end;

procedure TReader.ReadObject(const Name:string; Value:PObject);
begin
end;

  { TIOProperty }
constructor TIOProperty.Init(const AName:string; AOwner:PWriter);
begin
  inherited Init;
  Owner:=AOwner;
  Name:=NewStr(stUpCase(AName));
end;

destructor TIOProperty.Done;
begin
  DisposeStr(Name);
  inherited Done;
end;

function TIOProperty.DataSize:int32;
begin
  DataSize:=Length(DinamicStr(Name))+1;
end;

procedure TIOProperty.SaveToBin(S:PStream);
begin
  S^.WriteStr(Name);
end;

procedure TIOProperty.SaveToText(S:PStream);
begin
end;

constructor TIOProperty.LoadBin(S:PStream);
begin
  inherited Init;
  Name:=S^.ReadStr;
end;

constructor TIOProperty.LoadText(S:PStream);
begin
  inherited Init;
end;

  {TStringIOProperty}
constructor TStringIOProperty.Init(const AName:string; AOwner:PWriter; const AValue:string);
begin
  inherited Init(AName, AOwner);
  Value:=NewStr(AValue);
end;

destructor TStringIOProperty.Done;
begin
  DisposeStr(Value);
  inherited Done;
end;

constructor TStringIOProperty.LoadBin(S:PStream);
begin
  inherited LoadBin(S);
  Value:=S^.ReadStr;
end;

constructor TStringIOProperty.LoadText(S:PStream);
begin
  inherited LoadText(S);
end;

procedure TStringIOProperty.SaveToBin(S:PStream);
begin
  inherited SaveToBin(S);
  S^.WriteStr(Value);
end;

procedure TStringIOProperty.SaveToText(S:PStream);
begin
  inherited SaveToText(S);
end;

function TStringIOProperty.DataSize:int32;
begin
  DataSize:=Inherited DataSize+Length(DinamicStr(Value))+1;
end;


  {TIntegerIOProperty}
constructor TIntegerIOProperty.Init(const AName:string; AOwner:PWriter; const AValue:int32);
begin
  inherited Init(AName, AOwner);
  Value:=AValue;
end;

constructor TIntegerIOProperty.LoadBin(S:PStream);
begin
  inherited LoadBin(S);
  S^.Read(Value, SizeOf(Int32));
end;

constructor TIntegerIOProperty.LoadText(S:PStream);
begin
  inherited LoadText(S);
end;

procedure TIntegerIOProperty.SaveToBin(S:PStream);
begin
  inherited SaveToBin(S);
  S^.Write(Value, SizeOf(Int32));
end;

procedure TIntegerIOProperty.SaveToText(S:PStream);
begin
  inherited SaveToText(S);
end;

function TIntegerIOProperty.DataSize:int32;
begin
  DataSize:=Inherited DataSize+SizeOf(int32);
end;

  {TFloatIOProperty}
constructor TFloatIOProperty.Init(const AName:string; AOwner:PWriter; const AValue:Double);
begin
  inherited Init(AName, AOwner);
  Value:=AValue;
end;
constructor TFloatIOProperty.LoadBin(S:PStream);
begin
  inherited LoadBin(S);
  S^.Read(Value, SizeOf(Double));
end;
constructor TFloatIOProperty.LoadText(S:PStream);
begin
  inherited LoadText(S);
end;
procedure TFloatIOProperty.SaveToBin(S:PStream);
begin
  inherited SaveToBin(S);
  S^.Write(Value, SizeOf(Int32));
end;
procedure TFloatIOProperty.SaveToText(S:PStream);
begin
  inherited SaveToText(S);
end;

function TFloatIOProperty.DataSize:int32;
begin
  DataSize:=Inherited DataSize+SizeOf(Double);
end;

  { TBlobIOProperty }
constructor TBlobIOProperty.Init(const AName:string; AOwner:PWriter; const AValue:Pointer; const ASize:word);
begin
  inherited Init(AName,AOwner);
  Size:=ASize;
  if Size<>0 then
  begin
    GetMem(Value, Size);
    Move(AValue^, Value, Size);
  end;
end;

destructor TBlobIOProperty.Done;
begin
  if Size<>0 then FreeMem(Value, Size);
  inherited Done;
end;

constructor TBlobIOProperty.LoadBin(S:PStream);
begin
  inherited LoadBin(S);
  S^.Read(Size, SizeOf(word));
  if Size<>0 then
  begin
    GetMem(Value, Size);
    S^.Read(Value^, Size);
  end;
end;

constructor TBlobIOProperty.LoadText(S:PStream);
begin
  inherited LoadText(S);
end;

procedure TBlobIOProperty.SaveToBin(S:PStream);
begin
  inherited SaveToBin(S);
  S^.Write(Size, SizeOf(Word));
  if Size<>0 then S^.Write(Value^, Size);
end;

procedure TBlobIOProperty.SaveToText(S:PStream);
begin
end;

function TBlobIOProperty.DataSize:int32;
begin
  DataSize:=inherited DataSize + SizeOf(word) + Size;
end;

function CreateStringIOProperty(S:PStream):PIOProperty;{$IFNDEF FPC}far;{$ENDIF}
begin
  CreateStringIOProperty:=New(PStringIOProperty, LoadBin(S));
end;
function CreateIntegerIOProperty(S:PStream):PIOProperty;{$IFNDEF FPC}far;{$ENDIF}
begin
  CreateIntegerIOProperty:=New(PIntegerIOProperty, LoadBin(S));
end;
function CreateFloatIOProperty(S:PStream):PIOProperty;{$IFNDEF FPC}far;{$ENDIF}
begin
  CreateFloatIOProperty:=New(PFloatIOProperty, LoadBin(S));
end;
function CreateBlobIOProperty(S:PStream):PIOProperty;{$IFNDEF FPC}far;{$ENDIF}
begin
  CreateBlobIOProperty:=New(PBlobIOProperty, LoadBin(S));
end;

procedure RegisterClases;
begin
  RegisterType(RStrings);
  RegisterType(RBLOBObject);
  SaveExitProc:=ExitProc;
  ExitProc:=@DoExit;
  Registred:=true;
{$IFDEF VER70}
  RegisterIOProperty(1, Ofs(TypeOf(TStringIOProperty)^), CreateStringIOProperty);
  RegisterIOProperty(2, Ofs(TypeOf(TIntegerIOProperty)^), CreateIntegerIOProperty);
  RegisterIOProperty(3, Ofs(TypeOf(TFloatIOProperty)^), CreateFloatIOProperty);
  RegisterIOProperty(4, Ofs(TypeOf(TBlobIOProperty)^), CreateBlobIOProperty);
{$ENDIF}
end;

end.