unit Ext_Arh;
interface
uses Objects, Ext_Dos, Dos;
const
  arhOk                 = 0;
  arhFileNotOpen        = 1;
  arhEof                = 2;

type
  PArhObject = ^TArhObject;
  TArhObject = object(TObject)
    ArhStatus:word;
    constructor Init(AFileName:FNameStr);
    destructor Done;virtual;
    procedure FindFirst(var Rec:TSearchRec);virtual;
    procedure FindNext(var Rec:TSearchRec);virtual;
    function ValidArhive:boolean;virtual;
  private
    Stream:PStream;
  end;

type
  TZipFileHeader = record
    Signature :longint;
    Version   :word;
    GPBFlag   :word;
    Compress  :word;
    Date      :word;
    Time      :word;
    CRC32     :longint;
    CSize     :longint;
    USize     :longint;
    FNameLen  :word;
    ExtraField:word;
  end;

type
  PZipArhive = ^TZipArhive;
  TZipArhive = object(TArhObject)
    procedure FindFirst(var Rec:TSearchRec);virtual;
    procedure FindNext(var Rec:TSearchRec);virtual;
    function ValidArhive:boolean;virtual;
  private
    procedure ReadHeader(var Rec:TSearchRec);
  end;

implementation

  {TArhObject}
constructor TArhObject.Init(AFileName:FNameStr);
begin
  inherited Init;
  if FileExists(AFileName) then
  begin
    Stream:=New(PDosStream, Init(AFileName, stOpen));
    if (Stream^.Status<>stOk) or not ValidArhive then
    begin
      Stream^.Free;
      Stream:=nil;
      ArhStatus:=arhFileNotOpen;
    end
  end
  else ArhStatus:=arhFileNotOpen;
end;

destructor TArhObject.Done;
begin
  if Stream<>nil then
  begin
    Stream^.Free;
    Stream:=nil;
  end;
  inherited Done;
end;

procedure TArhObject.FindFirst(var Rec:TSearchRec);
begin abstract;
end;

procedure TArhObject.FindNext(var Rec:TSearchRec);
begin abstract;
end;

function TArhObject.ValidArhive:boolean;
begin
  ValidArhive:=true;
end;

  {TZipArhive}
procedure TZipArhive.FindFirst(var Rec:TSearchRec);
begin
  if (Stream=nil) or (Stream^.Status<>stOk) then
  begin
    ArhStatus:=arhEof;
    exit;
  end;
  Stream^.Seek(0);
  ReadHeader(Rec);
end;

procedure TZipArhive.FindNext(var Rec:TSearchRec);
begin
  if (Stream=nil) or (Stream^.Status<>stOk) then
  begin
    ArhStatus:=arhEof;
    exit;
  end;
  ReadHeader(Rec);
end;

procedure TZipArhive.ReadHeader(var Rec:TSearchRec);
var H:TZipFileHeader;
    C:char;
begin
  FillChar(Rec, SizeOf(TSearchRec), 0);
  Stream^.Read(H, SizeOf(TZipFileHeader));
  if H.FNameLen<>0 then
  begin
    repeat
      Stream^.Read(C, SizeOf(Char));
      if C<>#0 then  Rec.Name:=Rec.Name+C;
    until (Length(Rec.Name)=H.FNameLen) or (C=#0);
    if C<>#0 then
    begin
      Stream^.Seek(Stream^.GetPos+H.CSize+H.ExtraField);
      Rec.Size:=H.USize;
      Rec.Time:=H.Date+H.Time*Longint(256*256);
    end
    else ArhStatus:=arhEof;
  end
  else ArhStatus:=arhEof;
end;

function TZipArhive.ValidArhive:boolean;
var S:array[1..2] of char;
begin
  ValidArhive:=false;
  if (Stream=nil) or (Stream^.Status<>stOk) then exit;
  Stream^.Read(S,2);
  ValidArhive:=S='PK';
end;
end.