unit Intrpret;

interface
uses Objects, Parser, Clases;
const
  sErrorLine = 'ப ';
type
  PVariables = ^TVariables;
  PInterpretator = ^TInterpretator;
  TSetValueEvent=procedure(Sender:PInterpretator; const Name:String; var Value:TParserValue);
  TInterpretator = object(TObject)
    OnSetValue:TSetValueEvent;
    Variables:PVariables;
    Parser: TParser;
    Error: Boolean;
    constructor Init;
    destructor Done; virtual;
    procedure SetValue(const Name: String;var Value: TParserValue); virtual;
    procedure PrepareScript(var MemoFrom, MemoTo, MemoErr: TStrings); virtual;
    procedure DoScript(var Memo: TStrings); virtual;
  end;

  TVariables = object(TObject)
  private
    FList: PCollection;
  public
    procedure SetVariable(AName: String;var AValue: TParserValue);
    function Variable(AName: String): PParserValue;
    procedure SetValue(Index: Integer; var AValue: TParserValue);
    function _Value(Index: Integer): PParserValue;
    function _Name(Index: Integer): String;
    function Count: Integer;
    constructor Init;
    destructor Done;virtual;
    procedure Clear;
    procedure Delete(Index: Integer);
    function IndexOf(AName: String): Integer;
  end;

procedure GetValue(Owner:PObject; Sender:PParser; const Name:String; var Value:TParserValue);far;
procedure DoFunction(Owner:PObject; Sender:PParser; const Name:String; var p1,p2,p3, Value:string);far;

implementation
uses Ext_Str, Ext_Math;

type
  TCharArray = Array[0..31999] of Char;
  PCharArray = ^TCharArray;
  lrec = record
    name: String[16];
    n: Integer;
  end;

  PVariable = ^TVariable;
  TVariable = record
    Name: PString;
    Value: TParserValue;
  end;

const
  ttIf    = #1;
  ttGoto  = #2;
  ttProc  = #3;

var
  labels: Array[0..100] of lrec;
  labc: Integer;

{------------------------------------------------------------------------------}
constructor TVariables.Init;
begin
  inherited Init;
  FList := New(PCollection, Init(20,20));
end;

destructor TVariables.Done;
begin
  Clear;
  Dispose(FList, Done);
  inherited Done;
end;

procedure TVariables.Clear;
begin
  while FList^.Count > 0 do Delete(0);
end;

procedure TVariables.SetVariable(AName: String;var AValue: TParserValue);
var
  i: Integer;
  p: PVariable;
begin
  for i := 0 to FList^.Count - 1 do
    if AnsiCompareText(PVariable(FList^.At(i))^.Name^, stUpCase(AName)) = 0 then
    begin
      PVariable(FList^.At(i))^.Value.Assign(@AValue);
      Exit;
    end;
  GetMem(p, SizeOf(TVariable));
  FillChar(p^, SizeOf(TVariable), 0);
  p^.Name := NewStr(stUpCase(AName));
  p^.Value.Init;
  p^.Value.Assign(@AValue);
  FList^.Insert(p);
end;

function TVariables.Variable(AName: String): PParserValue;
var
  i: Integer;
begin
  Variable := nil;
  AName:=stUpCase(AName);
  for i := 0 to FList^.Count - 1 do
    if AnsiCompareText(PVariable(FList^.At(i))^.Name^, AName) = 0 then
    begin
      Variable := @(PVariable(FList^.At(i))^.Value);
      break;
    end;
end;

procedure TVariables.SetValue(Index: Integer;var AValue: TParserValue);
begin
  if (Index < 0) or (Index >= FList^.Count) then Exit;
  PVariable(FList^.At(Index))^.Value.Assign(@AValue);
end;

function TVariables._Value(Index: Integer): PParserValue;
var R:PParserValue;
begin
  _Value := nil;
  if (Index < 0) or (Index >= FList^.Count) then Exit;
  _Value := @(PVariable(FList^.At(Index))^.Value);
end;

function TVariables.IndexOf(AName: String): Integer;
var
  i: Integer;
begin
  IndexOf := -1;
  for i := 0 to FList^.Count - 1 do
    if AnsiCompareText(PVariable(FList^.At(i))^.Name^, AName) = 0 then
    begin
      IndexOf := i;
      break;
    end;
end;

function TVariables.Count: Integer;
begin
  Count := FList^.Count;
end;

function TVariables._Name(Index: Integer): String;
begin
  _Name := '';
  if (Index < 0) or (Index >= FList^.Count) then Exit;
  _Name := PVariable(FList^.At(Index))^.Name^;
end;

procedure TVariables.Delete(Index: Integer);
var
  p: PVariable;
begin
  if (Index < 0) or (Index >= FList^.Count) then Exit;
  p := FList^.At(Index);
  DisposeStr(p^.Name);
  p^.Value.Done;
  FreeMem(p, SizeOf(TVariable));
  FList^.AtDelete(Index);
end;

{------------------------------------------------------------------------------}
function Remain(S: String; From: Integer): String;
begin
  Remain := Copy(s, From, Length(s) - 1);
end;

function GetIdentify(const s: String; var i: Integer): String;
var
  k: Integer;
begin
  while (i <= Length(s)) and (s[i] = ' ') do
    Inc(i);
  k := i;
  while (i <= Length(s)) and (s[i] <> ' ') do
    Inc(i);
  GetIdentify := Copy(s, k, i - k);
end;

{-----------------------------------------------------------------------------}
constructor TInterpretator.Init;
begin
  inherited Init;
  Parser.Init(@Self);
  Parser.OnGetValue := GetValue;
{  Parser.OnFunction := DoFunction;}
  Variables:=New(PVariables, Init);
end;

destructor TInterpretator.Done;
begin
  Parser.Done;
  Variables^.Free;
  inherited Done;
end;

procedure TInterpretator.PrepareScript(var MemoFrom, MemoTo, MemoErr: TStrings);
var
  i, j, cur, lastp: Integer;
  s, bs: String;
  len: Integer;
  buf: PCharArray;

procedure DoCommand; forward;
procedure DoBegin; forward;
procedure DoIf; forward;
procedure DoRepeat; forward;
procedure DoWhile; forward;
procedure DoGoto; forward;
procedure DoEqual; forward;
procedure DoExpression; forward;
procedure DoSExpression; forward;
procedure DoTerm; forward;
procedure DoFactor; forward;
procedure DoVariable; forward;
procedure DoConst; forward;
procedure DoLabel; forward;
procedure DoFunc; forward;
procedure DoFuncId; forward;

  function last: Integer;
  begin
    Last := MemoTo.Count;
  end;

  function CopyArr(cur, n: Integer): String;
  var S:string;
  begin
{    SetLength(Result, n);
    Move(buf^[cur], Result[1], n);}
    S[0]:=char(n);
    Move(buf^[cur], S[1], n);
    CopyArr:=S;
  end;

  procedure AddLabel(s: String; n: Integer);
  var
    i: Integer;
    f: Boolean;
  begin
    f := True;
    for i := 0 to labc - 1 do
      if labels[i].name = s then f := False;
    if f then
    begin
      labels[labc].name := s;
      labels[labc].n := n;
      Inc(labc);
    end;
  end;

  procedure SkipSpace;
  begin
    while (buf^[cur] = ' ') and (cur < len) do Inc(cur);
  end;

  function GetToken: String;
  var
    j: Integer;
  begin
    SkipSpace;
    j := cur; Inc(cur);
    while (buf^[cur] > ' ') and (cur < len) do Inc(cur);
    GetToken:= AnsiUpperCase(CopyArr(j, cur - j));
  end;

  procedure AddError(s: String);
  var
    i, j, c: Integer;
    s1: String;
  begin
    Error := True;
    cur := lastp;
    SkipSpace;
    c := 0;
    for i := 0 to cur do
      if buf^[i] > ' ' then Inc(c);
    i := 0;
    j := 1;
    while c > 0 do
    begin
      s1 := MemoFrom.Str(i);
      j := 1;
      while (j <= Length(s1)) and (c > 0) do
      begin
        if s1[j] = '{' then break;
        if s1[j] > ' ' then Dec(c);
        Inc(j);
      end;
      if c = 0 then break;
      Inc(i);
    end;
    MemoErr.Add(sErrorLine + IntToStr(i + 1) + '/' + IntToStr(j - 1) + ': ' + s);
  end;

  procedure ProcessBrackets(var i: Integer);
  var
    c: Integer;
    fl1, fl2: Boolean;
  begin
    fl1 := True; fl2 := True; c := 0;
    Dec(i);
    repeat
      Inc(i);
      if fl1 and fl2 then
        if buf^[i] = '[' then
          Inc(c) else
          if buf^[i] = ']' then Dec(c);
      if fl1 then
        if buf^[i] = '"' then fl2 := not fl2;
      if fl2 then
        if buf^[i] = '''' then fl1 := not fl1;
    until (c = 0) or (i >= len);
  end;

  {----------------------------------------------}
  procedure DoDigit;
  begin
    while (buf^[cur] = ' ') and (cur < len) do Inc(cur);
    if buf^[cur] in ['0'..'9'] then
      while (buf^[cur] in ['0'..'9']) and (cur < len) do Inc(cur)
    else Error := True;
  end;

  procedure DoBegin;
  label 1;
  begin
  1:DoCommand;
    if Error then Exit;
    lastp := cur;
    bs := GetToken;
    if (bs = '') or (bs[1] = ';') then
    begin
      cur := cur - Length(bs) + 1;
      goto 1;
    end
    else if Pos('END', bs) = 1 then cur := cur - Length(bs) + 3
    else AddError(' ॡ ";"  "end"');
  end;

  procedure DoIf;
  var
    nsm, nl, nl1: Integer;
  begin
    nsm := cur;
    DoExpression;
    if Error then Exit;
    bs := ttIf + '  ' + CopyArr(nsm, cur - nsm);
    nl := last;
    MemoTo.Add(bs);
    lastp := cur;
    if GetToken = 'THEN' then
    begin
      DoCommand;
      if Error then Exit;
      nsm := cur;
      if GetToken = 'ELSE' then
      begin
        nl1 := last;
        MemoTo.Add(ttGoto + '  ');
        bs := MemoTo.Str(nl);
        bs[2] := Chr(last);
        bs[3] := Chr(last div 256);
        MemoTo.SetStr(nl,bs);
        DoCommand;
        bs := MemoTo.Str(nl1);
        bs[2] := Chr(last);
        bs[3] := Chr(last div 256);
        MemoTo.SetStr(nl1,bs);
      end
      else
      begin
        bs := MemoTo.Str(nl);
        bs[2] := Chr(last);
        bs[3] := Chr(last div 256);
        MemoTo.SetStr(nl,bs);
        cur := nsm;
      end;
    end
    else AddError(' ॡ "then"');
  end;

  procedure DoRepeat;
  label 1;
  var
    nl, nsm: Integer;
  begin
    nl := last;
  1:DoCommand;
    if Error then Exit;
    lastp := cur;
    bs := GetToken;
    if bs[1] = ';' then
    begin
      cur := cur - Length(bs) + 1;
      goto 1;
    end
    else if bs = 'UNTIL' then
    begin
      nsm := cur;
      DoExpression;
      MemoTo.Add(ttIf + Chr(nl) + Chr(nl div 256) + CopyArr(nsm, cur - nsm));
    end
     else AddError(' ॡ ";"  "until"');
  end;

  procedure DoWhile;
  var
    nl, nsm: Integer;
  begin
    nl := last;
    nsm := cur;
    DoExpression;
    if Error then Exit;
    MemoTo.Add(ttIf + '  ' + CopyArr(nsm, cur - nsm));
    lastp := cur;
    if GetToken = 'DO' then
    begin
      DoCommand;
      MemoTo.Add(ttGoto + Chr(nl) + Chr(nl div 256));
      bs := MemoTo.Str(nl);
      bs[2] := Chr(last);
      bs[3] := Chr(last div 256);
      MemoTo.SetStr(nl,bs);
    end
    else AddError(' ॡ "do"');
  end;

  procedure DoGoto;
  var
    nsm: Integer;
  begin
    SkipSpace;
    nsm := cur;
    lastp := cur;
    DoDigit;
    if Error then AddError('⪠  goto   ᫮');
    MemoTo.Add(ttGoto + Trim(CopyArr(nsm, cur - nsm)));
  end;

  procedure DoEqual;
  var
    s: String;
    n, nsm: Integer;
  begin
    nsm := cur;
    DoVariable;
    s := Trim(CopyArr(nsm, cur - nsm)) + ' ';
    lastp := cur;
    bs := GetToken;
    if (bs = ';') or (bs = '') or (bs = #0) then
    begin
      s := Trim(CopyArr(nsm, lastp - nsm));
      MemoTo.Add(ttProc + s + '(0)');
      cur := lastp;
    end
    else if Pos(':=', bs) = 1 then
    begin
      cur := cur - Length(bs) + 2;
      nsm := cur;
      DoExpression;
      n := Pos('[', s);
      if n <> 0 then
      begin
        s := ttProc + 'SETARRAY(' + Copy(s, 1, n - 1) + ', ' +
          Copy(s, n + 1, Length(s) - n - 2) + ', ' + CopyArr(nsm, cur - nsm) + ')';
      end
      else
        s := s + CopyArr(nsm, cur - nsm);
      MemoTo.Add(s);
    end
    else
      AddError(' ॡ ":="');
  end;
  {-------------------------------------}
  procedure DoExpression;
  var
    nsm: Integer;
  begin
    DoSExpression;
    nsm := cur;
    bs := GetToken;
    if (Pos('>=', bs) = 1) or (Pos('<=', bs) = 1) or (Pos('<>', bs) = 1) then
    begin
      cur := cur - Length(bs) + 2;
      DoSExpression;
    end
    else if (bs[1] = '>') or (bs[1] = '<') or (bs[1] = '=') then
    begin
      cur := cur - Length(bs) + 1;
      DoSExpression;
    end
    else cur := nsm;
  end;

  procedure DoSExpression;
  var
    nsm: Integer;
  begin
    DoTerm;
    nsm := cur;
    bs := GetToken;
    if (bs[1] = '+') or (bs[1] = '-') then
    begin
      cur := cur - Length(bs) + 1;
      DoSExpression;
    end
    else if Pos('OR', bs) = 1 then
    begin
      cur := cur - Length(bs) + 2;
      DoSExpression;
    end
    else cur := nsm;
  end;

  procedure DoTerm;
  var
    nsm: Integer;
  begin
    DoFactor;
    nsm := cur;
    bs := GetToken;
    if (bs[1] = '*') or (bs[1] = '/') then
    begin
      cur := cur - Length(bs) + 1;
      DoTerm;
    end
    else if (Pos('AND', bs) = 1) or (Pos('MOD', bs) = 1) then
    begin
      cur := cur - Length(bs) + 3;
      DoTerm;
    end
    else cur := nsm;
  end;

  procedure DoFactor;
  var
    nsm: Integer;
  begin
    nsm := cur;
    bs := GetToken;
    if bs[1] = '(' then
    begin
      cur := cur - Length(bs) + 1;
      DoExpression;
      SkipSpace;
      lastp := cur;
      if buf^[cur] = ')' then Inc(cur)
      else AddError(' ॡ ")"');
    end
    else if bs[1] = '[' then
    begin
      cur := cur - Length(bs);
      ProcessBrackets(cur);
      SkipSpace;
      lastp := cur;
      if buf^[cur] = ']' then Inc(cur)
      else AddError(' ॡ "]"');
    end
    else if (bs[1] = '+') or (bs[1] = '-') then
    begin
      cur := cur - Length(bs) + 1;
      DoExpression;
    end
    else if bs = 'NOT' then
    begin
      cur := cur - Length(bs) + 3;
      DoExpression;
    end
    else
    begin
      cur := nsm;
      DoVariable;
      if Error then
      begin
        Error := False;
        cur := nsm;
        DoConst;
        if Error then
        begin
          Error := False;
          cur := nsm;
          DoFunc;
        end;
      end;
    end;
  end;

  procedure DoVariable;
  begin
    SkipSpace;
    if (buf^[cur] in ['a'..'z', 'A'..'Z']) then
    begin
      Inc(cur);
      while buf^[cur] in ['0'..'9', '_', '.', 'A'..'Z', 'a'..'z'] do Inc(cur);
      if buf^[cur] = '(' then Error := True;
      if buf^[cur] = '[' then
      begin
        Inc(cur);
        DoExpression;
        if buf^[cur] <> ']' then
          Error := True else
          Inc(cur);
      end;
    end
    else Error := True;
  end;

  procedure DoConst;
  begin
    SkipSpace;
    if buf^[cur] = #$27 then
    begin
      Inc(cur);
      while (buf^[cur] <> #$27) and (cur < len) do Inc(cur);
      if cur = len then Error := True
      else Inc(cur);
    end
    else
    begin
      DoDigit;
      if buf^[cur] = '.' then
      begin
        Inc(cur);
        DoDigit;
      end;
    end;
  end;

  procedure DoLabel;
  begin
    DoDigit;
    if buf^[cur] = ':' then Inc(cur)
    else Error := True;
  end;

  procedure DoFunc;
  label 1;
  begin
    DoFuncId;
    if buf^[cur] = '(' then
    begin
      Inc(cur);
  1:  DoExpression;
      lastp := cur;
      SkipSpace;
      if buf^[cur] = ',' then
      begin
        Inc(cur);
        goto 1;
      end
      else if buf^[cur] = ')' then Inc(cur)
      else AddError(' ॡ ","  ")"');
    end;
  end;

  procedure DoFuncId;
  label 1;
  begin
    SkipSpace;
    if buf^[cur] in ['A'..'Z', 'a'..'z'] then
      while buf^[cur] in ['0'..'9', '_', 'A'..'Z', 'a'..'z'] do Inc(cur)
    else Error := True;
  end;

  procedure DoCommand;
  label 1;
  var
    nsm: Integer;
  begin
  1:Error := False;
    nsm := cur;
    lastp := cur;
    bs := GetToken;
    if bs = 'BEGIN' then DoBegin
    else if bs = 'IF' then DoIf
    else if bs = 'REPEAT' then DoRepeat
    else if bs = 'WHILE' then DoWhile
    else if bs = 'GOTO' then DoGoto
    else if Pos('END', bs) = 1 then
    begin
      cur := nsm;
      Error := False;
    end
    else
    begin
      cur := nsm;
      DoLabel;
      if Error then
      begin
        Error := False;
        cur := nsm;
        DoVariable;
        if not Error then
        begin
          cur := nsm;
          DoEqual;
        end
        else
        begin
          cur := nsm;
          Error := False;
          DoExpression;
          MemoTo.Add(ttProc + Trim(CopyArr(nsm, cur - nsm)));
        end;
      end
      else
      begin
        AddLabel(Trim(CopyArr(nsm, cur - nsm)), last);
        goto 1;
      end;
    end;
  end;
var ii:integer;
begin
  Error := False;
  GetMem(buf, 32000);
  FillChar(buf^, 32000, 0);
  len := 0;
  for i := 0 to MemoFrom.Count - 1 do
  begin
    s := ' ' + MemoFrom.Str(i);
    if Pos('//', s) <> 0 then Delete(S,Pos('//', s),255);
    if Pos('{', s) <> 0 then Delete(S,Pos('{', s),255);
    ii:=1;
    while ii<Length(S) do
    begin
      if S[ii]='''' then
      begin
        inc(ii);
        while (ii<Length(S)) and (S[ii]<>'''') do inc(ii);
        inc(ii);
      end
      else
      if S[ii]=#9 then S[ii]:=' '
      else
      if (ii<Length(S)) and (S[ii]=' ') and (S[ii+1]=' ') then Delete(S,ii,1)
      else inc(ii);
    end;
    Move(s[1], buf^[len], Length(s));
    Inc(len, Length(s));
  end;
  cur := 0; labc := 0;
  MemoTo.Clear;
  MemoErr.Clear;
  if len > 0 then
    DoCommand;
  FreeMem(buf, 32000);
  for i := 0 to MemoTo.Count - 1 do
    if Copy(MemoTo.Str(i),1,1) = ttGoto then
    begin
      s := Remain(MemoTo.Str(i), 2) + ':';
      for j := 0 to labc do
        if labels[j].name = s then
        begin
          s := MemoTo.Str(i);
          s[2] := Chr(labels[j].n);
          s[3] := Chr(labels[j].n div 256);
          MemoTo.SetStr(i, s);
          break;
        end;
    end
    else
    if Copy(MemoTo.Str(i),1,1) = ttIf then
    begin
      s := Parser.Str2OPZ(Remain(MemoTo.Str(i), 4));
      MemoTo.SetStr(i,Copy(MemoTo.Str(i), 1, 3) + s);
    end
    else if Copy(MemoTo.Str(i),1,1) = ttProc then
    begin
      s := Parser.Str2OPZ(Remain(MemoTo.Str(i), 2));
      MemoTo.SetStr(i,Copy(MemoTo.Str(i), 1, 1) + s);
    end
    else
    begin
      j := 1;
      GetIdentify(MemoTo.Str(i), j);
      len := j;
      s := Parser.Str2OPZ(Remain(MemoTo.Str(i), j));
      MemoTo.SetStr(i,Copy(MemoTo.Str(i), 1, len) + s);
    end;
end;

procedure TInterpretator.DoScript(var Memo: TStrings);
var
  i, j: Integer;
  s, s1: String;
  Result:TParserValue;
begin
  i := 0;
  Result.Init;
  while i < Memo.Count do
  begin
    s := Memo.Str(i);
    j := 1;
    if s[1] = ttIf then
    begin
      Parser.CalcOPZ(Remain(s, 4), Result);
      if not Result.AsBoolean  then
      begin
        i := Ord(s[2]) + Ord(s[3]) * 256;
        continue;
      end;
    end
    else if s[1] = ttGoto then
    begin
      i := Ord(s[2]) + Ord(s[3]) * 256;
      continue;
    end
    else
    if s[1] = ttProc then
    begin
      Result.Clear;
      Parser.CalcOPZ(Remain(s, 2),Result)
    end
    else
    begin
      s1 := GetIdentify(s, j);
      Result.Clear;
      Parser.CalcOPZ(Remain(s, j), Result);
      SetValue(s1, Result);
    end;
    Inc(i);
  end;
  Result.Done;
end;

procedure TInterpretator.SetValue(const Name: String;var Value: TParserValue);
begin
{/ abstract method}
  if Assigned(OnSetValue) then OnSetValue(@Self, Name, Value)
  else Variables^.SetVariable(Name, Value);
end;

procedure GetValue(Owner:PObject; Sender:PParser; const Name:String; var Value:TParserValue);
var I:PInterpretator absolute Owner;
begin
   Value.Assign(I^.Variables^.Variable(Name));
end;

procedure DoFunction(Owner:PObject; Sender:PParser; const Name:String; var p1,p2,p3, Value:string);
var R:TParserValue;
    I:PInterpretator absolute Owner;
begin
end;

end.
